var async = require("async"), path = require("path"), util = require("util"), http = require("http"), EventEmitter = require("events").EventEmitter, express = require("express"), io = require("socket.io"), Errors = require("./errors"), Const = require("./const");

var Server = module.exports = function(testCafe) {
    EventEmitter.call(this);
    this.app = express();
    this.appServer = http.createServer(this.app);
    this.config = testCafe.config;
    this.port = this.config.controlPanelPort;
    this.errMsgBuilder = testCafe.errMsgBuilder;
    this.vfs = testCafe.vfs;
    this.testRunner = testCafe.testRunner;
    this.api = testCafe;
    this.recorder = testCafe.recorder;
    this.socketIO = null;
    this.testsViewFirstRun = true;
    this.trialExpired = null;
    this.installTime = -1;
    this.remainingTime = null;
    this.checkTrialTimeout = null;
    this.version = null;
    this.checkUpdateTimeout = null;
    this.app.use(express.bodyParser()).use("/assets", express.static(path.join(__dirname, "../../_compiled_/control_panel"))).use("/ace", express.static(path.join(__dirname, "../../_compiled_/control_panel/ace"))).use(express.favicon(path.join(__dirname, "../../_compiled_/control_panel/images/favicon.ico"))).set("view engine", "ejs").set("view options", {
        layout: false
    }).set("views", path.join(__dirname, "../views"));
    var controlPanel = this;
    this.appServer.on("error", function(err) {
        if (err.code === "EADDRINUSE") {
            controlPanel.emit("fatalError", {
                code: Errors.CONTROL_PANEL_PORT_IS_ALREADY_IN_USE,
                port: controlPanel.port
            });
        }
    });
};

util.inherits(Server, EventEmitter);

Server.prototype.start = function() {
    var controlPanel = this;
    this.socketIO = io.listen(this.appServer, {
        log: false
    });
    this.appServer.listen(this.port);
    this.config.getCurrentVersion(function(curVersion) {
        controlPanel.version = {
            curVersion: curVersion,
            newVersion: null
        };
        controlPanel._checkVersion();
        controlPanel._setupSockets();
        controlPanel._setupRoutes();
    });
};

Server.prototype.close = function() {
    if (this.checkTrialTimeout) {
        clearTimeout(this.checkTrialTimeout);
    }
    if (this.checkUpdateTimeout) {
        clearTimeout(this.checkUpdateTimeout);
    }
    this.appServer.close();
};

Server.prototype._checkVersion = function() {
    var controlPanel = this;
    this.config.checkUpdates(function(curVersion, newVersion) {
        controlPanel.version = {
            curVersion: curVersion,
            newVersion: newVersion
        };
    });
    this.checkUpdateTimeout = setTimeout(function() {
        controlPanel._checkVersion.call(controlPanel);
    }, Const.VERSION_CHECK_TIME);
};

Server.prototype._checkTrial = function() {
    var currDate = new Date().getTime(), controlPanel = this;
    this.remainingTime = Const.TRIAL_TIME - (currDate - this.installTime);
    if (this.remainingTime <= 0 || this.remainingTime > Const.TRIAL_TIME) {
        this.trialExpired = true;
    }
    this.checkTrialTimeout = setTimeout(function() {
        controlPanel._checkTrial.call(controlPanel);
    }, Const.TRIAL_CHECK_TIME);
};

Server.prototype._setupRoutes = function() {
    var controlPanel = this, GET = function(route, handler) {
        controlPanel.app.get(route, handler.bind(controlPanel));
    }, POST = function(route, handler) {
        controlPanel.app.post(route, handler.bind(controlPanel));
    }, ALL = function(route, handler) {
        controlPanel.app.all(route, handler.bind(controlPanel));
    };
    GET("/", this._indexPageHandler);
    GET("/*", this._cachingPreventionHandler);
    POST("/*", this._cachingPreventionHandler);
    GET("/workers/", this._workersPageHandler);
    GET("/workers", this._workersPageHandler);
    ALL("/tests/*/", this._testBrowserHandler);
    ALL("/tests/*", this._testBrowserHandler);
    ALL("/tests", this._testBrowserHandler);
    POST("/tests_run/", this._runTests);
    POST("/tests_run", this._runTests);
    GET("/worker/add/:name/", this._addWorker);
    GET("/worker/add/:name", this._addWorker);
    GET("/worker/idle/:name/", this._workerIdleHandler);
    GET("/worker/idle/:name", this._workerIdleHandler);
    GET("/worker/list/", this._listWorkers);
    GET("/worker/list", this._listWorkers);
    POST("/worker/heartbeat/", this._workerHeartbeatHandler);
    POST("/worker/heartbeat", this._workerHeartbeatHandler);
    GET("/tasks/:taskID/", this._getDetailTaskReport);
    GET("/tasks/:taskID", this._getDetailTaskReport);
    GET("/tasks/", this._listTaskReports);
    GET("/tasks", this._listTaskReports);
    GET("/tasks_rev/:uid/", this._getTaskReportRevision);
    GET("/tasks_rev/:uid", this._getTaskReportRevision);
    POST("/get_fixture_code/", this._getFixtureCode);
    POST("/get_fixture_code", this._getFixtureCode);
    POST("/save_fixture_code/", this._saveFixtureCode);
    POST("/save_fixture_code", this._saveFixtureCode);
    POST("/rename_dir/", this._renameDir);
    POST("/rename_dir", this._renameDir);
    POST("/delete_dir/", this._deleteDir);
    POST("/delete_dir", this._deleteDir);
    POST("/create_fixture/", this._createFixture);
    POST("/create_fixture", this._createFixture);
    POST("/create_fixture_in_dir/", this._createFixtureInDir);
    POST("/create_fixture_in_dir", this._createFixtureInDir);
    POST("/edit_fixture/", this._editFixture);
    POST("/edit_fixture", this._editFixture);
    POST("/delete_fixture/", this._deleteFixture);
    POST("/delete_fixture", this._deleteFixture);
    POST("/rename_test/", this._renameTest);
    POST("/rename_test", this._renameTest);
    POST("/delete_test/", this._deleteTest);
    POST("/delete_test", this._deleteTest);
    GET("/dirs_by_path/", this._getDirsByPath);
    GET("/dirs_by_path", this._getDirsByPath);
    POST("/set_tests_dir/", this._setTestsDir);
    POST("/set_tests_dir", this._setTestsDir);
    POST("/recording/start/", this._record);
    POST("/recording/start", this._record);
    GET("/options/", this._getOptions);
    GET("/options", this._getOptions);
    GET("/files_by_path/", this._getFilesByPath);
    GET("/files_by_path", this._getFilesByPath);
    POST("/browser/configure/", this._configureBrowsers);
    POST("/browser/configure", this._configureBrowsers);
    POST("/browsers/detect/", this._detectBrowsers);
    POST("/browsers/detect", this._detectBrowsers);
    POST("/network/save/", this._configureNetwork);
    POST("/network/save", this._configureNetwork);
    GET("/check_updates/", this._checkUpdates);
    GET("/check_updates", this._checkUpdates);
    ALL("*", this._page404Handler);
};

Server.prototype._cachingPreventionHandler = function(req, res, next) {
    var controlPanel = this;
    res.set("cache-control", "no-cache, no-store, must-revalidate");
    function trialHandler() {
        if (controlPanel.trialExpired && Const.CLOSED_PAGE_PATTERN.test(req.originalUrl)) {
            res.render("trial_expired");
        } else {
            next();
        }
    }
    if (controlPanel.installTime < 0) {
        this.config.getInstallDate(function(val) {
            if (val) {
                controlPanel.installTime = val;
                controlPanel._checkTrial();
                trialHandler();
            } else {
                controlPanel.installTime = 0;
                next();
            }
        });
    } else {
        trialHandler();
    }
};

Server.prototype._indexPageHandler = function(req, res) {
    res.statusCode = 301;
    res.setHeader("Location", "/tests/");
    res.end();
};

Server.prototype._workersPageHandler = function(req, res) {
    var isUpdateRequest = req.param("workersUpdate"), currentWorker = req.param("worker") || "", workers = this.testRunner.workerPool.getWorkersInfo(), connectedWorkerSuggestionName = this.testRunner.workerPool.getWorkerNameFromUserAgent(req.headers["user-agent"]);
    Object.keys(workers).forEach(function(workerIndex) {
        workers[workerIndex].workerType = workers[workerIndex].connected ? Const.CONNECTED_WORKER_TYPE : Const.AUTO_CREATED_WORKER_TYPE;
    });
    if (!isUpdateRequest) {
        res.render("workers", {
            workers: workers,
            connectedWorkerSuggestionName: connectedWorkerSuggestionName,
            currentWorker: currentWorker,
            hostname: this.config.controlPanelHost,
            remainingTime: this.remainingTime
        });
    } else {
        res.render("partial/workers_container", {
            workers: workers
        });
    }
};

Server.prototype._getPathInfo = function(urlPath, callback) {
    var vfs = this.vfs, errMsgBuilder = this.errMsgBuilder, vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], testsBrowser = null, fixtureErrs = [], errs = [], isPage404 = false;
    this.api.listDirectory(vfsPath, function(err, pathInfo) {
        if (err && vfsPath.length) {
            isPage404 = true;
        } else {
            if (!err) {
                if (pathInfo.buildErrs) {
                    for (var errIndex = 0, errsLength = pathInfo.buildErrs.length; errIndex < errsLength; errIndex++) {
                        var error = pathInfo.buildErrs[errIndex];
                        if (error.fixtureUid) {
                            if (!error.filename) {
                                fixtureErrs[error.fixtureUid] = [];
                                continue;
                            }
                            var tmpBasePath = vfs.basePath.replace(/\//g, "\\"), errorMsg = {
                                errText: errMsgBuilder.build(error),
                                filename: error.filename.replace(tmpBasePath, "").replace(/\\/g, "/")
                            };
                            if (fixtureErrs.hasOwnProperty(error.fixtureUid)) {
                                fixtureErrs[error.fixtureUid].push(errorMsg);
                            } else {
                                fixtureErrs[error.fixtureUid] = [ errorMsg ];
                            }
                            errs.push(errorMsg);
                        } else {
                            errs.push(errMsgBuilder.build(error));
                        }
                    }
                }
                testsBrowser = {
                    errs: fixtureErrs,
                    currentUrl: urlPath ? encodeURI([ "/tests/", urlPath, "/" ].join("")) : "/tests/",
                    pathInfo: pathInfo,
                    breadcrumbs: []
                };
                for (var i = 0, url = "/tests/"; i < vfsPath.length; i++) {
                    url = encodeURI([ url, vfsPath[i], "/" ].join(""));
                    testsBrowser.breadcrumbs.push({
                        name: vfsPath[i],
                        url: url
                    });
                }
            }
        }
        callback(errs, testsBrowser, isPage404);
    });
};

Server.prototype._testBrowserHandler = function(req, res) {
    var controlPanel = this, slide = req.param("slide"), rebuild = req.param("rebuild"), completeRecordingUid = req.param("completeRecordingUid"), expandRowFilename = req.param("expand") || "", urlPath = req.params[0] || "", vfs = this.vfs, autoDetectTestsDir = "";
    this._getPathInfo(urlPath, function(errs, testsBrowser, isPage404) {
        var testsDir = vfs.getBasePath();
        if (isPage404) {
            res.status(404);
            res.render("404");
            return;
        }
        var sendRenderData = function() {
            if (slide) {
                res.render("partial/tests_browser", {
                    testsBrowser: testsBrowser,
                    testsDir: testsDir
                });
            } else {
                if (rebuild) {
                    res.render("partial/tests_revision_container", {
                        testsBrowser: testsBrowser,
                        testsDir: testsDir,
                        errMsgs: errs.length ? errs : null
                    });
                } else {
                    res.render("tests", {
                        autoDetectTestsDir: autoDetectTestsDir,
                        testsBrowser: testsBrowser,
                        testsDir: testsDir,
                        expandRowFilename: expandRowFilename,
                        errMsgs: errs.length ? errs : null,
                        completeRecordingUid: completeRecordingUid,
                        remainingTime: controlPanel.remainingTime
                    });
                }
            }
        };
        if (controlPanel.testsViewFirstRun) {
            controlPanel.testsViewFirstRun = false;
            controlPanel.config.readCfgFile(function(cfg, err) {
                if (!err && cfg && cfg.testsDir && cfg.testsDir !== controlPanel.config.testsDir && controlPanel.testsViewFirstRun) {
                    autoDetectTestsDir = cfg.testsDir;
                    sendRenderData();
                } else {
                    sendRenderData();
                }
            });
        } else {
            sendRenderData();
        }
    });
};

Server.prototype._runTests = function(req, res) {
    var opt = req.param("opt"), source = opt.source, workers = opt.workers || [], runTestErrs = [], returnUrl = opt.location || "", browsers = opt.browsers || [], emulateCursor = opt.emulateCursor !== "false", sourceType = opt.sourceType, api = this.api, errMsgBuilder = this.errMsgBuilder, options = {
        sourceType: sourceType,
        source: source,
        workers: workers,
        browsers: browsers,
        emulateCursor: emulateCursor
    }, onTaskStarted = null;
    if (Array.isArray(source)) {
        source.forEach(function(elm) {
            source[elm] = decodeURI(source[elm]);
        });
    } else {
        source = decodeURI(source);
    }
    if (!workers.length && !browsers.length) {
        var workerName = this.testRunner.workerPool.createWorkerFromControlPanelWindow(returnUrl, req.headers["user-agent"]);
        options.workers.push(workerName);
        onTaskStarted = function(errs) {
            res.send({
                workerName: workerName,
                errs: errs ? errMsgBuilder.build(runTestErrs.concat(errs)) : []
            });
        };
    } else {
        onTaskStarted = function(errs, taskUid, workers) {
            var workerList = workers || [], status = "success";
            if (errs) {
                runTestErrs = runTestErrs.concat(errMsgBuilder.build(errs));
            }
            if (runTestErrs.length || !workerList.length) {
                status = workerList.length ? "with-errors" : "fail";
            }
            res.status(200);
            if (status !== "fail") {
                res.send({
                    status: status,
                    errs: runTestErrs,
                    taskUid: taskUid
                });
            } else {
                res.send({
                    status: status,
                    errs: runTestErrs
                });
            }
        };
    }
    api.runTests(options, onTaskStarted);
};

Server.prototype._addWorker = function(req, res) {
    var noRedirect = req.param("noRedirect"), autoCreated = req.param("auto_created") || "", createdByQRCode = req.param("create_by_qr_code") || "", name = createdByQRCode ? this.testRunner.workerPool.getWorkerNameFromUserAgent(req.headers["user-agent"]) : req.params.name, location = null, validName = "", err = "";
    try {
        if (!autoCreated) {
            validName = this.testRunner.workerPool.connectWorker(name, req.headers["user-agent"]);
        } else {
            validName = this.testRunner.workerPool.createWorkerInNewBrowserWindow(name, req.headers["user-agent"]);
        }
        location = [ "/worker/idle/", encodeURIComponent(validName) ].join("");
    } catch (e) {
        validName = null;
        err = this.errMsgBuilder.build(e);
    }
    if (err) {
        res.status(403);
        res.send(err);
        return;
    }
    if (noRedirect) {
        res.send(location);
    } else {
        res.status(302);
        res.set("location", location);
        res.end();
    }
};

Server.prototype._workerIdleHandler = function(req, res) {
    var restartTestRun = req.param("restartTestRun"), name = req.params.name;
    if (this.testRunner.workerPool.get(name)) {
        if (restartTestRun) {
            this.testRunner.restartCurrentTestRun(name);
        }
        var workerIsConnected = this.testRunner.workerPool.get(name).connected, workerType = workerIsConnected ? Const.CONNECTED_WORKER : Const.AUTO_CREATED;
        res.render("worker_idle", {
            workerName: name,
            workerType: workerType
        });
    } else {
        res.status(404);
        res.render("404");
    }
};

Server.prototype._workerHeartbeatHandler = function(req, res) {
    var state = this.testRunner.getWorkerStatus(req.param("name"));
    if (state) {
        res.send(state);
    } else {
        res.status(204);
        res.end();
    }
};

Server.prototype._listWorkers = function(req, res) {
    var render = req.param("render"), browsers = this.api.config.browsers || {}, workerNames = this.api.listConnectedWorkers();
    this.config.validateBrowsers(browsers, function(browsers) {
        if (render) {
            res.render("partial/worker_list", {
                workerNames: workerNames,
                browsers: browsers
            });
        } else {
            res.send(workerNames);
        }
    });
};

Server.prototype._getDetailTaskReport = function(req, res) {
    var taskID = req.params.taskID, partial = req.param("partial"), taskReport = this.testRunner.reporter.getTaskReportByUid(taskID);
    if (taskReport) {
        if (partial) {
            res.render("partial/task_detail", {
                r: taskReport
            });
        } else {
            res.render("task", {
                r: taskReport,
                remainingTime: this.remainingTime
            });
        }
    } else {
        res.status(404);
        res.render("404");
    }
};

Server.prototype._listTaskReports = function(req, res) {
    var partial = req.param("partial");
    if (!partial) {
        res.render("task_list", {
            taskReports: this.testRunner.reporter.getTaskReports(),
            remainingTime: this.remainingTime
        });
    } else {
        res.render("partial/tasks_report_container", {
            taskReports: this.testRunner.reporter.getTaskReports()
        });
    }
};

Server.prototype._getTaskReportRevision = function(req, res) {
    var taskID = req.params.uid, isDetail = req.param("detail"), taskReport = this.testRunner.reporter.getTaskReportByUid(taskID);
    if (taskReport) {
        if (isDetail) {
            res.render("partial/task_detail", {
                r: taskReport
            });
        } else {
            res.render("partial/task_list_item", {
                r: taskReport
            });
        }
    } else {
        res.status(404);
        res.render("404");
    }
};

Server.prototype._renameDir = function(req, res) {
    var urlPath = req.param("curPath") || "", vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], newName = req.param("newName"), errMsgBuilder = this.errMsgBuilder;
    this.vfs.renameDir(vfsPath, newName, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

Server.prototype._deleteDir = function(req, res) {
    var urlPath = req.param("curPath") || "", vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], errMsgBuilder = this.errMsgBuilder;
    this.vfs.deleteDir(vfsPath, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

Server.prototype._createFixture = function(req, res) {
    var urlPath = req.param("curPath") || "", fixtureName = req.param("fixtureName"), fixturePage = req.param("fixturePage"), filename = req.param("filename"), fixtureUsername = req.param("fixtureUsername"), fixturePassword = req.param("fixturePassword"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], errMsgBuilder = this.errMsgBuilder;
    this.vfs.createFixture(vfsPath, fixtureName, filename, fixturePage, fixtureUsername, fixturePassword, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

Server.prototype._createFixtureInDir = function(req, res) {
    var urlPath = req.param("curPath") || "", fixtureName = req.param("fixtureName"), fixturePage = req.param("fixturePage"), fixtureUsername = req.param("fixtureUsername"), fixturePassword = req.param("fixturePassword"), filename = req.param("filename"), dirName = req.param("dirName"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], errMsgBuilder = this.errMsgBuilder;
    this.vfs.createFixtureInDir(vfsPath, dirName, fixtureName, filename, fixturePage, fixtureUsername, fixturePassword, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

Server.prototype._editFixture = function(req, res) {
    var urlPath = req.param("curPath") || "", newName = req.param("newName"), newPage = req.param("newPage"), oldFilename = req.param("oldFilename"), newFilename = req.param("newFilename"), newUsername = req.param("newUsername"), newPassword = req.param("newPassword"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], errMsgBuilder = this.errMsgBuilder;
    this.vfs.editFixture(vfsPath, oldFilename, newFilename, newName, newPage, newUsername, newPassword, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

Server.prototype._deleteFixture = function(req, res) {
    var urlPath = req.param("curPath") || "", fixtureFilename = req.param("fixtureFilename"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], errMsgBuilder = this.errMsgBuilder;
    this.vfs.deleteFixture(vfsPath, fixtureFilename, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

Server.prototype._deleteTest = function(req, res) {
    var urlPath = req.param("curPath") || "", filename = req.param("filename"), testName = req.param("testName"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], errMsgBuilder = this.errMsgBuilder;
    this.vfs.deleteTest(vfsPath, filename, testName, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

Server.prototype._renameTest = function(req, res) {
    var urlPath = req.param("curPath") || "", filename = req.param("filename"), oldName = req.param("oldName"), newName = req.param("newName"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], errMsgBuilder = this.errMsgBuilder;
    this.vfs.renameTest(vfsPath, filename, oldName, newName, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

Server.prototype._record = function(req, res) {
    var returnUrl = req.param("returnUrl"), fixtureUid = req.param("fixtureUid"), errMsgBuilder = this.errMsgBuilder;
    this.recorder.startRecording(fixtureUid, returnUrl, function(err, recordingUrl) {
        if (err) {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        } else {
            res.send(recordingUrl);
        }
    });
};

Server.prototype._getDirsByPath = function(req, res) {
    var fsPath = req.param("path"), errMsgBuilder = this.errMsgBuilder;
    if (fsPath) {
        this.vfs.getFileStructureByPath(decodeURIComponent(fsPath), false, function(err, dirs) {
            res.status(200);
            res.send(dirs);
        });
    } else {
        this.vfs.getLocalDrives(function(err, devicesList) {
            if (!err) {
                res.status(200);
                res.send(devicesList);
            } else {
                res.status(304);
                res.send(errMsgBuilder.build(err));
            }
        });
    }
};

Server.prototype._getFilesByPath = function(req, res) {
    var fsPath = decodeURIComponent(req.param("path")), errMsgBuilder = this.errMsgBuilder;
    if (fsPath) {
        this.vfs.getFileStructureByPath(fsPath, true, function(err, fs) {
            res.status(200);
            res.send(fs);
        });
    } else {
        this.vfs.getLocalDrives(function(err, devicesList) {
            if (!err) {
                res.status(200);
                res.send(devicesList);
            } else {
                res.status(304);
                res.send(errMsgBuilder.build(err));
            }
        });
    }
};

Server.prototype._setTestsDir = function(req, res) {
    var fsPath = decodeURIComponent(req.param("testsDir")), controlPanel = this, errMsgBuilder = this.errMsgBuilder;
    controlPanel.vfs.changeTestsDir(fsPath, function(err) {
        if (!err) {
            controlPanel.config.setTestsDir(fsPath, function(err) {
                if (!err) {
                    res.status(200);
                    res.end();
                } else {
                    res.status(403);
                    res.send(errMsgBuilder.build(err));
                }
            });
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

Server.prototype._getOptions = function(req, res) {
    var controlPanel = this, browsers = controlPanel.config.browsers || {}, isControlPanelPortFree = true, isServicePortFree = true, isHostnameAutoDetect = true, isControlPanelPortAutoDetect = true, isServicePortAutoDetect = true;
    controlPanel.config.readCfgFile(function(cfg, err) {
        if (!err && cfg) {
            if (cfg.controlPanelPort) {
                if (cfg.controlPanelPort !== controlPanel.port) {
                    isControlPanelPortFree = false;
                    isControlPanelPortAutoDetect = true;
                } else {
                    isControlPanelPortAutoDetect = false;
                }
            }
            if (cfg.servicePort) {
                if (cfg.servicePort !== controlPanel.config.servicePort) {
                    isServicePortFree = false;
                    isServicePortAutoDetect = true;
                } else {
                    isServicePortAutoDetect = false;
                }
            }
            if (cfg.hostname) {
                isHostnameAutoDetect = false;
            }
        }
        controlPanel.config.validateBrowsers(browsers, function(browsers) {
            res.status(200);
            res.render("options", {
                hostname: {
                    value: controlPanel.config.hostname,
                    autoDetect: isHostnameAutoDetect
                },
                controlPanelPort: {
                    value: controlPanel.port,
                    autoDetect: isControlPanelPortAutoDetect,
                    portFree: isControlPanelPortFree
                },
                proxyPort: {
                    value: controlPanel.config.servicePort,
                    autoDetect: isServicePortAutoDetect,
                    portFree: isServicePortFree
                },
                browsers: browsers,
                remainingTime: controlPanel.remainingTime
            });
        });
    });
};

Server.prototype._configureBrowsers = function(req, res) {
    var browserName = req.param("browserName"), oldName = req.param("oldName") || "", browserInfo = req.param("browserInfo") || null, config = this.config, errMsgBuilder = this.errMsgBuilder;
    var saveBrowsersAndSendResponse = function() {
        config.updateConfig({
            browsers: config.browsers
        }, function(err) {
            if (!err) {
                res.status(200);
                res.render("partial/browsers", {
                    browsers: config.browsers
                });
            } else {
                res.status(403);
                res.send({
                    saveError: errMsgBuilder.build(err)
                });
            }
        });
    };
    if (browserInfo) {
        var browsers = {};
        browsers[browserName] = browserInfo;
        if (oldName && oldName !== browserName) {
            delete config.browsers[oldName];
        }
        config.validateBrowsers(browsers, function(validatedBrowsers) {
            if (validatedBrowsers[browserName].pathExists) {
                config.browsers[browserName] = browserInfo;
                saveBrowsersAndSendResponse();
            } else {
                res.status(403);
                res.send({
                    pathError: true
                });
            }
        });
    } else {
        delete config.browsers[browserName];
        saveBrowsersAndSendResponse();
    }
};

Server.prototype._configureNetwork = function(req, res) {
    var controlPanel = this, options = req.param("options"), servicePort = options.service_port, controlPanelPort = options.control_panel_port, hostname = options.hostname, changingOptions = {}, warnings = {}, errMsgBuilder = this.errMsgBuilder;
    async.series({
        "Change proxy port": function(done) {
            if (servicePort === undefined) {
                done();
            } else {
                if (servicePort !== "") {
                    changingOptions.servicePort = parseInt(servicePort);
                    controlPanel.config.isPortFree(servicePort, function(isFree) {
                        if (!isFree) {
                            warnings.servicePort = true;
                        }
                        done();
                    });
                } else {
                    changingOptions.servicePort = servicePort;
                    done();
                }
            }
        },
        "Change control panel port": function(done) {
            if (controlPanelPort === undefined) {
                done();
            } else {
                if (controlPanelPort !== "") {
                    changingOptions.controlPanelPort = parseInt(controlPanelPort);
                    controlPanel.config.isPortFree(controlPanelPort, function(isFree) {
                        if (!isFree) {
                            warnings.controlPanelPort = true;
                        }
                        done();
                    });
                } else {
                    changingOptions.controlPanelPort = controlPanelPort;
                    done();
                }
            }
        },
        "Change hostname": function(done) {
            changingOptions.hostname = hostname;
            done();
        }
    }, function() {
        controlPanel.config.updateConfig(changingOptions, function(err) {
            if (!err) {
                res.status(200);
                res.send(warnings);
            } else {
                res.status(403);
                res.send(errMsgBuilder.build(err));
            }
        });
    });
};

Server.prototype._checkUpdates = function(req, res) {
    res.status(200);
    res.send(this.version);
};

Server.prototype._detectBrowsers = function(req, res) {
    var controlPanel = this;
    controlPanel.config.addInstalledBrowsers(function(browsers) {
        res.status(200);
        if (browsers && Object.keys(browsers).length) {
            res.render("partial/browsers", {
                browsers: controlPanel.config.browsers
            });
        } else {
            res.end();
        }
    });
};

Server.prototype._getFixtureCode = function(req, res) {
    var fixtureFile = req.param("fixtureFile"), urlPath = req.param("curPath"), vfsPath = urlPath ? decodeURI(urlPath) : "", fixturePath = path.join(vfsPath, fixtureFile);
    this.vfs.getFixtureCode(fixturePath, function(err, code) {
        if (err) {
            res.status(403);
            res.send(err);
        } else {
            res.writeHead(200, {
                "Content-Type": "text/plain"
            });
            res.end(code);
        }
    });
};

Server.prototype._saveFixtureCode = function(req, res) {
    var filename = req.param("filename"), urlPath = req.param("curPath"), code = req.param("code"), vfsPath = urlPath ? decodeURI(urlPath) : "", fixturePath = path.join(vfsPath, filename), errMsgBuilder = this.errMsgBuilder;
    this.vfs.saveFixtureCode(fixturePath, code, function(err) {
        if (err) {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        } else {
            res.status(200);
            res.end();
        }
    });
};

Server.prototype._page404Handler = function(req, res) {
    res.status(404);
    res.render("404");
};

Server.prototype._setupSockets = function() {
    var socketIO = this.socketIO, reporter = this.testRunner.reporter, workerPool = this.testRunner.workerPool;
    this.testRunner.on("taskAdded", function(workerNames) {
        workerNames.forEach(function(workerName) {
            socketIO.sockets.in(workerName).emit("taskAdded");
        });
        socketIO.sockets.in("registerSeveralWorkers").emit("taskAdded");
    });
    reporter.on("taskUpdated", function(data) {
        socketIO.sockets.emit("taskUpdated", {
            id: data.id
        });
    });
    reporter.on("taskDetailErrUpdated", function(data) {
        socketIO.sockets.emit("taskDetailErrUpdated", {
            id: data.id
        });
    });
    reporter.on("taskListTestFailed", function(data) {
        socketIO.sockets.emit("taskListTestFailed", {
            id: data.id
        });
    });
    this.vfs.on("suiteChanged", function() {
        socketIO.sockets.emit("vfsRevisionChanged");
    });
    workerPool.on("workerAdded", function() {
        socketIO.sockets.emit("workerAdded");
    });
    workerPool.on("workerDisconnected", function() {
        socketIO.sockets.emit("workerDisconnected");
    });
    socketIO.sockets.on("connection", function(socket) {
        socket.on("registerWorker", function(workerName) {
            if (workerName && !socket.disconnected) {
                socket.join(workerName);
            }
        });
        socket.on("registerSeveralWorkers", function() {
            if (!socket.disconnected) {
                socket.join("workerList");
            }
        });
    });
};