"@fixture A set of examples that illustrate how to use TestCafe API";

"@page http://testcafe.devexpress.com/Example";

'@test'['How to implement typing (act.type user action)'] = {
    'Type name': function () {
        act.type($('#Developer_Name'), 'Peter');
    },
    'Replace with last name': function () {
        act.type($('#Developer_Name'), 'Paker', {
            replace: true
        });
    },
    'Update last name': function () {
        act.type($('#Developer_Name'), 'r', {
            caretPos: 2
        });
    },
    'Check result': function () {
        eq($('#Developer_Name').val(), 'Parker');
    }
};

'@test'['How to click an array of labels and then check their states (act.click user action and ok() assertion)'] = {
    'Click two checkbox labels': function () {
        var labels = [$(':containsExcludeChildren(Support for testing on remote devices)'),
            $(':containsExcludeChildren(Reusing existing JavaScript code for testing)')];
        act.click(labels);
    },
    'Confirm checked state': function () {
        ok($('#testing-on-remote-devices').is(':checked'));
        ok($('#re-using-existing-javascript').is(':checked'));
    }
};

'@test'['How to implement typing, clicking at the specified position and then backspacing the symbol (act.type, act.click, act.press user actions and eq() assertion)'] = {
    'Type name': function () {
        act.type(getInput(), 'Peter Parker');
    },
    'Move caret position': function () {
        act.click(getInput(), {
            caretPos: 5
        });
    },
    'Erase a character': function () {
        act.press('backspace');
    },
    'Check result': function () {
        eq(getInput().val(), 'Pete Parker');
    }
};

'@test'['How to implement pressing specified button (act.press user action)'] = {
    'Type name': function () {
        act.type(getInput(), 'Peter Parker');
    },
    'Erase Peter': function () {
        act.press('home right . delete delete delete delete');
    },
    'Check result': function () {
        eq(getInput().val(), 'P. Parker');
    }
};

'@test'['How to implement a comparison (ok() and notOk() assertions)'] = {
    'Check comments area is disabled': function () {
        ok($('#Developer_Comments').is(':disabled'));
    },
    'Click label "I have tried TestCafe"': function () {
        var label = $(':containsExcludeChildren(I have tried TestCafe)');
        act.click(label);
    },
    'Check comments area is enabled': function () {
        notOk($('#Developer_Comments').is(':disabled'));
    }
};

'@test'['How to compare the input value with the specified one (eq() assertion)'] = {
    'Type name': function () {
        act.type(getInput(), 'Peter Parker');
    },
    'Submit form': function () {
        var submitButton = $('.button.blue.fix-width-180');
        act.click(submitButton);
    },
    'Check message': function () {
        var header = $('.article-header');
        eq(header.html(), 'Thank You, Peter Parker!');
    }
};

'@test'['How to implement dragging (act.drag user action)'] = {
    'Click label "I have tried TestCafe"': function () {
        var label = $(':containsExcludeChildren(I have tried TestCafe)');
        act.click(label);
    },
    'Check initial slider value': function () {
        eq($('#Developer_Rating').val(), 1);
    },
    'Drag slider handle': function () {
        var sliderHandle = $('.ui-slider-handle');
        act.drag(sliderHandle, 357, 0, {
            offsetX: 10,
            offsetY: 10
        });
    },
    'Check resulting slider value': function () {
        eq($('#Developer_Rating').val(), 7);
    }
};

'@test'['How to implement hovering (act.hover user action)'] = {
    'Hover over the combo box': function () {
        var div = $('.text-field');
        act.hover(div);
    },
    'Select "Both"': function () {
        var div = $(':containsExcludeChildren(Both)').eq(0);
        act.click(div);
    },
    'Check result': function () {
        var value = $('.text-field').html();
        eq(value, 'Both');
    }
};

'@test'['How to implement waiting (act.wait user action)'] = {
    'Initiate animation and wait': function () {
        $('.article-header').animate({
            opacity: 0
        }, 1e3);
        act.wait(1e4, isTransparent);
    },
    'Type': function () {
        var input = $('#Developer_Name');
        act.type(input, 'The wait is over!');
    }
};

'@test'['How to implement selection (act.select user action) '] = {
    '1.Type within “Your name:" input element': function () {
        act.type(getInput(), 'Test Cafe', {
            caretPos: 0
        });
    },
    '2.Select within "Your name:" input element': function () {
        act.select(getInput(), 7, 1);
    },
    '3.Confirm input selection state': function () {
        eq(getInput()[0].selectionStart, 1);
        eq(getInput()[0].selectionEnd, 7);
        if (getInput()[0].selectionDirection)
            eq(getInput()[0].selectionDirection, 'backward');
    }
};

'@test'['How to handle a confirmation dialog (handleConfirm() dialog handling function)'] = {
    '1.Click "Populate"': function () {
        var populateFormButton = $(':containsExcludeChildren(Populate)');
        handleConfirm('OK');
        act.click(populateFormButton);
    },
    '2.Click "Submit"': function () {
        this.autoGeneratedName = $('#Developer_Name').val();
        var submitButton = $('.button.blue.fix-width-180');
        act.click(submitButton);
    },
    'Check result': function () {
        var header = $('.article-header');
        var expectedResult = 'Thank You, ' + this.autoGeneratedName + '!';
        eq(header.html(), expectedResult);
    }
};

'@test'['A simple test on our example page (act.click, act.type, act.hover user actions and eq() assertion)'] = {
    '1.Click "MacOS"': function () {
        var label = $(':containsExcludeChildren(MacOS)');
        act.click(label);
    },
    '2.Type a name': function () {
        var input = $('#Developer_Name');
        act.type(input, 'Peter Parker', {
            offsetX: 59,
            offsetY: 22,
            caretPos: 0
        });
    },
    '3.Hover over "Visual recorder"': function () {
        var div = $('.text-field');
        act.hover(div);
    },
    '4.Click "Both"': function () {
        var div = $(':containsExcludeChildren(Both)').eq(0);
        act.click(div, {
            offsetX: 71,
            offsetY: 12
        });
    },
    '5.Click "Submit"': function () {
        var submitButton = $('#submit-button');
        act.click(submitButton, {
            offsetX: 126,
            offsetY: 38
        });
    },
    '6. Check Result': function () {
        var header = $('.article-header');
        eq(header.html(), 'Thank You, Peter Parker!');
    }

};

//helpers
var isTransparent = function () {
    return $('.article-header').css('opacity') == 0;
};

var getInput = function () {
    return $('#Developer_Name');
};
