var dns = require("dns"), tslErrorsDomain = require("domain").create(), http = require("http"), https = require("https"), os = require("os"), url = require("url"), util = require("util"), webauth = require("webauth"), Events = require("./events"), ERR = require("./server_errs"), exec = require("child_process").exec, injector = require("./injector"), sharedConst = require("./../shared/const"), urlUtil = require("./../shared/url_util");

process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";

var HTTPS_AGENT_CACHE_LIMIT = 1e3;

var httpAgent = new http.Agent(), httpsAgent = new https.Agent(), httpsAgentFallback = new https.Agent(), httpsAgentsCache = {};

httpAgent.maxSockets = httpsAgent.maxSockets = httpsAgentFallback.maxSockets = 128;

httpsAgent.options.secureProtocol = "SSLv3_method";

httpsAgentFallback.options.secureProtocol = "TLSv1_method";

var curUserDomain = "", curUserWorkstation = "", isWin = !!os.platform().match(/^win/);

if (isWin) {
    exec("echo %userdomain%", function(error, domain) {
        if (!error) {
            curUserDomain = domain.replace(/\s/g, "");
        }
    });
    exec("hostname", function(error, hostname) {
        if (!error) {
            curUserWorkstation = hostname.replace(/\s/g, "");
        }
    });
}

tslErrorsDomain.on("error", function(err) {
    if (err.code !== "ECONNRESET") {
        tslErrorsDomain.removeAllListeners("error");
        throw new Error(err);
    }
});

function getContentLength(bodyChunks) {
    var length = 0;
    bodyChunks.forEach(function(chunk) {
        length += chunk.length;
    });
    return length;
}

function testForPageMime(header) {
    return header && /(text\/html)|(application\/xhtml\+xml)|(application\/xml)|(application\/x-ms-application)/gi.test(header);
}

function getCharset(contentTypeHeader) {
    if (contentTypeHeader) {
        var charsetMatch = contentTypeHeader.match(/(?:^|;)\s*charset=(.+)(?:;|$)/i);
        if (charsetMatch) {
            return charsetMatch[1];
        }
    }
    return "iso-8859-1";
}

function isSSLProtocolErr(err) {
    return err.message && err.message.indexOf("SSL routines") > -1;
}

function isCSSResource(contentType, accept) {
    return contentType.indexOf("text/css") !== -1 || accept === "text/css";
}

function isScriptResource(contentType, accept) {
    return contentType.indexOf("application/javascript") !== -1 || contentType.indexOf("text/javascript") !== -1 || contentType.indexOf("application/x-javascript") !== -1 || accept === "application/javascript" || accept === "text/javascript" || accept === "application/x-javascript";
}

function isManifest(contentType) {
    return contentType.indexOf("text/cache-manifest") !== -1;
}

function isJSON(contentType) {
    return contentType.indexOf("application/json") !== -1;
}

var Server = module.exports = function(port, hostname, cookieShelf, serviceChannel) {
    this.port = port;
    this.hostname = hostname;
    this.events = new Events();
    this.cookieShelf = cookieShelf;
    this.serviceChannel = serviceChannel;
    var server = this;
    this.server = http.createServer(function(req, res) {
        var parsedProxyUrl = urlUtil.parseProxyUrl(req.url);
        var isXhr = !!req.headers[sharedConst.XHR_REQUEST_MARKER_HEADER], isIFrame = parsedProxyUrl && parsedProxyUrl.isIFrame, ctx = {
            jobInfo: null,
            req: req,
            reqBodyChunks: [],
            res: res,
            resHeaders: {},
            originResourceInfo: null,
            originResContentInfo: null,
            originReqOptions: {},
            originReq: null,
            originRes: null,
            originResBodyChunks: [],
            isXhr: isXhr,
            isIFrame: isIFrame,
            isPageReqCandidate: !isXhr && testForPageMime(req.headers["accept"]),
            originReqErr: false
        };
        server._onProxyRequest(ctx);
    });
    this.server.on("error", function(err) {
        if (err.code === "EADDRINUSE") {
            server.events.broadcast.emit("fatalError", {
                code: ERR.PROXY_PORT_IS_ALREADY_IN_USE,
                port: server.port
            });
        }
    });
};

Server.prototype.start = function() {
    this.server.listen(this.port);
};

Server.prototype.close = function() {
    this.server.close();
};

Server.prototype.getProxyUrl = function(originUrl, jobUid, jobOwnerToken, isIFrame) {
    return urlUtil.getProxyUrl(originUrl, this.hostname, this.port, jobUid, jobOwnerToken, isIFrame);
};

Server.prototype.injectInPage = function(ctx, callback) {
    var injectionOptions = {
        urlReplacer: this.getResourceUrlReplacer(ctx),
        styleUrl: this.serviceChannel.staticCssUrl,
        scripts: [ this.serviceChannel.hammerheadScriptUrl, this.serviceChannel.staticScriptUrl, this.serviceChannel.taskScriptUrl ]
    }, chunks = ctx.originResBodyChunks, encoding = ctx.originResContentInfo.encoding, charset = ctx.originResContentInfo.charset;
    injector.injectInPage(chunks, encoding, charset, injectionOptions, callback);
};

Server.prototype.getResourceUrlReplacer = function(ctx) {
    var server = this;
    return function(resourceUrl, isIFrame) {
        var baseUrl = urlUtil.getBaseUrl();
        baseUrl = baseUrl ? url.resolve(ctx.originResourceInfo.url, baseUrl) : "";
        var resolvedUrl = url.resolve(baseUrl || ctx.originResourceInfo.url, resourceUrl);
        try {
            return server.getProxyUrl(resolvedUrl, ctx.jobInfo.uid, ctx.jobInfo.ownerToken, isIFrame);
        } catch (err) {
            return resourceUrl;
        }
    };
};

Server.prototype._error = function(ctx, err) {
    var eventsInCtx = this.events.in(ctx);
    if (eventsInCtx.canHandle("error")) {
        eventsInCtx.emit("error", err);
    } else {
        ctx.res.statusCode = 500;
        ctx.res.end();
    }
};

Server.prototype._extractReqInfo = function(ctx) {
    if (ctx.req.url === "/favicon.ico") {
        ctx.res.statusCode = 404;
        ctx.res.end();
        return;
    }
    var parsingResult = urlUtil.parseProxyUrl(ctx.req.url), referer = ctx.req.headers["referer"], refererParsingResult = null;
    try {
        refererParsingResult = referer && urlUtil.parseProxyUrl(referer);
    } catch (err) {
        refererParsingResult = null;
    }
    if (refererParsingResult) {
        if (refererParsingResult.originResourceInfo.protocol === "https:" && refererParsingResult.originResourceInfo.port === "443" || refererParsingResult.originResourceInfo.protocol === "http:" && refererParsingResult.originResourceInfo.port === "80") {
            refererParsingResult.originResourceInfo.host = refererParsingResult.originResourceInfo.host.split(":")[0];
            refererParsingResult.originResourceInfo.port = "";
        }
    }
    if (!parsingResult) {
        if (!refererParsingResult) {
            return null;
        }
        var parsedReqUrl = url.parse(ctx.req.url, true), hostParts = parsedReqUrl.host && parsedReqUrl.host.split(":");
        parsingResult = {
            originResourceInfo: {
                protocol: parsedReqUrl.protocol || refererParsingResult.originResourceInfo.protocol,
                host: parsedReqUrl.host || refererParsingResult.originResourceInfo.host,
                hostname: hostParts ? hostParts[0] : refererParsingResult.originResourceInfo.hostname,
                port: hostParts ? hostParts[1] : refererParsingResult.originResourceInfo.port,
                pathname: parsedReqUrl.pathname,
                query: parsedReqUrl.query,
                hash: parsedReqUrl.hash
            },
            jobInfo: refererParsingResult.jobInfo
        };
    }
    parsingResult.originResourceInfo.url = urlUtil.formatUrl(parsingResult.originResourceInfo);
    parsingResult.originResourceInfo.domain = urlUtil.formatUrl({
        protocol: parsingResult.originResourceInfo.protocol,
        host: parsingResult.originResourceInfo.host,
        hostname: parsingResult.originResourceInfo.hostname,
        port: parsingResult.originResourceInfo.port
    });
    if (refererParsingResult) {
        var refererResourceInfo = refererParsingResult.originResourceInfo;
        parsingResult.originResourceInfo.referer = urlUtil.formatUrl(refererResourceInfo);
        parsingResult.originResourceInfo.reqOrigin = urlUtil.formatUrl({
            protocol: refererResourceInfo.protocol,
            host: refererResourceInfo.host,
            hostname: refererResourceInfo.hostname,
            port: refererResourceInfo.port
        });
    }
    return parsingResult;
};

Server.prototype._onProxyRequest = function(ctx) {
    ctx.req.on("data", function(chunk) {
        ctx.reqBodyChunks.push(chunk);
    });
    var server = this;
    ctx.req.on("end", function() {
        if (server.serviceChannel.shouldProcess(ctx)) {
            server.serviceChannel.process(ctx);
        } else {
            server._lookupOrigin(ctx);
        }
    });
};

Server.prototype._lookupOrigin = function(ctx) {
    var requestInfo = this._extractReqInfo(ctx);
    if (!requestInfo) {
        ctx.res.statusCode = 404;
        ctx.res.end();
        return;
    }
    ctx.originResourceInfo = requestInfo.originResourceInfo;
    ctx.jobInfo = requestInfo.jobInfo;
    if (ctx.originResourceInfo.hostname === "localhost" || ctx.originResourceInfo.hostname === "127.0.0.1") {
        ctx.originReqOptions.hostname = "127.0.0.1";
        this._prepareOriginRequest(ctx);
    } else {
        var server = this;
        dns.lookup(ctx.originResourceInfo.hostname, function(err, addr) {
            if (err || !addr) {
                server._error(ctx, {
                    code: ERR.PROXY_CANT_RESOLVE_ORIGIN_URL,
                    originUrl: ctx.originResourceInfo.url
                });
                return;
            }
            ctx.originReqOptions.hostname = addr;
            server._prepareOriginRequest(ctx);
        });
    }
};

Server.prototype._prepareOriginRequest = function(ctx) {
    ctx.originReqOptions.port = ctx.originResourceInfo.port;
    ctx.originReqOptions.method = ctx.req.method;
    ctx.originReqOptions.path = urlUtil.formatUrl({
        pathname: ctx.originResourceInfo.pathname,
        query: ctx.originResourceInfo.query
    }, ctx.req.url);
    ctx.originReqOptions.hash = ctx.originResourceInfo.hash;
    ctx.originReqOptions.headers = {};
    if (ctx.originResourceInfo.protocol === "http:") {
        ctx.originReqOptions.agent = httpAgent;
    } else {
        if (!httpsAgentsCache[ctx.originReqOptions.hostname]) {
            ctx.originReqOptions.agent = httpsAgent;
        } else {
            ctx.originReqOptions.agent = httpsAgentFallback;
        }
    }
    var server = this;
    Object.keys(ctx.req.headers).forEach(function(field) {
        switch (field) {
          case "host":
            ctx.originReqOptions.headers[field] = ctx.originResourceInfo.host;
            break;

          case "referer":
            if (ctx.originResourceInfo.referer) {
                ctx.originReqOptions.headers[field] = ctx.originResourceInfo.referer;
            }
            break;

          case "origin":
            ctx.originReqOptions.headers[field] = ctx.originResourceInfo.reqOrigin || ctx.req.headers["origin"];
            break;

          case "cookie":
            break;

          case sharedConst.XHR_REQUEST_MARKER_HEADER:
            break;

          default:
            ctx.originReqOptions.headers[field] = ctx.req.headers[field];
        }
    });
    if (ctx.isXhr && !ctx.req.headers["origin"] && ctx.originResourceInfo.domain !== ctx.originResourceInfo.reqOrigin) {
        ctx.originReqOptions.headers["origin"] = ctx.originResourceInfo.reqOrigin;
    }
    this.cookieShelf.getCookieHeader(ctx.jobInfo, ctx.originResourceInfo.url, function(cookieHeader) {
        ctx.originReqOptions.headers["cookie"] = cookieHeader;
        server._sendOriginRequest(ctx);
    });
};

Server.prototype._sendOriginRequest = function(ctx) {
    var server = this, isHttps = ctx.originResourceInfo.protocol === "https:", protocolInterface = isHttps ? https : http;
    if (isHttps) {
        ctx.originReqOptions.rejectUnauthorized = false;
    }
    var sendRequest = function() {
        tslErrorsDomain.run(function() {
            ctx.originReq = protocolInterface.request(ctx.originReqOptions, function(originRes) {
                ctx.originRes = originRes;
                if (originRes.statusCode === 401) {
                    server.events.in(ctx).emit("authCredentialsRequested", function(credentials) {
                        if (credentials) {
                            ctx.originReq.abort();
                            webauth.auth(ctx.originReqOptions, credentials, function(targetRes) {
                                ctx.originRes = targetRes;
                                server._onOriginResponse(ctx);
                            }, isHttps, originRes);
                        } else {
                            server._onOriginResponse(ctx);
                        }
                    });
                } else {
                    server._onOriginResponse(ctx);
                }
            });
        });
        ctx.originReq.on("error", function(err) {
            if (isSSLProtocolErr(err) && ctx.originReqOptions.agent.options.secureProtocol === httpsAgent.options.secureProtocol) {
                if (Object.keys(httpsAgentsCache).length > HTTPS_AGENT_CACHE_LIMIT) {
                    httpsAgentsCache = {};
                }
                httpsAgentsCache[ctx.originReqOptions.hostname] = true;
                ctx.originReqOptions.agent = httpsAgentFallback;
                sendRequest();
            } else {
                ctx.originReqErr = true;
            }
        });
        ctx.reqBodyChunks.forEach(function(chunk) {
            ctx.originReq.write(chunk);
        });
        ctx.originReq.end();
    };
    sendRequest();
};

Server.prototype._isSameOriginPolicyCompliantRequest = function(ctx) {
    var reqOrigin = ctx.originResourceInfo.reqOrigin;
    if (ctx.originResourceInfo.domain !== reqOrigin) {
        if (ctx.req.headers[sharedConst.XHR_REQUEST_MARKER_HEADER] === sharedConst.XHR_CORS_SUPPORTED_FLAG) {
            if (ctx.req.method === "OPTIONS") {
                return true;
            }
            var allowOriginHeader = ctx.originRes.headers["access-control-allow-origin"], credentials = !!ctx.originRes.headers["access-control-allow-credentials"];
            return util.isArray(allowOriginHeader) ? allowOriginHeader.indexOf("*") > -1 && !credentials || allowOriginHeader.indexOf(reqOrigin) > -1 : allowOriginHeader === "*" && !credentials || allowOriginHeader === reqOrigin;
        }
        return false;
    }
    return true;
};

Server.prototype._extractOriginResponseContentInfo = function(ctx) {
    var isPage = ctx.isPageReqCandidate, contentType = ctx.originRes.headers["content-type"] || "", accept = ctx.req.headers["accept"] || "";
    if (ctx.originRes.headers["content-type"]) {
        isPage = isPage && testForPageMime(contentType);
    }
    ctx.originResContentInfo = {
        encoding: ctx.originRes.headers["content-encoding"],
        charset: getCharset(contentType),
        isPage: isPage,
        isCSS: isCSSResource(contentType, accept),
        isScript: isScriptResource(contentType, accept),
        isManifest: isManifest(contentType),
        isJSON: isJSON(contentType)
    };
};

Server.prototype._onOriginResponse = function(ctx) {
    this._extractOriginResponseContentInfo(ctx);
    if (ctx.isXhr && !this._isSameOriginPolicyCompliantRequest(ctx)) {
        this._error(ctx, {
            code: ERR.PROXY_XHR_REQUEST_SAME_ORIGIN_POLICY_VIOLATION,
            reqOrigin: ctx.originResourceInfo.reqOrigin,
            xhrReqUrl: ctx.originResourceInfo.url
        });
        ctx.res.statusCode = 0;
        ctx.res.end();
        return;
    }
    var server = this;
    var eventsInCtx = server.events.in(ctx);
    if (eventsInCtx.canHandle("originResponse")) {
        eventsInCtx.emit("originResponse", function() {
            server._processOriginResponse(ctx);
        });
    } else {
        server._processOriginResponse(ctx);
    }
};

Server.prototype._processOriginResponse = function(ctx) {
    var server = this, contentInfo = ctx.originResContentInfo, shouldProcessContent = contentInfo.isPage || contentInfo.isCSS || contentInfo.isScript || contentInfo.isManifest || contentInfo.isJSON;
    if (shouldProcessContent) {
        ctx.originRes.on("data", function(chunk) {
            ctx.originResBodyChunks.push(chunk);
        });
        ctx.originRes.on("end", function() {
            server._prepareProcessedProxyResponse(ctx);
        });
    } else {
        server._sendProxyResponseHeaders(ctx, false, function() {
            ctx.originRes.pipe(ctx.res);
        });
    }
};

Server.prototype._sendProxyResponseHeaders = function(ctx, recalculateContentLength, callback) {
    var server = this, cookies = ctx.originRes.headers["set-cookie"] || "";
    server.cookieShelf.setCookieByServer(ctx.jobInfo, ctx.originResourceInfo.url, cookies, function() {
        Object.keys(ctx.originRes.headers).forEach(function(field) {
            switch (field) {
              case "set-cookie":
              case "content-security-policy":
              case "x-content-security-policy":
              case "x-webkit-csp":
              case "access-control-allow-origin":
                return;

              case "location":
                var location = ctx.originRes.headers[field];
                var parsedLocation = url.parse(location);
                if (!parsedLocation.host) {
                    location = url.resolve(ctx.originResourceInfo.url, location);
                }
                ctx.resHeaders[field] = server.getProxyUrl(location, ctx.jobInfo.uid, ctx.jobInfo.ownerToken);
                return;

              default:
                ctx.resHeaders[field] = ctx.originRes.headers[field];
            }
        });
        if (recalculateContentLength) {
            ctx.resHeaders["content-length"] = getContentLength(ctx.originResBodyChunks);
        }
        ctx.res.writeHeader(ctx.originRes.statusCode, ctx.resHeaders);
        ctx.res.addTrailers(ctx.originRes.trailers);
        callback();
    });
};

Server.prototype._hasOriginReqError = function(ctx) {
    if (!ctx.originReqErr) {
        return false;
    }
    var contentLength = ctx.originRes.headers["content-length"];
    return !ctx.originRes || Object.keys(ctx.originRes.headers).length === 0 || typeof contentLength !== "undefined" && contentLength !== getContentLength(ctx.originResBodyChunks);
};

Server.prototype._processResourceContent = function(ctx, callback) {
    if (ctx.originRes.statusCode !== 200) {
        callback();
        return;
    }
    var urlReplacer = this.getResourceUrlReplacer(ctx), contentInfo = ctx.originResContentInfo;
    if (contentInfo.isPage) {
        var eventsInCtx = this.events.in(ctx);
        this.injectInPage(ctx, function(err, html) {
            if (!err) {
                ctx.originResBodyChunks = [ html ];
            } else {
                if (eventsInCtx.canHandle("injectionError")) {
                    eventsInCtx.emit("injectionError", err);
                }
            }
            callback();
        });
    } else {
        if (contentInfo.isCSS) {
            injector.injectInStylesheet(ctx.originResBodyChunks, ctx.originResContentInfo.encoding, ctx.originResContentInfo.charset, urlReplacer, function(err, processedStylesheet) {
                if (!err) {
                    ctx.originResBodyChunks = [ processedStylesheet ];
                }
                callback();
            });
        } else {
            if (contentInfo.isScript && !ctx.isXhr) {
                injector.injectInScript(ctx.originResBodyChunks, ctx.originResContentInfo.encoding, ctx.originResContentInfo.charset, function(err, processedScript) {
                    if (!err) {
                        ctx.originResBodyChunks = [ processedScript ];
                    }
                    callback();
                });
            } else {
                if (contentInfo.isManifest) {
                    injector.injectInManifest(ctx.originResBodyChunks, ctx.originResContentInfo.encoding, ctx.originResContentInfo.charset, urlReplacer, function(err, processedManifest) {
                        if (!err) {
                            ctx.originResBodyChunks = [ processedManifest ];
                        }
                        callback();
                    });
                } else {
                    if (contentInfo.isJSON) {
                        injector.injectInJSON(ctx.originResBodyChunks, ctx.originResContentInfo.encoding, ctx.originResContentInfo.charset, function(err, processedJSON) {
                            if (!err) {
                                ctx.originResBodyChunks = [ processedJSON ];
                            }
                            callback();
                        });
                    } else {
                        callback();
                    }
                }
            }
        }
    }
};

Server.prototype._prepareProcessedProxyResponse = function(ctx) {
    if (this._hasOriginReqError(ctx)) {
        this._error(ctx, {
            code: ERR.PROXY_ORIGIN_SERVER_CONNECTION_TERMINATED,
            originUrl: ctx.originResourceInfo.url
        });
        return;
    }
    var server = this;
    this._processResourceContent(ctx, function() {
        server._sendProxyResponseHeaders(ctx, true, function() {
            server._sendProcessedProxyResponseBody(ctx);
        });
    });
};

Server.prototype._sendProcessedProxyResponseBody = function(ctx) {
    try {
        ctx.originResBodyChunks.forEach(function(chunk) {
            ctx.res.write(chunk);
        });
        ctx.res.end();
    } catch (err) {}
};