(function() {
    var JSProcessor = {};
    var isNode = typeof module !== "undefined" && module.exports;
    var JSParsingTools = isNode ? require("./js_parsing_tools") : HammerheadClient.get("Shared.JSParsingTools");
    JSProcessor.GET_LOCATION_METH_NAME = "__get$Loc";
    JSProcessor.SET_LOCATION_METH_NAME = "__set$Loc";
    JSProcessor.SET_PROPERTY_METH_NAME = "__set$";
    JSProcessor.GET_PROPERTY_METH_NAME = "__get$";
    JSProcessor.PROCESS_SCRIPT_METH_NAME = "__proc$Script";
    JSProcessor.MOCK_ACCESSORS = [ 'var __w$undef_ = typeof window === "undefined",\r\n', JSProcessor.GET_LOCATION_METH_NAME, "=__w$undef_?function(l){return l}:window.", JSProcessor.GET_LOCATION_METH_NAME, ",\r\n", JSProcessor.SET_LOCATION_METH_NAME, "=__w$undef_?function(l,v){return l = v}:window.", JSProcessor.SET_LOCATION_METH_NAME, ",\r\n", JSProcessor.SET_PROPERTY_METH_NAME, "=__w$undef_?function(o,p,v){return o[p] = v}:window.", JSProcessor.SET_PROPERTY_METH_NAME, ",\r\n", JSProcessor.GET_PROPERTY_METH_NAME, "=__w$undef_?function(o,p){return o[p]}:window.", JSProcessor.GET_PROPERTY_METH_NAME, ",\r\n", JSProcessor.PROCESS_SCRIPT_METH_NAME, "=__w$undef_?function(s){return s}:window.", JSProcessor.PROCESS_SCRIPT_METH_NAME, ";\r\n" ].join("");
    var SYNTAX = {
        AssignmentExpression: "AssignmentExpression",
        ArrayExpression: "ArrayExpression",
        BlockStatement: "BlockStatement",
        BinaryExpression: "BinaryExpression",
        BreakStatement: "BreakStatement",
        CallExpression: "CallExpression",
        CatchClause: "CatchClause",
        ConditionalExpression: "ConditionalExpression",
        ContinueStatement: "ContinueStatement",
        DoWhileStatement: "DoWhileStatement",
        DebuggerStatement: "DebuggerStatement",
        EmptyStatement: "EmptyStatement",
        ExpressionStatement: "ExpressionStatement",
        ForStatement: "ForStatement",
        ForInStatement: "ForInStatement",
        FunctionDeclaration: "FunctionDeclaration",
        FunctionExpression: "FunctionExpression",
        Identifier: "Identifier",
        IfStatement: "IfStatement",
        Literal: "Literal",
        LabeledStatement: "LabeledStatement",
        LogicalExpression: "LogicalExpression",
        MemberExpression: "MemberExpression",
        NewExpression: "NewExpression",
        ObjectExpression: "ObjectExpression",
        Program: "Program",
        Property: "Property",
        ReturnStatement: "ReturnStatement",
        SequenceExpression: "SequenceExpression",
        SwitchStatement: "SwitchStatement",
        SwitchCase: "SwitchCase",
        ThisExpression: "ThisExpression",
        ThrowStatement: "ThrowStatement",
        TryStatement: "TryStatement",
        UnaryExpression: "UnaryExpression",
        UpdateExpression: "UpdateExpression",
        VariableDeclaration: "VariableDeclaration",
        VariableDeclarator: "VariableDeclarator",
        WhileStatement: "WhileStatement",
        WithStatement: "WithStatement"
    };
    var codegenOpts = {
        comment: true,
        format: {
            compact: true,
            quotes: "double",
            escapeless: true
        }
    };
    JSProcessor.isJSON = function(code) {
        if (JSProcessor.isObject(code)) {
            try {
                JSON.parse(code);
                return true;
            } catch (e) {}
        }
        return false;
    };
    JSProcessor.isArray = function(code) {
        return /^\s*\[[\s\S]*\]\s*$/.test(code);
    };
    JSProcessor.isObject = function(code) {
        return /^\s*\{[\s\S]*\}\s*$/.test(code);
    };
    JSProcessor.GET_LOCATION_METH_NAME = "__get$Loc";
    JSProcessor.SET_LOCATION_METH_NAME = "__set$Loc";
    JSProcessor.SET_PROPERTY_METH_NAME = "__set$";
    JSProcessor.GET_PROPERTY_METH_NAME = "__get$";
    JSProcessor.PROCESS_SCRIPT_METH_NAME = "__proc$Script";
    JSProcessor.isScriptProcessed = function(code) {
        return new RegExp([ JSProcessor.GET_LOCATION_METH_NAME, JSProcessor.SET_LOCATION_METH_NAME, JSProcessor.SET_PROPERTY_METH_NAME, JSProcessor.GET_PROPERTY_METH_NAME, JSProcessor.PROCESS_SCRIPT_METH_NAME ].join("|").replace(/\$/, "\\$")).test(code);
    };
    JSProcessor.wrappedProperties = {
        protocol: true,
        host: true,
        hostname: true,
        port: true,
        pathname: true,
        origin: true,
        search: true,
        data: true,
        text: true,
        innerHTML: true,
        src: true,
        href: true,
        action: true,
        domain: true,
        cookie: true,
        referrer: true,
        location: true,
        onbeforeunload: true,
        URL: true,
        manifest: true,
        target: true,
        background: true,
        backgroundImage: true,
        listStyle: true,
        listStyleImage: true,
        cursor: true,
        borderImage: true,
        cssText: true
    };
    function needToWrap(property) {
        return JSProcessor.wrappedProperties[property];
    }
    JSProcessor.isDataScript = function(code) {
        return JSProcessor.isObject(code) || JSProcessor.isArray(code);
    };
    JSProcessor.process = function(code) {
        var isJSON = JSProcessor.isJSON(code), isObject = JSProcessor.isObject(code);
        codegenOpts.json = isJSON;
        var ast = null;
        try {
            ast = JSParsingTools.parse(isObject && !isJSON ? "(" + code + ")" : code);
        } catch (e) {
            return code;
        }
        var modified = modify(ast);
        if (modified) {
            var result = JSParsingTools.generate(ast, codegenOpts);
            if (!/;\s*$/.test(code)) {
                result = result.replace(/;\s*$/, "");
            }
            if (isObject && !isJSON) {
                result = result.replace(/^\(|\)$/g, "");
            }
            return result;
        }
        return code;
    };
    var modifiers = [ {
        modifier: evalArgument,
        condition: function(astNode) {
            if (astNode.type === SYNTAX.CallExpression) {
                if (astNode.callee.type === SYNTAX.Identifier && /^(eval|setTimeout|setInterval)$/.test(astNode.callee.name)) {
                    return true;
                }
                if (astNode.callee.type === SYNTAX.MemberExpression && /^(eval|setTimeout|setInterval)$/.test(astNode.callee.property.name || astNode.callee.property.value)) {
                    return true;
                }
            }
            return false;
        }
    }, {
        modifier: callEvalArgument,
        condition: function(astNode) {
            if (astNode.type === SYNTAX.CallExpression) {
                if (astNode.callee.type === SYNTAX.MemberExpression && astNode.callee.property.name === "call") {
                    var obj = astNode.callee.object;
                    if (obj.type === SYNTAX.MemberExpression && /^(eval|setTimeout|setInterval)$/.test(obj.property.value || obj.property.name)) {
                        return true;
                    }
                }
                if (astNode.callee.type === SYNTAX.MemberExpression && astNode.callee.property.name === "call" && /^(eval|setTimeout|setInterval)$/.test(astNode.callee.object.name)) {
                    return true;
                }
            }
            return false;
        }
    }, {
        modifier: applyEvalArgument,
        condition: function(astNode) {
            if (astNode.type === SYNTAX.CallExpression) {
                if (astNode.callee.type === SYNTAX.MemberExpression && astNode.callee.property.name === "apply") {
                    var obj = astNode.callee.object;
                    if (obj.type === SYNTAX.MemberExpression && /^(eval|setTimeout|setInterval)$/.test(obj.property.value || obj.property.name)) {
                        return true;
                    }
                }
                if (astNode.callee.type === SYNTAX.MemberExpression && astNode.callee.property.name === "apply" && /^(eval|setTimeout|setInterval)$/.test(astNode.callee.object.name)) {
                    return true;
                }
            }
            return false;
        }
    }, {
        modifier: getLocation,
        condition: function(astNode, parent) {
            if (astNode.type === SYNTAX.Identifier) {
                if (astNode.name !== "location") {
                    return false;
                }
                if (parent.type === SYNTAX.VariableDeclarator && parent.id === astNode) {
                    return false;
                }
                if (parent.type === SYNTAX.AssignmentExpression && parent.left === astNode) {
                    return false;
                }
                if (parent.type === SYNTAX.MemberExpression) {
                    return false;
                }
                if (parent.type === SYNTAX.Property && parent.key === astNode) {
                    return false;
                }
                if (parent.type === SYNTAX.UpdateExpression && parent.operator === "++" || parent.operator === "--") {
                    return false;
                }
                if ((parent.type === SYNTAX.FunctionExpression || parent.type === SYNTAX.FunctionDeclaration) && parent.params.indexOf(astNode) !== -1) {
                    return false;
                }
                if (parent.type === SYNTAX.CallExpression && parent.callee.name === JSProcessor.GET_LOCATION_METH_NAME) {
                    return false;
                }
                return true;
            }
            return false;
        }
    }, {
        modifier: getLocationMember,
        condition: function(astNode) {
            if (astNode.type === SYNTAX.MemberExpression) {
                if (astNode.object.name === "location") {
                    return true;
                }
            }
            return false;
        }
    }, {
        modifier: setLocation,
        condition: function(astNode) {
            if (astNode.type === SYNTAX.AssignmentExpression && astNode.operator === "=") {
                var leftOperand = astNode.left;
                if (leftOperand.type === SYNTAX.Identifier && leftOperand.name === "location") {
                    return true;
                }
            }
            return false;
        }
    }, {
        modifier: memberSet,
        condition: function(astNode) {
            if (astNode.type === SYNTAX.AssignmentExpression && !astNode.computed && astNode.operator === "=") {
                var leftOperand = astNode.left;
                if (leftOperand.type === SYNTAX.MemberExpression && leftOperand.property.type === SYNTAX.Identifier) {
                    if (needToWrap(leftOperand.property.name)) {
                        return true;
                    }
                }
            }
            return false;
        }
    }, {
        modifier: memberGet,
        condition: function(astNode, parent) {
            if (astNode.type === SYNTAX.MemberExpression && !astNode.computed) {
                if (!needToWrap(astNode.property.name)) {
                    return false;
                }
                if (parent.type === SYNTAX.AssignmentExpression && parent.left === astNode) {
                    return false;
                }
                if (parent.type === SYNTAX.UnaryExpression && parent.operator === "delete") {
                    return false;
                }
                if (parent.type === SYNTAX.UpdateExpression && parent.operator === "++" || parent.operator === "--") {
                    return false;
                }
                if (parent.type === SYNTAX.CallExpression && parent.callee === astNode) {
                    return false;
                }
                if (parent.type === SYNTAX.NewExpression && parent.callee === astNode) {
                    return false;
                }
                return true;
            }
            return false;
        }
    }, {
        modifier: computedMemberGet,
        condition: function(astNode, parent) {
            if (astNode.type === SYNTAX.MemberExpression && astNode.computed) {
                if (parent.type === SYNTAX.AssignmentExpression && parent.left === astNode) {
                    return false;
                }
                if (parent.type === SYNTAX.CallExpression && parent.callee === astNode) {
                    return false;
                }
                if (parent.type === SYNTAX.UnaryExpression && parent.operator === "delete") {
                    return false;
                }
                if (parent.type === SYNTAX.UpdateExpression && parent.operator === "++" || parent.operator === "--") {
                    return false;
                }
                if (parent.type === SYNTAX.CallExpression && parent.callee === astNode) {
                    return false;
                }
                if (parent.type === SYNTAX.NewExpression && parent.callee === astNode) {
                    return false;
                }
                if (astNode.property.type === SYNTAX.Literal && !needToWrap(astNode.property.value)) {
                    return false;
                }
                return true;
            }
            return false;
        }
    }, {
        modifier: computedMemberSet,
        condition: function(astNode) {
            if (astNode.type === SYNTAX.AssignmentExpression && astNode.operator === "=") {
                if (astNode.left.type === SYNTAX.MemberExpression && astNode.left.computed) {
                    if (astNode.left.property.type === SYNTAX.Literal) {
                        return needToWrap(astNode.left.property.value);
                    }
                    return true;
                }
            }
            return false;
        }
    }, {
        modifier: concatOperator,
        condition: function(astNode) {
            if (astNode.type === SYNTAX.AssignmentExpression && astNode.operator === "+=") {
                return true;
            }
            return false;
        }
    } ];
    function modify(ast, parent, key) {
        var modified = false;
        if (!ast || typeof ast !== "object") {
            return modified;
        }
        if (ast.type) {
            for (var i = 0; i < modifiers.length; i++) {
                if (modifiers[i].condition(ast, parent)) {
                    var needToModify = modifiers[i].modifier(ast, parent, key);
                    modified = true;
                    if (needToModify) {
                        modified = modify(parent[key], parent, key) || modified;
                    }
                }
            }
        }
        for (var astKey in ast) {
            if (ast.hasOwnProperty(astKey)) {
                var childNode = ast[astKey];
                if (Object.prototype.toString.call(childNode) === "[object Array]") {
                    for (var j = 0; j < childNode.length; j++) {
                        modified = modify(childNode[j], ast, astKey) || modified;
                    }
                } else {
                    modified = modify(childNode, ast, astKey) || modified;
                }
            }
        }
        return modified;
    }
    function getProcessScriptMethAst(args) {
        return {
            type: SYNTAX.CallExpression,
            callee: {
                type: SYNTAX.Identifier,
                name: JSProcessor.PROCESS_SCRIPT_METH_NAME
            },
            arguments: [ args[0] ]
        };
    }
    function getGetLocationMethAst() {
        return {
            type: SYNTAX.CallExpression,
            callee: {
                type: SYNTAX.Identifier,
                name: JSProcessor.GET_LOCATION_METH_NAME
            },
            arguments: [ {
                type: SYNTAX.Identifier,
                name: "location"
            } ]
        };
    }
    function getSetLocationMethAst(value) {
        return {
            type: SYNTAX.CallExpression,
            callee: {
                type: SYNTAX.MemberExpression,
                computed: false,
                object: {
                    type: SYNTAX.FunctionExpression,
                    id: null,
                    params: [],
                    defaults: [],
                    body: {
                        type: SYNTAX.BlockStatement,
                        body: [ {
                            type: SYNTAX.ReturnStatement,
                            argument: {
                                type: SYNTAX.LogicalExpression,
                                operator: "||",
                                left: {
                                    type: SYNTAX.CallExpression,
                                    callee: {
                                        type: SYNTAX.Identifier,
                                        name: JSProcessor.SET_LOCATION_METH_NAME
                                    },
                                    arguments: [ {
                                        type: SYNTAX.Identifier,
                                        name: "location"
                                    }, value ]
                                },
                                right: {
                                    type: SYNTAX.AssignmentExpression,
                                    operator: "=",
                                    left: {
                                        type: SYNTAX.Identifier,
                                        name: "location"
                                    },
                                    right: value
                                }
                            }
                        } ]
                    },
                    rest: null,
                    generator: false,
                    expression: false
                },
                property: {
                    type: SYNTAX.Identifier,
                    name: "apply"
                }
            },
            arguments: [ {
                type: SYNTAX.ThisExpression
            } ]
        };
    }
    function getSetMethAst(propertyName, obj, value) {
        return {
            type: SYNTAX.CallExpression,
            callee: {
                type: SYNTAX.Identifier,
                name: JSProcessor.SET_PROPERTY_METH_NAME
            },
            arguments: [ obj, {
                type: SYNTAX.Literal,
                value: propertyName,
                raw: '"' + propertyName + '"'
            }, value ]
        };
    }
    function getGetMethAst(propertyName, owner) {
        return {
            type: SYNTAX.CallExpression,
            callee: {
                type: SYNTAX.Identifier,
                name: JSProcessor.GET_PROPERTY_METH_NAME
            },
            arguments: [ owner, {
                type: SYNTAX.Literal,
                value: propertyName,
                raw: '"' + propertyName + '"'
            } ]
        };
    }
    function getGetComputedMethAst(property, owner) {
        return {
            type: SYNTAX.CallExpression,
            callee: {
                type: SYNTAX.Identifier,
                name: JSProcessor.GET_PROPERTY_METH_NAME
            },
            arguments: [ owner, property ]
        };
    }
    function getSetComputedMethAst(property, owner, value) {
        return {
            type: SYNTAX.CallExpression,
            callee: {
                type: SYNTAX.Identifier,
                name: JSProcessor.SET_PROPERTY_METH_NAME
            },
            arguments: [ owner, property, value ]
        };
    }
    function getConcatOperatorAst(left, right) {
        return {
            type: SYNTAX.AssignmentExpression,
            operator: "=",
            left: left,
            right: {
                type: SYNTAX.BinaryExpression,
                operator: "+",
                left: left,
                right: right
            }
        };
    }
    function updateAstNode(node, newNode, parent, key) {
        if (key === "arguments" || key === "elements" || key === "expressions") {
            var index = parent[key].indexOf(node);
            parent[key][index] = newNode;
        } else {
            parent[key] = newNode;
        }
    }
    function evalArgument(astNode) {
        if (!astNode.arguments.length) {
            return false;
        }
        var newArg = getProcessScriptMethAst(astNode.arguments);
        astNode.arguments[0] = newArg;
        return false;
    }
    function callEvalArgument(astNode) {
        var newArg = getProcessScriptMethAst([ astNode.arguments[1] ]);
        astNode.arguments[1] = newArg;
        return false;
    }
    function applyEvalArgument(astNode) {
        var newArg = getProcessScriptMethAst([ astNode.arguments[1].elements[0] ]);
        astNode.arguments[1].elements[0] = newArg;
        return false;
    }
    function getLocation(astNode, parent, key) {
        var newNode = getGetLocationMethAst();
        updateAstNode(astNode, newNode, parent, key);
        return false;
    }
    function getLocationMember(astNode) {
        var newNode = getGetLocationMethAst(astNode.object);
        astNode.object = newNode;
        return false;
    }
    function setLocation(astNode, parent, key) {
        var newNode = getSetLocationMethAst(astNode.right);
        updateAstNode(astNode, newNode, parent, key);
        return false;
    }
    function memberSet(astNode, parent, key) {
        var newNode = getSetMethAst(astNode.left.property.name, astNode.left.object, astNode.right);
        updateAstNode(astNode, newNode, parent, key);
        return true;
    }
    function memberGet(astNode, parent, key) {
        var newNode = getGetMethAst(astNode.property.name, astNode.object);
        updateAstNode(astNode, newNode, parent, key);
        return true;
    }
    function computedMemberGet(astNode, parent, key) {
        var newNode = getGetComputedMethAst(astNode.property, astNode.object);
        updateAstNode(astNode, newNode, parent, key);
        return true;
    }
    function computedMemberSet(astNode, parent, key) {
        var newNode = getSetComputedMethAst(astNode.left.property, astNode.left.object, astNode.right);
        updateAstNode(astNode, newNode, parent, key);
        return true;
    }
    function concatOperator(astNode, parent, key) {
        var newNode = getConcatOperatorAst(astNode.left, astNode.right);
        updateAstNode(astNode, newNode, parent, key);
        return true;
    }
    if (typeof module !== "undefined" && module.exports) {
        module.exports = JSProcessor;
    } else {
        HammerheadClient.define("Shared.JSProcessor", function() {
            this.exports = JSProcessor;
        });
    }
})();