(function() {
    var PageProc = {};
    var isNode = typeof module !== "undefined" && module.exports;
    var SharedConst = null, UrlUtil = null, NativeMethods = null, JSProcessor = null, $ = null;
    if (isNode) {
        SharedConst = require("./const");
        UrlUtil = require("./url_util");
        JSProcessor = require("./js_processor");
        module.exports = PageProc;
    } else {
        HammerheadClient.define("Shared.PageProc", function(require) {
            $ = require("jQuery");
            SharedConst = require("Shared.Const");
            UrlUtil = require("Shared.UrlUtil");
            NativeMethods = require("DOMSandbox.NativeMethods");
            JSProcessor = require("Shared.JSProcessor");
            this.exports = PageProc;
        });
    }
    var CSS_URL_PROPERTY_VALUE_PATTERN = /(url\s*\(\s*)(?:(')([^\s']+)(')|(")([^\s"]+)(")|([^\s\)]+))(\s*\))|(@import\s+)(?:(')([^\s']+)(')|(")([^\s"]+)("))/g, EVENTS = [ "onblur", "onchange", "onclick", "oncontextmenu", "oncopy", "oncut", "ondblclick", "onerror", "onfocus", "onfocusin", "onfocusout", "onhashchange", "onkeydown", "onkeypress", "onkeyup", "onload", "onmousedown", "onmouseenter", "onmouseleave", "onmousemove", "onmouseout", "onmouseover", "onmouseup", "onmousewheel", "onpaste", "onreset", "onresize", "onscroll", "onselect", "onsubmit", "ontextinput", "onunload", "onwheel" ], EVENT_HANDLERS_SELECTOR = "[" + EVENTS.join("],[") + "]", BOM_REGEX = new RegExp("^(\\xEF\\xBB\\xBF|" + "\\xFE\\xFF|" + "\\xFF\\xFE|" + "\\x00\\x00\\xFE\\xFF|" + "\\xFF\\xFE\\x00\\x00|" + "\\x2B\\x2F\\x76\\x38|" + "\\x2B\\x2F\\x76\\x39|" + "\\x2B\\x2F\\x76\\x2B|" + "\\x2B\\x2F\\x76\\x2F|" + "\\xF7\\x64\\x4C|" + "\\xDD\\x73\\x66\\x73|" + "\\x0E\\xFE\\xFF|" + "\\xFB\\xEE\\x28|" + "\\x84\\x31\\x95\\x33)"), HTML_COMMENT_REG_EX = /^(\s)*<!--([\s\S]*)-->(\s)*$/, CDATA_REG_EX = /^(\s)*\/\/<!\[CDATA\[([\s\S]*)\/\/\]\]>(\s)*$/, URL_ATTRS = [ "href", "src", "action", "manifest", "data" ], URL_ATTR_TAGS = {
        href: [ "a", "link", "image", "area", "base" ],
        src: [ "img", "embed", "script", "source", "video", "audio", "input", "frame", "iframe" ],
        action: [ "form" ],
        manifest: [ "html" ],
        data: [ "object" ]
    }, TARGET_ATTR_TAGS = {
        a: true,
        form: true,
        area: true,
        base: true
    };
    PageProc.URL_ATTRS = URL_ATTRS;
    PageProc.URL_ATTR_TAGS = URL_ATTR_TAGS;
    PageProc.TARGET_ATTR_TAGS = TARGET_ATTR_TAGS;
    var getAttr = function($el, attr) {
        return isNode ? $el[0].attribs[attr] : NativeMethods.getAttribute.call($el[0], attr);
    };
    var getTagName = function($el) {
        return isNode ? $el[0].name : $el[0].tagName;
    };
    var setAttr = function($el, attr, value) {
        return isNode ? $el[0].attribs[attr] = value : NativeMethods.setAttribute.call($el[0], attr, value);
    };
    var processTargetBlank = function($el) {
        var attrValue = getAttr($el, "target");
        attrValue = attrValue && attrValue.replace(/\s/g, "");
        if (attrValue === "_blank" || attrValue === "blank") {
            setAttr($el, "target", "_self");
        }
    }, processStylesheetElement = function($el, urlReplacer) {
        if (isNode) {
            var contentChild = $el[0].children.length ? $el[0].children[0] : null;
            if (contentChild && contentChild.data && urlReplacer) {
                contentChild.data = PageProc.processStylesheet(contentChild.data, urlReplacer);
            }
        }
    }, processStyleAttr = function($el, urlReplacer) {
        var style = getAttr($el, "style");
        if (style) {
            setAttr($el, "style", PageProc.processStylesheet(style, urlReplacer));
        }
    }, processScriptElement = function($script) {
        var scriptContent = "";
        if (isNode) {
            var contentChild = $script[0].children.length ? $script[0].children[0] : null;
            scriptContent = !!contentChild ? contentChild.data : "";
        } else {
            scriptContent = $script[0].text;
            var scriptProcessedOnServer = JSProcessor.isScriptProcessed(scriptContent);
            if (scriptProcessedOnServer) {
                return;
            }
        }
        if (!scriptContent) {
            return;
        }
        var scriptType = getAttr($script, "type"), executableScriptTypesRegEx = /(application\/((x-)?ecma|(x-)?java)script)|(text\/)(javascript(1\.{0-5})?|((x-)?ecma|x-java|js|live)script)/, isExecutableScript = !scriptType || executableScriptTypesRegEx.test(scriptType);
        if (isExecutableScript) {
            var result = scriptContent, hasCommentPrefix = HTML_COMMENT_REG_EX.test(result), hasCDATA = CDATA_REG_EX.test(result);
            if (hasCommentPrefix) {
                result = result.replace(HTML_COMMENT_REG_EX, "$2");
            }
            if (hasCDATA) {
                result = result.replace(CDATA_REG_EX, "$2");
            }
            result = PageProc.processScript(result);
            if (hasCommentPrefix) {
                result = "\n<!--\n" + result + "\n//-->";
            }
            if (hasCDATA) {
                result = "\n//<![CDATA[\n" + result + "//]]>";
            }
            if (isNode) {
                $script[0].children[0].data = result;
            } else {
                $script[0].text = result;
            }
        }
    }, processEvtHandlers = function($el) {
        if (isNode) {
            for (var i = 0; i < EVENTS.length; i++) {
                var attr = getAttr($el, EVENTS[i]);
                if (attr) {
                    setAttr($el, EVENTS[i], PageProc.processScript(attr, true));
                }
            }
        }
    }, processUrlAttrs = function($el, urlReplacer, pattern) {
        if (urlReplacer && pattern.urlAttr) {
            var storedUrlAttr = PageProc.getStoredAttrName(pattern.urlAttr), resourceUrl = getAttr($el, pattern.urlAttr), processedOnServer = !!getAttr($el, storedUrlAttr);
            if ((resourceUrl || resourceUrl === "") && !processedOnServer) {
                var isIframe = getTagName($el).toLowerCase() === "iframe";
                if (isIframe) {
                    var isRelativePath = !UrlUtil.parseUrl(resourceUrl).host;
                    if (!isRelativePath) {
                        var originUrlObj = UrlUtil.parseProxyUrl(urlReplacer("/")).originResourceInfo, originUrl = UrlUtil.formatUrl(originUrlObj);
                        if (!UrlUtil.sameOriginCheck(originUrl, resourceUrl)) {
                            return;
                        }
                    }
                }
                if (UrlUtil.isSupportedProtocol(resourceUrl)) {
                    setAttr($el, pattern.urlAttr, resourceUrl ? urlReplacer(resourceUrl, isIframe) : "");
                    setAttr($el, storedUrlAttr, resourceUrl);
                }
            }
        }
    }, processAutoComplete = function($el) {
        setAttr($el, "autocomplete", "off");
    }, processMetaElement = function($el, urlReplacer, pattern) {
        if (getAttr($el, "http-equiv").toLowerCase() === "refresh") {
            var attr = getAttr($el, pattern.urlAttr);
            attr = attr.replace(/(url=)(.*)$/i, function() {
                return arguments[1] + urlReplacer(arguments[2]);
            });
            setAttr($el, pattern.urlAttr, attr);
        }
    };
    var ELEMENT_PROCESSOR_PATTERNS = [ {
        selector: URL_ATTR_TAGS.href.join(", "),
        urlAttr: "href",
        elementProcessors: [ processTargetBlank, processUrlAttrs ]
    }, {
        selector: URL_ATTR_TAGS.src.join(", "),
        urlAttr: "src",
        elementProcessors: [ processTargetBlank, processUrlAttrs ]
    }, {
        selector: URL_ATTR_TAGS.action.join(", "),
        urlAttr: "action",
        elementProcessors: [ processTargetBlank, processUrlAttrs ]
    }, {
        selector: URL_ATTR_TAGS.manifest.join(", "),
        urlAttr: "manifest",
        elementProcessors: [ processUrlAttrs ]
    }, {
        selector: URL_ATTR_TAGS.data.join(", "),
        urlAttr: "data",
        elementProcessors: [ processUrlAttrs ]
    }, {
        selector: "meta[http-equiv]",
        urlAttr: "content",
        elementProcessors: [ processMetaElement ]
    }, {
        selector: "*",
        elementProcessors: [ processStyleAttr ]
    }, {
        selector: "script",
        elementProcessors: [ processScriptElement ]
    }, {
        selector: "style",
        elementProcessors: [ processStylesheetElement ]
    }, {
        selector: "input",
        elementProcessors: [ processAutoComplete ]
    }, {
        selector: EVENT_HANDLERS_SELECTOR,
        elementProcessors: [ processEvtHandlers ]
    } ];
    var getElementForSelectorCheck = function($el) {
        if (!isNode && $.browser.msie && parseInt($.browser.version, 10) === 9 && $el[0].tagName.toLowerCase() === "script") {
            var $clone = $(NativeMethods.cloneNode.call($el[0], false));
            $clone[0].src = $clone[0].innerHTML = "";
            return $clone;
        }
        return $el;
    };
    var isTestCafeElement = function($el) {
        var el = $el[0];
        return typeof el.className === "string" && el.className.indexOf(SharedConst.TEST_CAFE_UI_CLASSNAME_POSTFIX) > -1;
    };
    PageProc.processPage = function($, urlReplacer) {
        for (var i = 0; i < ELEMENT_PROCESSOR_PATTERNS.length; i++) {
            var pattern = ELEMENT_PROCESSOR_PATTERNS[i];
            $(pattern.selector).each(function() {
                if (!this[SharedConst.ELEMENT_PROCESSED_FLAG]) {
                    for (var j = 0; j < pattern.elementProcessors.length; j++) {
                        pattern.elementProcessors[j]($(this), urlReplacer, pattern);
                    }
                }
            });
        }
    };
    PageProc.processElement = function($el, urlReplacer) {
        var $elementForSelectorCheck = getElementForSelectorCheck($el);
        for (var i = 0; i < ELEMENT_PROCESSOR_PATTERNS.length; i++) {
            var pattern = ELEMENT_PROCESSOR_PATTERNS[i];
            if ($elementForSelectorCheck.is(pattern.selector) && !isTestCafeElement($el)) {
                for (var j = 0; j < pattern.elementProcessors.length; j++) {
                    pattern.elementProcessors[j]($el, urlReplacer, pattern);
                }
            }
        }
    };
    PageProc.getBOM = function(text) {
        var match = text.match(BOM_REGEX);
        return match ? match[0] : null;
    };
    PageProc.getStoredAttrName = function(attr) {
        return attr + SharedConst.DOM_SANDBOX_STORED_ATTR_POSTFIX;
    };
    PageProc.processScript = function(text, withoutHeader) {
        var bom = PageProc.getBOM(text);
        if (bom) {
            text = text.replace(bom, "");
        }
        text = JSProcessor.process(text);
        if (!JSProcessor.isDataScript(text) && !withoutHeader && text.indexOf(SharedConst.DOM_SANDBOX_OVERRIDE_DOM_METHOD_NAME) === -1) {
            var overrideDomMeth = 'window["' + SharedConst.DOM_SANDBOX_OVERRIDE_DOM_METHOD_NAME + '"]';
            text = '\r\ntypeof window !== "undefined" && ' + overrideDomMeth + " && " + overrideDomMeth + "();\r\n" + JSProcessor.MOCK_ACCESSORS + text;
        }
        return bom ? bom + text : text;
    };
    PageProc.processManifest = function(manifest, urlReplacer) {
        var lines = manifest.split("\n"), trim = function(str) {
            return str.replace(/^\s+|\s+$/g, "");
        };
        for (var i = 0; i < lines.length; i++) {
            var line = trim(lines[i]);
            if (line && line !== "CACHE MANIFEST" && line !== "NETWORK:" && line !== "FALLBACK:" && line !== "CACHE:" && line[0] !== "#" && line !== "*") {
                var isFallbackItem = line.indexOf(" ") !== -1;
                if (isFallbackItem) {
                    var urls = line.split(" ");
                    lines[i] = urlReplacer(urls[0]) + " " + urlReplacer(urls[1]);
                } else {
                    lines[i] = urlReplacer(line);
                }
            }
        }
        return lines.join("\n");
    };
    PageProc.processStylesheet = function(css, urlReplacer) {
        if (typeof css !== "string") {
            return css;
        }
        css = css.replace(/\s*:\s*hover(\W)/gi, "[" + SharedConst.TEST_CAFE_HOVER_PSEUDO_CLASS_ATTR + "]$1");
        return css.replace(CSS_URL_PROPERTY_VALUE_PATTERN, function() {
            var prefix = arguments[1] || arguments[10], openQuote = arguments[2] || arguments[5] || arguments[11] || arguments[14] || "", resourceUrl = arguments[3] || arguments[6] || arguments[8] || arguments[12] || arguments[15], closeQuote = arguments[4] || arguments[7] || arguments[13] || arguments[16] || "", postfix = arguments[9] || "";
            var proxyUrl = urlReplacer(resourceUrl);
            return prefix + openQuote + proxyUrl + closeQuote + postfix;
        });
    };
    PageProc.isCrossDomainIframe = function(el, value) {
        var $el = $(el), isIframe = getTagName($el).toLowerCase() === "iframe";
        if (isIframe) {
            var src = value || getAttr($el, PageProc.getStoredAttrName("src")) || getAttr($el, "src");
            if (src && UrlUtil.isSupportedProtocol(src)) {
                return !UrlUtil.sameOriginCheck(location.toString(), src);
            }
        }
        return false;
    };
})();