var ERR = require("./../server/server_errs"), async = require("async"), child_process = require("child_process"), dns = require("dns"), fs = require("fs"), os = require("os"), path = require("path"), http = require("http"), InstalledBrowsers = require("./installed_browsers"), TCPPorts = require("./tcp_ports"), Validator = require("./validator");

var CFG_FILE_NAME = "config.json", PACKAGE_JSON_FILE_NAME = "package.json", DEFAULT_CONTROL_PANEL_PORT = 1337, DEFAULT_SERVICE_PORT = 1338, CHECKING_HOST_PORT = 1400, CHECKING_REQUEST_TIMEOUT = 3e3, WIN_PLATFORM = /^win/.test(process.platform), MAC_PLATFORM = process.platform === "darwin", WIN_DEFAULT_TESTS_DIR = WIN_PLATFORM ? path.join(process.env["USERPROFILE"], "Documents", "TestCafe", "Tests") : "", EXAMPLES_DIR = path.join(__dirname, "../../examples"), DEFAULT_TESTS_DIR = path.join(__dirname, "../../my_tests"), INSTALL_DATE_REG_KEY = "HKEY_CLASSES_ROOT\\Licenses\\6F0F8269-1516-44C6-BD30-0E90BE27871C", LICENSE_DATE_REG_KEY = "HKEY_CLASSES_ROOT\\Licenses\\0378852D-D597-4A32-B6D9-680A16A3CDA6";

function copyDirRecursiveSync(from, to) {
    fs.readdirSync(from).forEach(function(fileName) {
        var filePath = path.join(from, fileName), stats = fs.statSync(filePath);
        if (stats.isDirectory()) {
            fs.mkdirSync(path.join(to, fileName));
            copyDirRecursiveSync(filePath, path.join(to, fileName));
        } else {
            fs.writeFileSync(path.join(to, fileName), fs.readFileSync(filePath));
        }
    });
}

function mkdirRecursiveSync(dirPath) {
    var dirPathArr = dirPath.split(path.sep), currentPath = dirPathArr.shift();
    while (dirPathArr.length) {
        currentPath = path.join(currentPath, dirPathArr.shift());
        if (!fs.existsSync(currentPath)) {
            fs.mkdirSync(currentPath);
        }
    }
}

var Config = module.exports = function() {
    this.initedFromObj = false;
    this.hostname = "";
    this.controlPanelHost = "";
    this.controlPanelPort = 0;
    this.servicePort = 0;
    this.testsDir = "";
    this.browsers = null;
};

Config.TESTCAFE_DIR = path.join(__dirname, "../../");

Config.CFG_FILE_PATH = path.join(Config.TESTCAFE_DIR, CFG_FILE_NAME);

Config.PACKAGE_JSON_FILE_PATH = path.join(Config.TESTCAFE_DIR, PACKAGE_JSON_FILE_NAME);

Config.VERSION_PATTERN = /^[0-9]+\.[0-9]+\.[0-9]+$/;

Config.HOSTNAME_PATTERN = /^[a-zA-Z0-9\-\.]+$/;

Config.CHECK_NEW_VERSION_REQUEST_CONFIG = {
    hostname: "testcafe.devexpress.com",
    port: 80,
    path: "/version.json",
    method: "GET"
};

Config.prototype.initConfig = function(startupCfg, callback) {
    if (startupCfg) {
        this._initFromObj(startupCfg, callback);
    } else {
        this._initFromFile(callback);
    }
};

Config.prototype._initFromObj = function(cfgObj, callback) {
    this.initedFromObj = true;
    var errs = Validator.validate(cfgObj, true);
    this.hostname = this._getHostname(cfgObj.hostname);
    this.servicePort = cfgObj.servicePort;
    this.controlPanelPort = cfgObj.controlPanelPort;
    this.testsDir = cfgObj.testsDir;
    this.browsers = cfgObj.browsers;
    this.controlPanelHost = [ this.hostname, ":", this.controlPanelPort ].join("");
    callback(errs.length ? errs : null);
};

Config.prototype._initFromFile = function(callback) {
    var config = this;
    this.readCfgFile(function(readedCfg, readErr) {
        config._prepareUserCfg(readedCfg || {}, function(validCfg, validErrs) {
            config.controlPanelPort = validCfg.controlPanelPort;
            config.servicePort = validCfg.servicePort;
            config.testsDir = validCfg.testsDir;
            config.hostname = validCfg.hostname;
            config.browsers = validCfg.browsers;
            config._getConfig(function() {
                var readed = !!readedCfg;
                callback(readed ? readErr || validErrs : null);
            });
        });
    });
};

Config.prototype.setTestsDir = function(testsDir, callback) {
    var configReadErr = {
        code: ERR.CONFIG_READ_FILE_FAILED
    }, configWriteErr = {
        code: ERR.CONFIG_SAVING_FAILED
    };
    if (!this.initedFromObj) {
        var fileCfg = null;
        if (!fs.existsSync(Config.CFG_FILE_PATH)) {
            fs.writeFileSync(Config.CFG_FILE_PATH, "{}");
        }
        try {
            fileCfg = JSON.parse(fs.readFileSync(Config.CFG_FILE_PATH).toString());
        } catch (err) {
            callback(err && configReadErr);
            return;
        }
        fileCfg.testsDir = this.testsDir = testsDir;
        fs.writeFile(Config.CFG_FILE_PATH, JSON.stringify(fileCfg, null, 5), function(err) {
            callback(err && configWriteErr);
        });
    } else {
        this.testsDir = testsDir;
        callback();
    }
};

Config.prototype.validateBrowsers = function(browsers, callback) {
    var result = {}, browserNames = Object.keys(browsers);
    async.forEachSeries(browserNames, function(browserName, done) {
        var browser = browsers[browserName], path = browsers[browserName].path;
        fs.stat(path, function(err, stat) {
            result[browserName] = {
                path: path,
                cmd: browser.cmd,
                icon: browser.icon
            };
            result[browserName].pathExists = err || !MAC_PLATFORM && !stat.isFile() ? false : true;
            done();
        });
    }, function() {
        callback(result);
    });
};

Config.prototype.checkUpdates = function(callback) {
    var newVersion = "", config = this;
    this.getCurrentVersion(function(curVersion) {
        config._getActualVersion(function(version) {
            if (Config.VERSION_PATTERN.test(curVersion) && Config.VERSION_PATTERN.test(version)) {
                var curVersionArr = curVersion.split(".").map(function(number) {
                    return parseInt(number, 10);
                }), actualVersionArr = version.split(".").map(function(number) {
                    return parseInt(number, 10);
                });
                if (curVersionArr[0] < actualVersionArr[0] || curVersionArr[0] === actualVersionArr[0] && (curVersionArr[1] < actualVersionArr[1] || curVersionArr[1] === actualVersionArr[1] && curVersionArr[2].toString() < actualVersionArr[2].toString())) {
                    newVersion = version;
                }
            }
            callback(curVersion, newVersion);
        });
    });
};

Config.prototype.updateConfig = function(configOptions, callback) {
    var config = this;
    var writeToConfig = function(cfg) {
        Object.keys(cfg).forEach(function(key) {
            if (!cfg[key]) {
                delete cfg[key];
            }
        });
        fs.writeFile(Config.CFG_FILE_PATH, JSON.stringify(cfg, false, 5), function(err) {
            callback(err ? {
                code: ERR.CONFIG_SAVING_FAILED
            } : null);
        });
    };
    fs.exists(Config.CFG_FILE_PATH, function(exists) {
        if (exists) {
            config.readCfgFile(function(cfg, err) {
                if (!err) {
                    Object.keys(configOptions).forEach(function(key) {
                        cfg[key] = configOptions[key];
                    });
                    writeToConfig(cfg);
                } else {
                    writeToConfig(configOptions);
                }
            });
        } else {
            writeToConfig(configOptions);
        }
    });
};

Config.prototype.readCfgFile = function(callback) {
    var cfgData = null, cfg = null;
    if (!fs.existsSync(Config.CFG_FILE_PATH)) {
        callback(null);
        return;
    }
    try {
        cfgData = fs.readFileSync(Config.CFG_FILE_PATH).toString();
    } catch (err) {
        callback(null, err);
        return;
    }
    try {
        cfg = JSON.parse(cfgData);
    } catch (parserErr) {
        callback(null, {
            code: ERR.CONFIG_PARSING_FAILED,
            parserErr: parserErr
        });
        return;
    }
    callback(cfg);
};

Config.prototype.isPortFree = function(port, callback) {
    TCPPorts.isPortFree(port, callback);
};

Config.prototype.addInstalledBrowsers = function(callback) {
    var config = this;
    config._getBrowsers(function(browsers) {
        if (browsers && Object.keys(browsers).length) {
            config.updateConfig({
                browsers: browsers
            }, function() {
                callback(browsers);
            });
            config.browsers = browsers;
        } else {
            callback(null);
        }
    });
};

Config.prototype.getCurrentVersion = function(callback) {
    fs.readFile(Config.PACKAGE_JSON_FILE_PATH, "utf8", function(err, jsonData) {
        callback(!err ? JSON.parse(jsonData).version : "");
    });
};

Config.prototype._getActualVersion = function(callback) {
    var version = "", body = "";
    var requestTimeout = setTimeout(function() {
        req.abort();
    }, CHECKING_REQUEST_TIMEOUT);
    var req = http.request(Config.CHECK_NEW_VERSION_REQUEST_CONFIG, function(res) {
        res.on("data", function(chunk) {
            body += chunk;
        });
        res.on("end", function() {
            try {
                version = JSON.parse(body.replace("﻿", "")).version;
            } catch (e) {
                version = "";
            }
            clearTimeout(requestTimeout);
            callback(version);
        });
    });
    req.on("error", function() {
        clearTimeout(requestTimeout);
        callback(version);
    });
    req.end();
};

Config.prototype._getConfig = function(callback) {
    var config = this;
    this.testsDir = this._getTestsDir(this.testsDir);
    this._getQualifiedHostname(this.hostname, function(hostname) {
        config.hostname = hostname;
        config._initPorts(function(controlPanelPort, servicePort) {
            config.controlPanelPort = controlPanelPort;
            config.servicePort = servicePort;
            config.controlPanelHost = [ config.hostname, ":", config.controlPanelPort ].join("");
            config._getBrowsers(function(browsers) {
                if (!config.browsers || !config.browsers.length) {
                    config.updateConfig({
                        browsers: browsers
                    }, function() {
                        config.browsers = browsers;
                        callback();
                    });
                } else {
                    config.browsers = browsers;
                    callback();
                }
            });
        });
    });
};

Config.prototype._getTestsDir = function(cfgTestsDir) {
    if (!fs.existsSync(cfgTestsDir)) {
        var testsDir = WIN_PLATFORM ? WIN_DEFAULT_TESTS_DIR : DEFAULT_TESTS_DIR;
        if (!fs.existsSync(testsDir)) {
            mkdirRecursiveSync(testsDir);
            copyDirRecursiveSync(EXAMPLES_DIR, testsDir);
        }
        return testsDir;
    }
    return cfgTestsDir;
};

Config.prototype._checkHostname = function(host, callback) {
    TCPPorts.getFree(CHECKING_HOST_PORT, 0, function(port) {
        var testServer = http.createServer(function(req, res) {
            res.writeHead(200, {
                "Content-Type": "text/plain"
            });
            res.end();
        }).listen(port, host, function() {
            dns.lookup(host, function(err) {
                testServer.close();
                callback(!err);
            });
        });
        testServer.on("error", function() {
            callback(false);
        });
    });
};

Config.prototype._getQualifiedHostname = function(cfgHostname, callback) {
    var hostname = cfgHostname || os.hostname(), config = this;
    config._checkHostname(hostname, function(isHostnameAvailable) {
        if (!isHostnameAvailable) {
            if (MAC_PLATFORM) {
                hostname = [ hostname, ".local" ].join("");
                config._checkHostname(hostname, function(isHostnameAvailable) {
                    if (!isHostnameAvailable) {
                        hostname = "127.0.0.1";
                    }
                    returnHostname();
                });
            } else {
                hostname = "127.0.0.1";
                returnHostname();
            }
        } else {
            returnHostname();
        }
    });
    function returnHostname() {
        callback(config._getHostname(hostname));
    }
};

Config.prototype._getHostname = function(cfgHostname) {
    var hostname = cfgHostname || os.hostname();
    return hostname && Config.HOSTNAME_PATTERN.test(hostname) ? hostname : "127.0.0.1";
};

Config.prototype._prepareUserCfg = function(cfgObj, callback) {
    var errs = Validator.validate(cfgObj);
    for (var i = errs.length - 1; i >= 0; i--) {
        delete cfgObj[errs[i].property];
    }
    callback(cfgObj, errs.length ? errs : null);
};

Config.prototype._initPorts = function(callback) {
    var controlPanelPort = this.controlPanelPort || DEFAULT_CONTROL_PANEL_PORT, servicePort = this.servicePort || DEFAULT_SERVICE_PORT, result = {};
    async.series({
        "Get control panel port": function(done) {
            TCPPorts.isPortFree(controlPanelPort, function(isFree) {
                if (isFree) {
                    result.controlPanelPort = controlPanelPort;
                    done();
                } else {
                    TCPPorts.getFree(controlPanelPort, null, function(port) {
                        result.controlPanelPort = port;
                        done();
                    });
                }
            });
        },
        "Get service port": function(done) {
            TCPPorts.isPortFree(servicePort, function(isFree) {
                if (isFree && servicePort !== result.controlPanelPort) {
                    result.servicePort = servicePort;
                    done();
                } else {
                    TCPPorts.getFree(servicePort, result.controlPanelPort, function(port) {
                        result.servicePort = port;
                        done();
                    });
                }
            });
        }
    }, function() {
        callback(result.controlPanelPort, result.servicePort);
    });
};

Config.prototype._getFreeBrowserName = function(name) {
    if (!this.browsers || !this.browsers.hasOwnProperty(name)) {
        return name;
    }
    for (var i = 2; ;i++) {
        var browserName = [ name, i ].join(" ");
        if (!this.browsers.hasOwnProperty(browserName)) {
            return browserName;
        }
    }
};

Config.prototype._getBrowsers = function(callback) {
    var config = this, browsers = config.browsers || {};
    InstalledBrowsers.get(function(installedBrowsers) {
        Object.keys(installedBrowsers).forEach(function(installedBrowserName) {
            var installedBrowserPath = installedBrowsers[installedBrowserName].path, alreadyExists = false;
            Object.keys(browsers).forEach(function(browserName) {
                if (browsers[browserName].path === installedBrowserPath) {
                    alreadyExists = true;
                }
            });
            if (!alreadyExists) {
                browsers[config._getFreeBrowserName(installedBrowserName)] = installedBrowsers[installedBrowserName];
            }
        });
        callback(browsers);
    });
};

Config.prototype._getRegistryKeyValue = function(key, field, callback) {
    var cmd = 'reg query "' + key + '" /v ' + field;
    child_process.exec(cmd, function(err, stdout) {
        try {
            var fieldString = stdout.split("\r\n")[2].split("   ")[3];
            callback(fieldString);
        } catch (e) {
            callback(null);
        }
    });
};

Config.prototype.getInstallDate = function(callback) {
    var config = this;
    if (!WIN_PLATFORM) {
        callback(0);
        return;
    }
    this.getCurrentVersion(function(version) {
        var versionArr = version.split("."), fieldName = [ versionArr[0], versionArr[1] ].join("");
        config._getRegistryKeyValue(LICENSE_DATE_REG_KEY, fieldName, function(license) {
            if (license) {
                callback(0);
            } else {
                config._getRegistryKeyValue(INSTALL_DATE_REG_KEY, fieldName, function(val) {
                    if (!val) {
                        callback(0);
                    } else {
                        val = parseInt(val, 16);
                        var day = Math.floor(val / 1234), month = Math.floor(val % 1234 / 38) - 1, year = Math.floor(2e3 + val % 1234 % 38);
                        val = new Date(year, month, day).getTime();
                        callback(val);
                    }
                });
            }
        });
    });
};