var fs = require("fs"), javascriptParser = require("uglify-js").parser, astProcessor = require("uglify-js").uglify, Common = require("./common"), Hammerhead = require("./../../hammerhead"), Ast = require("./ast"), FuncCallValidator = require("./func_call_validator"), StepsFacilityAnalyzer = require("./steps_facility_analyzer"), ErrCodes = require("./err_codes");

function extractMixins(ast, descriptor, errs) {
    var walker = astProcessor.ast_walker(), mixinAnalyzer = new StepsFacilityAnalyzer(true, descriptor.rawMixinsStepData, errs), requireCodeRebuild = false;
    walker.with_walkers({
        string: function() {
            var astPath = walker.stack(), topStatement = astPath[1][0];
            if (this[1] === Common.MIXIN_DECLARATION_MARKER) {
                topStatement.remove = true;
                requireCodeRebuild = true;
                mixinAnalyzer.run(walker.stack(), descriptor.filename);
            }
        }
    }, function() {
        walker.walk(ast);
    });
    return requireCodeRebuild;
}

function analyzeRequireCode(descriptor, errs) {
    var requireAst = null;
    try {
        requireAst = javascriptParser.parse(descriptor.jsCode, false, true);
    } catch (parserErr) {
        errs.push({
            code: ErrCodes.JAVASCRIPT_PARSING_FAILED,
            filename: descriptor.filename,
            parserErr: parserErr
        });
    }
    if (requireAst) {
        if (Common.MIXIN_FOOTPRINT_REGEXP.test(descriptor.jsCode)) {
            var rebuildCode = extractMixins(requireAst, descriptor, errs);
            if (rebuildCode) {
                requireAst = Ast.getRemainderAst(requireAst);
                descriptor.jsCode = requireAst ? astProcessor.gen_code(requireAst, {
                    beautify: true
                }) : "";
            }
        }
        if (Common.ACTION_FUNC_FOOTPRINT_REGEXP.test(descriptor.jsCode)) {
            FuncCallValidator.validate(requireAst, descriptor.filename, errs, true);
        }
    }
}

exports.build = function(requireFilename, ownerFilename, callback) {
    fs.readFile(requireFilename, function(readErr, data) {
        var errs = [], descriptor = {
            hasErrs: false,
            jsCode: null,
            rawMixinsStepData: {},
            filename: requireFilename
        };
        if (readErr) {
            errs.push({
                code: ErrCodes.FAILED_LOAD_REQUIRE,
                filename: ownerFilename,
                req: requireFilename
            });
        } else {
            descriptor.jsCode = data.toString().trim();
            analyzeRequireCode(descriptor, errs);
        }
        if (!errs.length) {
            descriptor.jsCode = Hammerhead.wrapDomAccessors(descriptor.jsCode, true);
        }
        if (descriptor.jsCode && descriptor.jsCode[descriptor.jsCode.length - 1] !== ";") {
            descriptor.jsCode += ";";
        }
        descriptor.hasErrs = !!errs.length;
        callback(errs, descriptor);
    });
};