(function() {
    var JSParsingTools = {};
    var acornExports = {};
    (function(exports) {
        (function(root, mod) {
            if (typeof exports == "object" && typeof module == "object") {
                return mod(exports);
            }
            if (typeof define == "function" && define.amd) {
                return define([ "exports" ], mod);
            }
            mod(root.acorn || (root.acorn = {}));
        })(this, function(exports) {
            "use strict";
            exports.version = "0.4.1";
            var options, input, inputLen, sourceFile;
            exports.parse = function(inpt, opts) {
                input = String(inpt);
                inputLen = input.length;
                setOptions(opts);
                initTokenState();
                return parseTopLevel(options.program);
            };
            var defaultOptions = exports.defaultOptions = {
                ecmaVersion: 5,
                strictSemicolons: false,
                allowTrailingCommas: true,
                forbidReserved: false,
                locations: false,
                onComment: null,
                ranges: false,
                program: null,
                sourceFile: null,
                directSourceFile: null
            };
            function setOptions(opts) {
                options = opts || {};
                for (var opt in defaultOptions) {
                    if (!Object.prototype.hasOwnProperty.call(options, opt)) {
                        options[opt] = defaultOptions[opt];
                    }
                }
                sourceFile = options.sourceFile || null;
            }
            var getLineInfo = exports.getLineInfo = function(input, offset) {
                for (var line = 1, cur = 0; ;) {
                    lineBreak.lastIndex = cur;
                    var match = lineBreak.exec(input);
                    if (match && match.index < offset) {
                        ++line;
                        cur = match.index + match[0].length;
                    } else {
                        break;
                    }
                }
                return {
                    line: line,
                    column: offset - cur
                };
            };
            exports.tokenize = function(inpt, opts) {
                input = String(inpt);
                inputLen = input.length;
                setOptions(opts);
                initTokenState();
                var t = {};
                function getToken(forceRegexp) {
                    readToken(forceRegexp);
                    t.start = tokStart;
                    t.end = tokEnd;
                    t.startLoc = tokStartLoc;
                    t.endLoc = tokEndLoc;
                    t.type = tokType;
                    t.value = tokVal;
                    return t;
                }
                getToken.jumpTo = function(pos, reAllowed) {
                    tokPos = pos;
                    if (options.locations) {
                        tokCurLine = 1;
                        tokLineStart = lineBreak.lastIndex = 0;
                        var match;
                        while ((match = lineBreak.exec(input)) && match.index < pos) {
                            ++tokCurLine;
                            tokLineStart = match.index + match[0].length;
                        }
                    }
                    tokRegexpAllowed = reAllowed;
                    skipSpace();
                };
                return getToken;
            };
            var tokPos;
            var tokStart, tokEnd;
            var tokStartLoc, tokEndLoc;
            var tokType, tokVal;
            var tokRegexpAllowed;
            var tokCurLine, tokLineStart;
            var lastStart, lastEnd, lastEndLoc;
            var inFunction, labels, strict;
            function raise(pos, message) {
                var loc = getLineInfo(input, pos);
                message += " (" + loc.line + ":" + loc.column + ")";
                var err = new SyntaxError(message);
                err.pos = pos;
                err.loc = loc;
                err.raisedAt = tokPos;
                throw err;
            }
            var empty = [];
            var _num = {
                type: "num"
            }, _regexp = {
                type: "regexp"
            }, _string = {
                type: "string"
            };
            var _name = {
                type: "name"
            }, _eof = {
                type: "eof"
            };
            var _break = {
                keyword: "break"
            }, _case = {
                keyword: "case",
                beforeExpr: true
            }, _catch = {
                keyword: "catch"
            };
            var _continue = {
                keyword: "continue"
            }, _debugger = {
                keyword: "debugger"
            }, _default = {
                keyword: "default"
            };
            var _do = {
                keyword: "do",
                isLoop: true
            }, _else = {
                keyword: "else",
                beforeExpr: true
            };
            var _finally = {
                keyword: "finally"
            }, _for = {
                keyword: "for",
                isLoop: true
            }, _function = {
                keyword: "function"
            };
            var _if = {
                keyword: "if"
            }, _return = {
                keyword: "return",
                beforeExpr: true
            }, _switch = {
                keyword: "switch"
            };
            var _throw = {
                keyword: "throw",
                beforeExpr: true
            }, _try = {
                keyword: "try"
            }, _var = {
                keyword: "var"
            };
            var _while = {
                keyword: "while",
                isLoop: true
            }, _with = {
                keyword: "with"
            }, _new = {
                keyword: "new",
                beforeExpr: true
            };
            var _this = {
                keyword: "this"
            };
            var _null = {
                keyword: "null",
                atomValue: null
            }, _true = {
                keyword: "true",
                atomValue: true
            };
            var _false = {
                keyword: "false",
                atomValue: false
            };
            var _in = {
                keyword: "in",
                binop: 7,
                beforeExpr: true
            };
            var keywordTypes = {
                "break": _break,
                "case": _case,
                "catch": _catch,
                "continue": _continue,
                "debugger": _debugger,
                "default": _default,
                "do": _do,
                "else": _else,
                "finally": _finally,
                "for": _for,
                "function": _function,
                "if": _if,
                "return": _return,
                "switch": _switch,
                "throw": _throw,
                "try": _try,
                "var": _var,
                "while": _while,
                "with": _with,
                "null": _null,
                "true": _true,
                "false": _false,
                "new": _new,
                "in": _in,
                "instanceof": {
                    keyword: "instanceof",
                    binop: 7,
                    beforeExpr: true
                },
                "this": _this,
                "typeof": {
                    keyword: "typeof",
                    prefix: true,
                    beforeExpr: true
                },
                "void": {
                    keyword: "void",
                    prefix: true,
                    beforeExpr: true
                },
                "delete": {
                    keyword: "delete",
                    prefix: true,
                    beforeExpr: true
                }
            };
            var _bracketL = {
                type: "[",
                beforeExpr: true
            }, _bracketR = {
                type: "]"
            }, _braceL = {
                type: "{",
                beforeExpr: true
            };
            var _braceR = {
                type: "}"
            }, _parenL = {
                type: "(",
                beforeExpr: true
            }, _parenR = {
                type: ")"
            };
            var _comma = {
                type: ",",
                beforeExpr: true
            }, _semi = {
                type: ";",
                beforeExpr: true
            };
            var _colon = {
                type: ":",
                beforeExpr: true
            }, _dot = {
                type: "."
            }, _question = {
                type: "?",
                beforeExpr: true
            };
            var _slash = {
                binop: 10,
                beforeExpr: true
            }, _eq = {
                isAssign: true,
                beforeExpr: true
            };
            var _assign = {
                isAssign: true,
                beforeExpr: true
            };
            var _incDec = {
                postfix: true,
                prefix: true,
                isUpdate: true
            }, _prefix = {
                prefix: true,
                beforeExpr: true
            };
            var _logicalOR = {
                binop: 1,
                beforeExpr: true
            };
            var _logicalAND = {
                binop: 2,
                beforeExpr: true
            };
            var _bitwiseOR = {
                binop: 3,
                beforeExpr: true
            };
            var _bitwiseXOR = {
                binop: 4,
                beforeExpr: true
            };
            var _bitwiseAND = {
                binop: 5,
                beforeExpr: true
            };
            var _equality = {
                binop: 6,
                beforeExpr: true
            };
            var _relational = {
                binop: 7,
                beforeExpr: true
            };
            var _bitShift = {
                binop: 8,
                beforeExpr: true
            };
            var _plusMin = {
                binop: 9,
                prefix: true,
                beforeExpr: true
            };
            var _multiplyModulo = {
                binop: 10,
                beforeExpr: true
            };
            exports.tokTypes = {
                bracketL: _bracketL,
                bracketR: _bracketR,
                braceL: _braceL,
                braceR: _braceR,
                parenL: _parenL,
                parenR: _parenR,
                comma: _comma,
                semi: _semi,
                colon: _colon,
                dot: _dot,
                question: _question,
                slash: _slash,
                eq: _eq,
                name: _name,
                eof: _eof,
                num: _num,
                regexp: _regexp,
                string: _string
            };
            for (var kw in keywordTypes) {
                exports.tokTypes["_" + kw] = keywordTypes[kw];
            }
            function makePredicate(words) {
                words = words.split(" ");
                var f = "", cats = [];
                out: for (var i = 0; i < words.length; ++i) {
                    for (var j = 0; j < cats.length; ++j) {
                        if (cats[j][0].length == words[i].length) {
                            cats[j].push(words[i]);
                            continue out;
                        }
                    }
                    cats.push([ words[i] ]);
                }
                function compareTo(arr) {
                    if (arr.length == 1) {
                        return f += "return str === " + JSON.stringify(arr[0]) + ";";
                    }
                    f += "switch(str){";
                    for (var i = 0; i < arr.length; ++i) {
                        f += "case " + JSON.stringify(arr[i]) + ":";
                    }
                    f += "return true}return false;";
                }
                if (cats.length > 3) {
                    cats.sort(function(a, b) {
                        return b.length - a.length;
                    });
                    f += "switch(str.length){";
                    for (var i = 0; i < cats.length; ++i) {
                        var cat = cats[i];
                        f += "case " + cat[0].length + ":";
                        compareTo(cat);
                    }
                    f += "}";
                } else {
                    compareTo(words);
                }
                return new Function("str", f);
            }
            var isReservedWord3 = makePredicate("abstract boolean byte char class double enum export extends final float goto implements import int interface long native package private protected public short static super synchronized throws transient volatile");
            var isReservedWord5 = makePredicate("class enum extends super const export import");
            var isStrictReservedWord = makePredicate("implements interface let package private protected public static yield");
            var isStrictBadIdWord = makePredicate("eval arguments");
            var isKeyword = makePredicate("break case catch continue debugger default do else finally for function if return switch throw try var while with null true false instanceof typeof void delete new in this");
            var nonASCIIwhitespace = /[\u1680\u180e\u2000-\u200a\u202f\u205f\u3000\ufeff]/;
            var nonASCIIidentifierStartChars = "ªµºÀ-ÖØ-öø-ˁˆ-ˑˠ-ˤˬˮͰ-ʹͶͷͺ-ͽΆΈ-ΊΌΎ-ΡΣ-ϵϷ-ҁҊ-ԧԱ-Ֆՙա-ևא-תװ-ײؠ-يٮٯٱ-ۓەۥۦۮۯۺ-ۼۿܐܒ-ܯݍ-ޥޱߊ-ߪߴߵߺࠀ-ࠕࠚࠤࠨࡀ-ࡘࢠࢢ-ࢬऄ-हऽॐक़-ॡॱ-ॷॹ-ॿঅ-ঌএঐও-নপ-রলশ-হঽৎড়ঢ়য়-ৡৰৱਅ-ਊਏਐਓ-ਨਪ-ਰਲਲ਼ਵਸ਼ਸਹਖ਼-ੜਫ਼ੲ-ੴઅ-ઍએ-ઑઓ-નપ-રલળવ-હઽૐૠૡଅ-ଌଏଐଓ-ନପ-ରଲଳଵ-ହଽଡ଼ଢ଼ୟ-ୡୱஃஅ-ஊஎ-ஐஒ-கஙசஜஞடணதந-பம-ஹௐఅ-ఌఎ-ఐఒ-నప-ళవ-హఽౘౙౠౡಅ-ಌಎ-ಐಒ-ನಪ-ಳವ-ಹಽೞೠೡೱೲഅ-ഌഎ-ഐഒ-ഺഽൎൠൡൺ-ൿඅ-ඖක-නඳ-රලව-ෆก-ะาำเ-ๆກຂຄງຈຊຍດ-ທນ-ຟມ-ຣລວສຫອ-ະາຳຽເ-ໄໆໜ-ໟༀཀ-ཇཉ-ཬྈ-ྌက-ဪဿၐ-ၕၚ-ၝၡၥၦၮ-ၰၵ-ႁႎႠ-ჅჇჍა-ჺჼ-ቈቊ-ቍቐ-ቖቘቚ-ቝበ-ኈኊ-ኍነ-ኰኲ-ኵኸ-ኾዀዂ-ዅወ-ዖዘ-ጐጒ-ጕጘ-ፚᎀ-ᎏᎠ-Ᏼᐁ-ᙬᙯ-ᙿᚁ-ᚚᚠ-ᛪᛮ-ᛰᜀ-ᜌᜎ-ᜑᜠ-ᜱᝀ-ᝑᝠ-ᝬᝮ-ᝰក-ឳៗៜᠠ-ᡷᢀ-ᢨᢪᢰ-ᣵᤀ-ᤜᥐ-ᥭᥰ-ᥴᦀ-ᦫᧁ-ᧇᨀ-ᨖᨠ-ᩔᪧᬅ-ᬳᭅ-ᭋᮃ-ᮠᮮᮯᮺ-ᯥᰀ-ᰣᱍ-ᱏᱚ-ᱽᳩ-ᳬᳮ-ᳱᳵᳶᴀ-ᶿḀ-ἕἘ-Ἕἠ-ὅὈ-Ὅὐ-ὗὙὛὝὟ-ώᾀ-ᾴᾶ-ᾼιῂ-ῄῆ-ῌῐ-ΐῖ-Ίῠ-Ῥῲ-ῴῶ-ῼⁱⁿₐ-ₜℂℇℊ-ℓℕℙ-ℝℤΩℨK-ℭℯ-ℹℼ-ℿⅅ-ⅉⅎⅠ-ↈⰀ-Ⱞⰰ-ⱞⱠ-ⳤⳫ-ⳮⳲⳳⴀ-ⴥⴧⴭⴰ-ⵧⵯⶀ-ⶖⶠ-ⶦⶨ-ⶮⶰ-ⶶⶸ-ⶾⷀ-ⷆⷈ-ⷎⷐ-ⷖⷘ-ⷞⸯ々-〇〡-〩〱-〵〸-〼ぁ-ゖゝ-ゟァ-ヺー-ヿㄅ-ㄭㄱ-ㆎㆠ-ㆺㇰ-ㇿ㐀-䶵一-鿌ꀀ-ꒌꓐ-ꓽꔀ-ꘌꘐ-ꘟꘪꘫꙀ-ꙮꙿ-ꚗꚠ-ꛯꜗ-ꜟꜢ-ꞈꞋ-ꞎꞐ-ꞓꞠ-Ɦꟸ-ꠁꠃ-ꠅꠇ-ꠊꠌ-ꠢꡀ-ꡳꢂ-ꢳꣲ-ꣷꣻꤊ-ꤥꤰ-ꥆꥠ-ꥼꦄ-ꦲꧏꨀ-ꨨꩀ-ꩂꩄ-ꩋꩠ-ꩶꩺꪀ-ꪯꪱꪵꪶꪹ-ꪽꫀꫂꫛ-ꫝꫠ-ꫪꫲ-ꫴꬁ-ꬆꬉ-ꬎꬑ-ꬖꬠ-ꬦꬨ-ꬮꯀ-ꯢ가-힣ힰ-ퟆퟋ-ퟻ豈-舘並-龎ﬀ-ﬆﬓ-ﬗיִײַ-ﬨשׁ-זּטּ-לּמּנּסּףּפּצּ-ﮱﯓ-ﴽﵐ-ﶏﶒ-ﷇﷰ-ﷻﹰ-ﹴﹶ-ﻼＡ-Ｚａ-ｚｦ-ﾾￂ-ￇￊ-ￏￒ-ￗￚ-ￜ";
            var nonASCIIidentifierChars = "̀-ͯ҃-֑҇-ׇֽֿׁׂׅׄؐ-ؚؠ-ىٲ-ۓۧ-ۨۻ-ۼܰ-݊ࠀ-ࠔࠛ-ࠣࠥ-ࠧࠩ-࠭ࡀ-ࡗࣤ-ࣾऀ-ःऺ-़ा-ॏ॑-ॗॢ-ॣ०-९ঁ-ঃ়া-ৄেৈৗয়-ৠਁ-ਃ਼ਾ-ੂੇੈੋ-੍ੑ੦-ੱੵઁ-ઃ઼ા-ૅે-ૉો-્ૢ-ૣ૦-૯ଁ-ଃ଼ା-ୄେୈୋ-୍ୖୗୟ-ୠ୦-୯ஂா-ூெ-ைொ-்ௗ௦-௯ఁ-ఃె-ైొ-్ౕౖౢ-ౣ౦-౯ಂಃ಼ಾ-ೄೆ-ೈೊ-್ೕೖೢ-ೣ೦-೯ംഃെ-ൈൗൢ-ൣ൦-൯ංඃ්ා-ුූෘ-ෟෲෳิ-ฺเ-ๅ๐-๙ິ-ູ່-ໍ໐-໙༘༙༠-༩༹༵༷ཁ-ཇཱ-྄྆-྇ྍ-ྗྙ-ྼ࿆က-ဩ၀-၉ၧ-ၭၱ-ၴႂ-ႍႏ-ႝ፝-፟ᜎ-ᜐᜠ-ᜰᝀ-ᝐᝲᝳក-ឲ៝០-៩᠋-᠍᠐-᠙ᤠ-ᤫᤰ-᤻ᥑ-ᥭᦰ-ᧀᧈ-ᧉ᧐-᧙ᨀ-ᨕᨠ-ᩓ᩠-᩿᩼-᪉᪐-᪙ᭆ-ᭋ᭐-᭙᭫-᭳᮰-᮹᯦-᯳ᰀ-ᰢ᱀-᱉ᱛ-ᱽ᳐-᳒ᴀ-ᶾḁ-ἕ‌‍‿⁀⁔⃐-⃥⃜⃡-⃰ⶁ-ⶖⷠ-ⷿ〡-〨゙゚Ꙁ-ꙭꙴ-꙽ꚟ꛰-꛱ꟸ-ꠀ꠆ꠋꠣ-ꠧꢀ-ꢁꢴ-꣄꣐-꣙ꣳ-ꣷ꤀-꤉ꤦ-꤭ꤰ-ꥅꦀ-ꦃ꦳-꧀ꨀ-ꨧꩀ-ꩁꩌ-ꩍ꩐-꩙ꩻꫠ-ꫩꫲ-ꫳꯀ-ꯡ꯬꯭꯰-꯹ﬠ-ﬨ︀-️︠-︦︳︴﹍-﹏０-９＿";
            var nonASCIIidentifierStart = new RegExp("[" + nonASCIIidentifierStartChars + "]");
            var nonASCIIidentifier = new RegExp("[" + nonASCIIidentifierStartChars + nonASCIIidentifierChars + "]");
            var newline = /[\n\r\u2028\u2029]/;
            var lineBreak = /\r\n|[\n\r\u2028\u2029]/g;
            var isIdentifierStart = exports.isIdentifierStart = function(code) {
                if (code < 65) {
                    return code === 36;
                }
                if (code < 91) {
                    return true;
                }
                if (code < 97) {
                    return code === 95;
                }
                if (code < 123) {
                    return true;
                }
                return code >= 170 && nonASCIIidentifierStart.test(String.fromCharCode(code));
            };
            var isIdentifierChar = exports.isIdentifierChar = function(code) {
                if (code < 48) {
                    return code === 36;
                }
                if (code < 58) {
                    return true;
                }
                if (code < 65) {
                    return false;
                }
                if (code < 91) {
                    return true;
                }
                if (code < 97) {
                    return code === 95;
                }
                if (code < 123) {
                    return true;
                }
                return code >= 170 && nonASCIIidentifier.test(String.fromCharCode(code));
            };
            function line_loc_t() {
                this.line = tokCurLine;
                this.column = tokPos - tokLineStart;
            }
            function initTokenState() {
                tokCurLine = 1;
                tokPos = tokLineStart = 0;
                tokRegexpAllowed = true;
                skipSpace();
            }
            function finishToken(type, val) {
                tokEnd = tokPos;
                if (options.locations) {
                    tokEndLoc = new line_loc_t();
                }
                tokType = type;
                skipSpace();
                tokVal = val;
                tokRegexpAllowed = type.beforeExpr;
            }
            function skipBlockComment() {
                var startLoc = options.onComment && options.locations && new line_loc_t();
                var start = tokPos, end = input.indexOf("*/", tokPos += 2);
                if (end === -1) {
                    raise(tokPos - 2, "Unterminated comment");
                }
                tokPos = end + 2;
                if (options.locations) {
                    lineBreak.lastIndex = start;
                    var match;
                    while ((match = lineBreak.exec(input)) && match.index < tokPos) {
                        ++tokCurLine;
                        tokLineStart = match.index + match[0].length;
                    }
                }
                if (options.onComment) {
                    options.onComment(true, input.slice(start + 2, end), start, tokPos, startLoc, options.locations && new line_loc_t());
                }
            }
            function skipLineComment() {
                var start = tokPos;
                var startLoc = options.onComment && options.locations && new line_loc_t();
                var ch = input.charCodeAt(tokPos += 2);
                while (tokPos < inputLen && ch !== 10 && ch !== 13 && ch !== 8232 && ch !== 8233) {
                    ++tokPos;
                    ch = input.charCodeAt(tokPos);
                }
                if (options.onComment) {
                    options.onComment(false, input.slice(start + 2, tokPos), start, tokPos, startLoc, options.locations && new line_loc_t());
                }
            }
            function skipSpace() {
                while (tokPos < inputLen) {
                    var ch = input.charCodeAt(tokPos);
                    if (ch === 32) {
                        ++tokPos;
                    } else {
                        if (ch === 13) {
                            ++tokPos;
                            var next = input.charCodeAt(tokPos);
                            if (next === 10) {
                                ++tokPos;
                            }
                            if (options.locations) {
                                ++tokCurLine;
                                tokLineStart = tokPos;
                            }
                        } else {
                            if (ch === 10 || ch === 8232 || ch === 8233) {
                                ++tokPos;
                                if (options.locations) {
                                    ++tokCurLine;
                                    tokLineStart = tokPos;
                                }
                            } else {
                                if (ch > 8 && ch < 14) {
                                    ++tokPos;
                                } else {
                                    if (ch === 47) {
                                        var next = input.charCodeAt(tokPos + 1);
                                        if (next === 42) {
                                            skipBlockComment();
                                        } else {
                                            if (next === 47) {
                                                skipLineComment();
                                            } else {
                                                break;
                                            }
                                        }
                                    } else {
                                        if (ch === 160) {
                                            ++tokPos;
                                        } else {
                                            if (ch >= 5760 && nonASCIIwhitespace.test(String.fromCharCode(ch))) {
                                                ++tokPos;
                                            } else {
                                                break;
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
            function readToken_dot() {
                var next = input.charCodeAt(tokPos + 1);
                if (next >= 48 && next <= 57) {
                    return readNumber(true);
                }
                ++tokPos;
                return finishToken(_dot);
            }
            function readToken_slash() {
                var next = input.charCodeAt(tokPos + 1);
                if (tokRegexpAllowed) {
                    ++tokPos;
                    return readRegexp();
                }
                if (next === 61) {
                    return finishOp(_assign, 2);
                }
                return finishOp(_slash, 1);
            }
            function readToken_mult_modulo() {
                var next = input.charCodeAt(tokPos + 1);
                if (next === 61) {
                    return finishOp(_assign, 2);
                }
                return finishOp(_multiplyModulo, 1);
            }
            function readToken_pipe_amp(code) {
                var next = input.charCodeAt(tokPos + 1);
                if (next === code) {
                    return finishOp(code === 124 ? _logicalOR : _logicalAND, 2);
                }
                if (next === 61) {
                    return finishOp(_assign, 2);
                }
                return finishOp(code === 124 ? _bitwiseOR : _bitwiseAND, 1);
            }
            function readToken_caret() {
                var next = input.charCodeAt(tokPos + 1);
                if (next === 61) {
                    return finishOp(_assign, 2);
                }
                return finishOp(_bitwiseXOR, 1);
            }
            function readToken_plus_min(code) {
                var next = input.charCodeAt(tokPos + 1);
                if (next === code) {
                    if (next == 45 && input.charCodeAt(tokPos + 2) == 62 && newline.test(input.slice(lastEnd, tokPos))) {
                        tokPos += 3;
                        skipLineComment();
                        skipSpace();
                        return readToken();
                    }
                    return finishOp(_incDec, 2);
                }
                if (next === 61) {
                    return finishOp(_assign, 2);
                }
                return finishOp(_plusMin, 1);
            }
            function readToken_lt_gt(code) {
                var next = input.charCodeAt(tokPos + 1);
                var size = 1;
                if (next === code) {
                    size = code === 62 && input.charCodeAt(tokPos + 2) === 62 ? 3 : 2;
                    if (input.charCodeAt(tokPos + size) === 61) {
                        return finishOp(_assign, size + 1);
                    }
                    return finishOp(_bitShift, size);
                }
                if (next == 33 && code == 60 && input.charCodeAt(tokPos + 2) == 45 && input.charCodeAt(tokPos + 3) == 45) {
                    tokPos += 4;
                    skipLineComment();
                    skipSpace();
                    return readToken();
                }
                if (next === 61) {
                    size = input.charCodeAt(tokPos + 2) === 61 ? 3 : 2;
                }
                return finishOp(_relational, size);
            }
            function readToken_eq_excl(code) {
                var next = input.charCodeAt(tokPos + 1);
                if (next === 61) {
                    return finishOp(_equality, input.charCodeAt(tokPos + 2) === 61 ? 3 : 2);
                }
                return finishOp(code === 61 ? _eq : _prefix, 1);
            }
            function getTokenFromCode(code) {
                switch (code) {
                  case 46:
                    return readToken_dot();

                  case 40:
                    ++tokPos;
                    return finishToken(_parenL);

                  case 41:
                    ++tokPos;
                    return finishToken(_parenR);

                  case 59:
                    ++tokPos;
                    return finishToken(_semi);

                  case 44:
                    ++tokPos;
                    return finishToken(_comma);

                  case 91:
                    ++tokPos;
                    return finishToken(_bracketL);

                  case 93:
                    ++tokPos;
                    return finishToken(_bracketR);

                  case 123:
                    ++tokPos;
                    return finishToken(_braceL);

                  case 125:
                    ++tokPos;
                    return finishToken(_braceR);

                  case 58:
                    ++tokPos;
                    return finishToken(_colon);

                  case 63:
                    ++tokPos;
                    return finishToken(_question);

                  case 48:
                    var next = input.charCodeAt(tokPos + 1);
                    if (next === 120 || next === 88) {
                        return readHexNumber();
                    }

                  case 49:
                  case 50:
                  case 51:
                  case 52:
                  case 53:
                  case 54:
                  case 55:
                  case 56:
                  case 57:
                    return readNumber(false);

                  case 34:
                  case 39:
                    return readString(code);

                  case 47:
                    return readToken_slash(code);

                  case 37:
                  case 42:
                    return readToken_mult_modulo();

                  case 124:
                  case 38:
                    return readToken_pipe_amp(code);

                  case 94:
                    return readToken_caret();

                  case 43:
                  case 45:
                    return readToken_plus_min(code);

                  case 60:
                  case 62:
                    return readToken_lt_gt(code);

                  case 61:
                  case 33:
                    return readToken_eq_excl(code);

                  case 126:
                    return finishOp(_prefix, 1);
                }
                return false;
            }
            function readToken(forceRegexp) {
                if (!forceRegexp) {
                    tokStart = tokPos;
                } else {
                    tokPos = tokStart + 1;
                }
                if (options.locations) {
                    tokStartLoc = new line_loc_t();
                }
                if (forceRegexp) {
                    return readRegexp();
                }
                if (tokPos >= inputLen) {
                    return finishToken(_eof);
                }
                var code = input.charCodeAt(tokPos);
                if (isIdentifierStart(code) || code === 92) {
                    return readWord();
                }
                var tok = getTokenFromCode(code);
                if (tok === false) {
                    var ch = String.fromCharCode(code);
                    if (ch === "\\" || nonASCIIidentifierStart.test(ch)) {
                        return readWord();
                    }
                    raise(tokPos, "Unexpected character '" + ch + "'");
                }
                return tok;
            }
            function finishOp(type, size) {
                var str = input.slice(tokPos, tokPos + size);
                tokPos += size;
                finishToken(type, str);
            }
            function readRegexp() {
                var content = "", escaped, inClass, start = tokPos;
                for (;;) {
                    if (tokPos >= inputLen) {
                        raise(start, "Unterminated regular expression");
                    }
                    var ch = input.charAt(tokPos);
                    if (newline.test(ch)) {
                        raise(start, "Unterminated regular expression");
                    }
                    if (!escaped) {
                        if (ch === "[") {
                            inClass = true;
                        } else {
                            if (ch === "]" && inClass) {
                                inClass = false;
                            } else {
                                if (ch === "/" && !inClass) {
                                    break;
                                }
                            }
                        }
                        escaped = ch === "\\";
                    } else {
                        escaped = false;
                    }
                    ++tokPos;
                }
                var content = input.slice(start, tokPos);
                ++tokPos;
                var mods = readWord1();
                if (mods && !/^[gmsiy]*$/.test(mods)) {
                    raise(start, "Invalid regexp flag");
                }
                return finishToken(_regexp, new RegExp(content, mods));
            }
            function readInt(radix, len) {
                var start = tokPos, total = 0;
                for (var i = 0, e = len == null ? Infinity : len; i < e; ++i) {
                    var code = input.charCodeAt(tokPos), val;
                    if (code >= 97) {
                        val = code - 97 + 10;
                    } else {
                        if (code >= 65) {
                            val = code - 65 + 10;
                        } else {
                            if (code >= 48 && code <= 57) {
                                val = code - 48;
                            } else {
                                val = Infinity;
                            }
                        }
                    }
                    if (val >= radix) {
                        break;
                    }
                    ++tokPos;
                    total = total * radix + val;
                }
                if (tokPos === start || len != null && tokPos - start !== len) {
                    return null;
                }
                return total;
            }
            function readHexNumber() {
                tokPos += 2;
                var val = readInt(16);
                if (val == null) {
                    raise(tokStart + 2, "Expected hexadecimal number");
                }
                if (isIdentifierStart(input.charCodeAt(tokPos))) {
                    raise(tokPos, "Identifier directly after number");
                }
                return finishToken(_num, val);
            }
            function readNumber(startsWithDot) {
                var start = tokPos, isFloat = false, octal = input.charCodeAt(tokPos) === 48;
                if (!startsWithDot && readInt(10) === null) {
                    raise(start, "Invalid number");
                }
                if (input.charCodeAt(tokPos) === 46) {
                    ++tokPos;
                    readInt(10);
                    isFloat = true;
                }
                var next = input.charCodeAt(tokPos);
                if (next === 69 || next === 101) {
                    next = input.charCodeAt(++tokPos);
                    if (next === 43 || next === 45) {
                        ++tokPos;
                    }
                    if (readInt(10) === null) {
                        raise(start, "Invalid number");
                    }
                    isFloat = true;
                }
                if (isIdentifierStart(input.charCodeAt(tokPos))) {
                    raise(tokPos, "Identifier directly after number");
                }
                var str = input.slice(start, tokPos), val;
                if (isFloat) {
                    val = parseFloat(str);
                } else {
                    if (!octal || str.length === 1) {
                        val = parseInt(str, 10);
                    } else {
                        if (/[89]/.test(str) || strict) {
                            raise(start, "Invalid number");
                        } else {
                            val = parseInt(str, 8);
                        }
                    }
                }
                return finishToken(_num, val);
            }
            function readString(quote) {
                tokPos++;
                var out = "";
                for (;;) {
                    if (tokPos >= inputLen) {
                        raise(tokStart, "Unterminated string constant");
                    }
                    var ch = input.charCodeAt(tokPos);
                    if (ch === quote) {
                        ++tokPos;
                        return finishToken(_string, out);
                    }
                    if (ch === 92) {
                        ch = input.charCodeAt(++tokPos);
                        var octal = /^[0-7]+/.exec(input.slice(tokPos, tokPos + 3));
                        if (octal) {
                            octal = octal[0];
                        }
                        while (octal && parseInt(octal, 8) > 255) {
                            octal = octal.slice(0, -1);
                        }
                        if (octal === "0") {
                            octal = null;
                        }
                        ++tokPos;
                        if (octal) {
                            if (strict) {
                                raise(tokPos - 2, "Octal literal in strict mode");
                            }
                            out += String.fromCharCode(parseInt(octal, 8));
                            tokPos += octal.length - 1;
                        } else {
                            switch (ch) {
                              case 110:
                                out += "\n";
                                break;

                              case 114:
                                out += "\r";
                                break;

                              case 120:
                                out += String.fromCharCode(readHexChar(2));
                                break;

                              case 117:
                                out += String.fromCharCode(readHexChar(4));
                                break;

                              case 85:
                                out += String.fromCharCode(readHexChar(8));
                                break;

                              case 116:
                                out += "	";
                                break;

                              case 98:
                                out += "\b";
                                break;

                              case 118:
                                out += "";
                                break;

                              case 102:
                                out += "\f";
                                break;

                              case 48:
                                out += "\x00";
                                break;

                              case 13:
                                if (input.charCodeAt(tokPos) === 10) {
                                    ++tokPos;
                                }

                              case 10:
                                if (options.locations) {
                                    tokLineStart = tokPos;
                                    ++tokCurLine;
                                }
                                break;

                              default:
                                out += String.fromCharCode(ch);
                                break;
                            }
                        }
                    } else {
                        if (ch === 13 || ch === 10 || ch === 8232 || ch === 8233) {
                            raise(tokStart, "Unterminated string constant");
                        }
                        out += String.fromCharCode(ch);
                        ++tokPos;
                    }
                }
            }
            function readHexChar(len) {
                var n = readInt(16, len);
                if (n === null) {
                    raise(tokStart, "Bad character escape sequence");
                }
                return n;
            }
            var containsEsc;
            function readWord1() {
                containsEsc = false;
                var word, first = true, start = tokPos;
                for (;;) {
                    var ch = input.charCodeAt(tokPos);
                    if (isIdentifierChar(ch)) {
                        if (containsEsc) {
                            word += input.charAt(tokPos);
                        }
                        ++tokPos;
                    } else {
                        if (ch === 92) {
                            if (!containsEsc) {
                                word = input.slice(start, tokPos);
                            }
                            containsEsc = true;
                            if (input.charCodeAt(++tokPos) != 117) {
                                raise(tokPos, "Expecting Unicode escape sequence \\uXXXX");
                            }
                            ++tokPos;
                            var esc = readHexChar(4);
                            var escStr = String.fromCharCode(esc);
                            if (!escStr) {
                                raise(tokPos - 1, "Invalid Unicode escape");
                            }
                            if (!(first ? isIdentifierStart(esc) : isIdentifierChar(esc))) {
                                raise(tokPos - 4, "Invalid Unicode escape");
                            }
                            word += input.substr(tokPos - 6, 6);
                        } else {
                            break;
                        }
                    }
                    first = false;
                }
                return containsEsc ? word : input.slice(start, tokPos);
            }
            function readWord() {
                var word = readWord1();
                var type = _name;
                if (!containsEsc) {
                    if (isKeyword(word)) {
                        type = keywordTypes[word];
                    } else {
                        if (options.forbidReserved && (options.ecmaVersion === 3 ? isReservedWord3 : isReservedWord5)(word) || strict && isStrictReservedWord(word)) {
                            raise(tokStart, "The keyword '" + word + "' is reserved");
                        }
                    }
                }
                return finishToken(type, word);
            }
            function next() {
                lastStart = tokStart;
                lastEnd = tokEnd;
                lastEndLoc = tokEndLoc;
                readToken();
            }
            function setStrict(strct) {
                strict = strct;
                tokPos = lastEnd;
                if (options.locations) {
                    while (tokPos < tokLineStart) {
                        tokLineStart = input.lastIndexOf("\n", tokLineStart - 2) + 1;
                        --tokCurLine;
                    }
                }
                skipSpace();
                readToken();
            }
            function node_t() {
                this.type = null;
                this.start = tokStart;
                this.end = null;
            }
            function node_loc_t() {
                this.start = tokStartLoc;
                this.end = null;
                if (sourceFile !== null) {
                    this.source = sourceFile;
                }
            }
            function startNode() {
                var node = new node_t();
                if (options.locations) {
                    node.loc = new node_loc_t();
                }
                if (options.directSourceFile) {
                    node.sourceFile = options.directSourceFile;
                }
                if (options.ranges) {
                    node.range = [ tokStart, 0 ];
                }
                return node;
            }
            function startNodeFrom(other) {
                var node = new node_t();
                node.start = other.start;
                if (options.locations) {
                    node.loc = new node_loc_t();
                    node.loc.start = other.loc.start;
                }
                if (options.ranges) {
                    node.range = [ other.range[0], 0 ];
                }
                return node;
            }
            function finishNode(node, type) {
                node.type = type;
                node.end = lastEnd;
                if (options.locations) {
                    node.loc.end = lastEndLoc;
                }
                if (options.ranges) {
                    node.range[1] = lastEnd;
                }
                return node;
            }
            function isUseStrict(stmt) {
                return options.ecmaVersion >= 5 && stmt.type === "ExpressionStatement" && stmt.expression.type === "Literal" && stmt.expression.value === "use strict";
            }
            function eat(type) {
                if (tokType === type) {
                    next();
                    return true;
                }
            }
            function canInsertSemicolon() {
                return !options.strictSemicolons && (tokType === _eof || tokType === _braceR || newline.test(input.slice(lastEnd, tokStart)));
            }
            function semicolon() {
                if (!eat(_semi) && !canInsertSemicolon()) {
                    unexpected();
                }
            }
            function expect(type) {
                if (tokType === type) {
                    next();
                } else {
                    unexpected();
                }
            }
            function unexpected() {
                raise(tokStart, "Unexpected token");
            }
            function checkLVal(expr) {
                if (expr.type !== "Identifier" && expr.type !== "MemberExpression") {
                    raise(expr.start, "Assigning to rvalue");
                }
                if (strict && expr.type === "Identifier" && isStrictBadIdWord(expr.name)) {
                    raise(expr.start, "Assigning to " + expr.name + " in strict mode");
                }
            }
            function parseTopLevel(program) {
                lastStart = lastEnd = tokPos;
                if (options.locations) {
                    lastEndLoc = new line_loc_t();
                }
                inFunction = strict = null;
                labels = [];
                readToken();
                var node = program || startNode(), first = true;
                if (!program) {
                    node.body = [];
                }
                while (tokType !== _eof) {
                    var stmt = parseStatement();
                    node.body.push(stmt);
                    if (first && isUseStrict(stmt)) {
                        setStrict(true);
                    }
                    first = false;
                }
                return finishNode(node, "Program");
            }
            var loopLabel = {
                kind: "loop"
            }, switchLabel = {
                kind: "switch"
            };
            function parseStatement() {
                if (tokType === _slash || tokType === _assign && tokVal == "/=") {
                    readToken(true);
                }
                var starttype = tokType, node = startNode();
                switch (starttype) {
                  case _break:
                  case _continue:
                    next();
                    var isBreak = starttype === _break;
                    if (eat(_semi) || canInsertSemicolon()) {
                        node.label = null;
                    } else {
                        if (tokType !== _name) {
                            unexpected();
                        } else {
                            node.label = parseIdent();
                            semicolon();
                        }
                    }
                    for (var i = 0; i < labels.length; ++i) {
                        var lab = labels[i];
                        if (node.label == null || lab.name === node.label.name) {
                            if (lab.kind != null && (isBreak || lab.kind === "loop")) {
                                break;
                            }
                            if (node.label && isBreak) {
                                break;
                            }
                        }
                    }
                    if (i === labels.length) {
                        raise(node.start, "Unsyntactic " + starttype.keyword);
                    }
                    return finishNode(node, isBreak ? "BreakStatement" : "ContinueStatement");

                  case _debugger:
                    next();
                    semicolon();
                    return finishNode(node, "DebuggerStatement");

                  case _do:
                    next();
                    labels.push(loopLabel);
                    node.body = parseStatement();
                    labels.pop();
                    expect(_while);
                    node.test = parseParenExpression();
                    semicolon();
                    return finishNode(node, "DoWhileStatement");

                  case _for:
                    next();
                    labels.push(loopLabel);
                    expect(_parenL);
                    if (tokType === _semi) {
                        return parseFor(node, null);
                    }
                    if (tokType === _var) {
                        var init = startNode();
                        next();
                        parseVar(init, true);
                        finishNode(init, "VariableDeclaration");
                        if (init.declarations.length === 1 && eat(_in)) {
                            return parseForIn(node, init);
                        }
                        return parseFor(node, init);
                    }
                    var init = parseExpression(false, true);
                    if (eat(_in)) {
                        checkLVal(init);
                        return parseForIn(node, init);
                    }
                    return parseFor(node, init);

                  case _function:
                    next();
                    return parseFunction(node, true);

                  case _if:
                    next();
                    node.test = parseParenExpression();
                    node.consequent = parseStatement();
                    node.alternate = eat(_else) ? parseStatement() : null;
                    return finishNode(node, "IfStatement");

                  case _return:
                    if (!inFunction) {
                        raise(tokStart, "'return' outside of function");
                    }
                    next();
                    if (eat(_semi) || canInsertSemicolon()) {
                        node.argument = null;
                    } else {
                        node.argument = parseExpression();
                        semicolon();
                    }
                    return finishNode(node, "ReturnStatement");

                  case _switch:
                    next();
                    node.discriminant = parseParenExpression();
                    node.cases = [];
                    expect(_braceL);
                    labels.push(switchLabel);
                    for (var cur, sawDefault; tokType != _braceR; ) {
                        if (tokType === _case || tokType === _default) {
                            var isCase = tokType === _case;
                            if (cur) {
                                finishNode(cur, "SwitchCase");
                            }
                            node.cases.push(cur = startNode());
                            cur.consequent = [];
                            next();
                            if (isCase) {
                                cur.test = parseExpression();
                            } else {
                                if (sawDefault) {
                                    raise(lastStart, "Multiple default clauses");
                                }
                                sawDefault = true;
                                cur.test = null;
                            }
                            expect(_colon);
                        } else {
                            if (!cur) {
                                unexpected();
                            }
                            cur.consequent.push(parseStatement());
                        }
                    }
                    if (cur) {
                        finishNode(cur, "SwitchCase");
                    }
                    next();
                    labels.pop();
                    return finishNode(node, "SwitchStatement");

                  case _throw:
                    next();
                    if (newline.test(input.slice(lastEnd, tokStart))) {
                        raise(lastEnd, "Illegal newline after throw");
                    }
                    node.argument = parseExpression();
                    semicolon();
                    return finishNode(node, "ThrowStatement");

                  case _try:
                    next();
                    node.block = parseBlock();
                    node.handler = null;
                    if (tokType === _catch) {
                        var clause = startNode();
                        next();
                        expect(_parenL);
                        clause.param = parseIdent();
                        if (strict && isStrictBadIdWord(clause.param.name)) {
                            raise(clause.param.start, "Binding " + clause.param.name + " in strict mode");
                        }
                        expect(_parenR);
                        clause.guard = null;
                        clause.body = parseBlock();
                        node.handler = finishNode(clause, "CatchClause");
                    }
                    node.guardedHandlers = empty;
                    node.finalizer = eat(_finally) ? parseBlock() : null;
                    if (!node.handler && !node.finalizer) {
                        raise(node.start, "Missing catch or finally clause");
                    }
                    return finishNode(node, "TryStatement");

                  case _var:
                    next();
                    parseVar(node);
                    semicolon();
                    return finishNode(node, "VariableDeclaration");

                  case _while:
                    next();
                    node.test = parseParenExpression();
                    labels.push(loopLabel);
                    node.body = parseStatement();
                    labels.pop();
                    return finishNode(node, "WhileStatement");

                  case _with:
                    if (strict) {
                        raise(tokStart, "'with' in strict mode");
                    }
                    next();
                    node.object = parseParenExpression();
                    node.body = parseStatement();
                    return finishNode(node, "WithStatement");

                  case _braceL:
                    return parseBlock();

                  case _semi:
                    next();
                    return finishNode(node, "EmptyStatement");

                  default:
                    var maybeName = tokVal, expr = parseExpression();
                    if (starttype === _name && expr.type === "Identifier" && eat(_colon)) {
                        for (var i = 0; i < labels.length; ++i) {
                            if (labels[i].name === maybeName) {
                                raise(expr.start, "Label '" + maybeName + "' is already declared");
                            }
                        }
                        var kind = tokType.isLoop ? "loop" : tokType === _switch ? "switch" : null;
                        labels.push({
                            name: maybeName,
                            kind: kind
                        });
                        node.body = parseStatement();
                        labels.pop();
                        node.label = expr;
                        return finishNode(node, "LabeledStatement");
                    } else {
                        node.expression = expr;
                        semicolon();
                        return finishNode(node, "ExpressionStatement");
                    }
                }
            }
            function parseParenExpression() {
                expect(_parenL);
                var val = parseExpression();
                expect(_parenR);
                return val;
            }
            function parseBlock(allowStrict) {
                var node = startNode(), first = true, strict = false, oldStrict;
                node.body = [];
                expect(_braceL);
                while (!eat(_braceR)) {
                    var stmt = parseStatement();
                    node.body.push(stmt);
                    if (first && allowStrict && isUseStrict(stmt)) {
                        oldStrict = strict;
                        setStrict(strict = true);
                    }
                    first = false;
                }
                if (strict && !oldStrict) {
                    setStrict(false);
                }
                return finishNode(node, "BlockStatement");
            }
            function parseFor(node, init) {
                node.init = init;
                expect(_semi);
                node.test = tokType === _semi ? null : parseExpression();
                expect(_semi);
                node.update = tokType === _parenR ? null : parseExpression();
                expect(_parenR);
                node.body = parseStatement();
                labels.pop();
                return finishNode(node, "ForStatement");
            }
            function parseForIn(node, init) {
                node.left = init;
                node.right = parseExpression();
                expect(_parenR);
                node.body = parseStatement();
                labels.pop();
                return finishNode(node, "ForInStatement");
            }
            function parseVar(node, noIn) {
                node.declarations = [];
                node.kind = "var";
                for (;;) {
                    var decl = startNode();
                    decl.id = parseIdent();
                    if (strict && isStrictBadIdWord(decl.id.name)) {
                        raise(decl.id.start, "Binding " + decl.id.name + " in strict mode");
                    }
                    decl.init = eat(_eq) ? parseExpression(true, noIn) : null;
                    node.declarations.push(finishNode(decl, "VariableDeclarator"));
                    if (!eat(_comma)) {
                        break;
                    }
                }
                return node;
            }
            function parseExpression(noComma, noIn) {
                var expr = parseMaybeAssign(noIn);
                if (!noComma && tokType === _comma) {
                    var node = startNodeFrom(expr);
                    node.expressions = [ expr ];
                    while (eat(_comma)) {
                        node.expressions.push(parseMaybeAssign(noIn));
                    }
                    return finishNode(node, "SequenceExpression");
                }
                return expr;
            }
            function parseMaybeAssign(noIn) {
                var left = parseMaybeConditional(noIn);
                if (tokType.isAssign) {
                    var node = startNodeFrom(left);
                    node.operator = tokVal;
                    node.left = left;
                    next();
                    node.right = parseMaybeAssign(noIn);
                    checkLVal(left);
                    return finishNode(node, "AssignmentExpression");
                }
                return left;
            }
            function parseMaybeConditional(noIn) {
                var expr = parseExprOps(noIn);
                if (eat(_question)) {
                    var node = startNodeFrom(expr);
                    node.test = expr;
                    node.consequent = parseExpression(true);
                    expect(_colon);
                    node.alternate = parseExpression(true, noIn);
                    return finishNode(node, "ConditionalExpression");
                }
                return expr;
            }
            function parseExprOps(noIn) {
                return parseExprOp(parseMaybeUnary(), -1, noIn);
            }
            function parseExprOp(left, minPrec, noIn) {
                var prec = tokType.binop;
                if (prec != null && (!noIn || tokType !== _in)) {
                    if (prec > minPrec) {
                        var node = startNodeFrom(left);
                        node.left = left;
                        node.operator = tokVal;
                        var op = tokType;
                        next();
                        node.right = parseExprOp(parseMaybeUnary(), prec, noIn);
                        var exprNode = finishNode(node, op === _logicalOR || op === _logicalAND ? "LogicalExpression" : "BinaryExpression");
                        return parseExprOp(exprNode, minPrec, noIn);
                    }
                }
                return left;
            }
            function parseMaybeUnary() {
                if (tokType.prefix) {
                    var node = startNode(), update = tokType.isUpdate;
                    node.operator = tokVal;
                    node.prefix = true;
                    tokRegexpAllowed = true;
                    next();
                    node.argument = parseMaybeUnary();
                    if (update) {
                        checkLVal(node.argument);
                    } else {
                        if (strict && node.operator === "delete" && node.argument.type === "Identifier") {
                            raise(node.start, "Deleting local variable in strict mode");
                        }
                    }
                    return finishNode(node, update ? "UpdateExpression" : "UnaryExpression");
                }
                var expr = parseExprSubscripts();
                while (tokType.postfix && !canInsertSemicolon()) {
                    var node = startNodeFrom(expr);
                    node.operator = tokVal;
                    node.prefix = false;
                    node.argument = expr;
                    checkLVal(expr);
                    next();
                    expr = finishNode(node, "UpdateExpression");
                }
                return expr;
            }
            function parseExprSubscripts() {
                return parseSubscripts(parseExprAtom());
            }
            function parseSubscripts(base, noCalls) {
                if (eat(_dot)) {
                    var node = startNodeFrom(base);
                    node.object = base;
                    node.property = parseIdent(true);
                    node.computed = false;
                    return parseSubscripts(finishNode(node, "MemberExpression"), noCalls);
                } else {
                    if (eat(_bracketL)) {
                        var node = startNodeFrom(base);
                        node.object = base;
                        node.property = parseExpression();
                        node.computed = true;
                        expect(_bracketR);
                        return parseSubscripts(finishNode(node, "MemberExpression"), noCalls);
                    } else {
                        if (!noCalls && eat(_parenL)) {
                            var node = startNodeFrom(base);
                            node.callee = base;
                            node.arguments = parseExprList(_parenR, false);
                            return parseSubscripts(finishNode(node, "CallExpression"), noCalls);
                        } else {
                            return base;
                        }
                    }
                }
            }
            function parseExprAtom() {
                switch (tokType) {
                  case _this:
                    var node = startNode();
                    next();
                    return finishNode(node, "ThisExpression");

                  case _name:
                    return parseIdent();

                  case _num:
                  case _string:
                  case _regexp:
                    var node = startNode();
                    node.value = tokVal;
                    node.raw = input.slice(tokStart, tokEnd);
                    next();
                    return finishNode(node, "Literal");

                  case _null:
                  case _true:
                  case _false:
                    var node = startNode();
                    node.value = tokType.atomValue;
                    node.raw = tokType.keyword;
                    next();
                    return finishNode(node, "Literal");

                  case _parenL:
                    var tokStartLoc1 = tokStartLoc, tokStart1 = tokStart;
                    next();
                    var val = parseExpression();
                    val.start = tokStart1;
                    val.end = tokEnd;
                    if (options.locations) {
                        val.loc.start = tokStartLoc1;
                        val.loc.end = tokEndLoc;
                    }
                    if (options.ranges) {
                        val.range = [ tokStart1, tokEnd ];
                    }
                    expect(_parenR);
                    return val;

                  case _bracketL:
                    var node = startNode();
                    next();
                    node.elements = parseExprList(_bracketR, true, true);
                    return finishNode(node, "ArrayExpression");

                  case _braceL:
                    return parseObj();

                  case _function:
                    var node = startNode();
                    next();
                    return parseFunction(node, false);

                  case _new:
                    return parseNew();

                  default:
                    unexpected();
                }
            }
            function parseNew() {
                var node = startNode();
                next();
                node.callee = parseSubscripts(parseExprAtom(), true);
                if (eat(_parenL)) {
                    node.arguments = parseExprList(_parenR, false);
                } else {
                    node.arguments = empty;
                }
                return finishNode(node, "NewExpression");
            }
            function parseObj() {
                var node = startNode(), first = true, sawGetSet = false;
                node.properties = [];
                next();
                while (!eat(_braceR)) {
                    if (!first) {
                        expect(_comma);
                        if (options.allowTrailingCommas && eat(_braceR)) {
                            break;
                        }
                    } else {
                        first = false;
                    }
                    var prop = {
                        type: "Property",
                        key: parsePropertyName()
                    }, isGetSet = false, kind;
                    if (eat(_colon)) {
                        prop.value = parseExpression(true);
                        kind = prop.kind = "init";
                    } else {
                        if (options.ecmaVersion >= 5 && prop.key.type === "Identifier" && (prop.key.name === "get" || prop.key.name === "set")) {
                            isGetSet = sawGetSet = true;
                            kind = prop.kind = prop.key.name;
                            prop.key = parsePropertyName();
                            if (tokType !== _parenL) {
                                unexpected();
                            }
                            prop.value = parseFunction(startNode(), false);
                        } else {
                            unexpected();
                        }
                    }
                    if (prop.key.type === "Identifier" && (strict || sawGetSet)) {
                        for (var i = 0; i < node.properties.length; ++i) {
                            var other = node.properties[i];
                            if (other.key.name === prop.key.name) {
                                var conflict = kind == other.kind || isGetSet && other.kind === "init" || kind === "init" && (other.kind === "get" || other.kind === "set");
                                if (conflict && !strict && kind === "init" && other.kind === "init") {
                                    conflict = false;
                                }
                                if (conflict) {
                                    raise(prop.key.start, "Redefinition of property");
                                }
                            }
                        }
                    }
                    node.properties.push(prop);
                }
                return finishNode(node, "ObjectExpression");
            }
            function parsePropertyName() {
                if (tokType === _num || tokType === _string) {
                    return parseExprAtom();
                }
                return parseIdent(true);
            }
            function parseFunction(node, isStatement) {
                if (tokType === _name) {
                    node.id = parseIdent();
                } else {
                    if (isStatement) {
                        unexpected();
                    } else {
                        node.id = null;
                    }
                }
                node.params = [];
                var first = true;
                expect(_parenL);
                while (!eat(_parenR)) {
                    if (!first) {
                        expect(_comma);
                    } else {
                        first = false;
                    }
                    node.params.push(parseIdent());
                }
                var oldInFunc = inFunction, oldLabels = labels;
                inFunction = true;
                labels = [];
                node.body = parseBlock(true);
                inFunction = oldInFunc;
                labels = oldLabels;
                if (strict || node.body.body.length && isUseStrict(node.body.body[0])) {
                    for (var i = node.id ? -1 : 0; i < node.params.length; ++i) {
                        var id = i < 0 ? node.id : node.params[i];
                        if (isStrictReservedWord(id.name) || isStrictBadIdWord(id.name)) {
                            raise(id.start, "Defining '" + id.name + "' in strict mode");
                        }
                        if (i >= 0) {
                            for (var j = 0; j < i; ++j) {
                                if (id.name === node.params[j].name) {
                                    raise(id.start, "Argument name clash in strict mode");
                                }
                            }
                        }
                    }
                }
                return finishNode(node, isStatement ? "FunctionDeclaration" : "FunctionExpression");
            }
            function parseExprList(close, allowTrailingComma, allowEmpty) {
                var elts = [], first = true;
                while (!eat(close)) {
                    if (!first) {
                        expect(_comma);
                        if (allowTrailingComma && options.allowTrailingCommas && eat(close)) {
                            break;
                        }
                    } else {
                        first = false;
                    }
                    if (allowEmpty && tokType === _comma) {
                        elts.push(null);
                    } else {
                        elts.push(parseExpression(true));
                    }
                }
                return elts;
            }
            function parseIdent(liberal) {
                var node = startNode();
                node.name = tokType === _name ? tokVal : liberal && !options.forbidReserved && tokType.keyword || unexpected();
                tokRegexpAllowed = false;
                next();
                return finishNode(node, "Identifier");
            }
        });
    }).call(acornExports);
    JSParsingTools.parse = acornExports.parse || acornExports.acorn.parse;
    var codeGetExports = {};
    (function(exports) {
        (function(global) {
            function require(file, parentModule) {
                if ({}.hasOwnProperty.call(require.cache, file)) {
                    return require.cache[file];
                }
                var resolved = require.resolve(file);
                if (!resolved) {
                    throw new Error("Failed to resolve module " + file);
                }
                var module$ = {
                    id: file,
                    require: require,
                    filename: file,
                    exports: {},
                    loaded: false,
                    parent: parentModule,
                    children: []
                };
                if (parentModule) {
                    parentModule.children.push(module$);
                }
                var dirname = file.slice(0, file.lastIndexOf("/") + 1);
                require.cache[file] = module$.exports;
                resolved.call(module$.exports, module$, module$.exports, dirname, file);
                module$.loaded = true;
                return require.cache[file] = module$.exports;
            }
            require.modules = {};
            require.cache = {};
            require.resolve = function(file) {
                return {}.hasOwnProperty.call(require.modules, file) ? require.modules[file] : void 0;
            };
            require.define = function(file, fn) {
                require.modules[file] = fn;
            };
            var process = function() {
                var cwd = "/";
                return {
                    title: "browser",
                    version: "v0.10.21",
                    browser: true,
                    env: {},
                    argv: [],
                    nextTick: global.setImmediate || function(fn) {
                        setTimeout(fn, 0);
                    },
                    cwd: function() {
                        return cwd;
                    },
                    chdir: function(dir) {
                        cwd = dir;
                    }
                };
            }();
            require.define("/tools/entry-point.js", function(module, exports, __dirname, __filename) {
                (function() {
                    "use strict";
                    global.escodegen = require("/escodegen.js", module);
                })();
            });
            require.define("/escodegen.js", function(module, exports, __dirname, __filename) {
                (function() {
                    "use strict";
                    var Syntax, Precedence, BinaryPrecedence, SourceNode, estraverse, esutils, isArray, base, indent, json, renumber, hexadecimal, quotes, escapeless, newline, space, parentheses, semicolons, safeConcatenation, directive, extra, parse, sourceMap, FORMAT_MINIFY, FORMAT_DEFAULTS;
                    estraverse = require("/node_modules/estraverse/estraverse.js", module);
                    esutils = require("/node_modules/esutils/lib/utils.js", module);
                    Syntax = {
                        AssignmentExpression: "AssignmentExpression",
                        ArrayExpression: "ArrayExpression",
                        ArrayPattern: "ArrayPattern",
                        ArrowFunctionExpression: "ArrowFunctionExpression",
                        BlockStatement: "BlockStatement",
                        BinaryExpression: "BinaryExpression",
                        BreakStatement: "BreakStatement",
                        CallExpression: "CallExpression",
                        CatchClause: "CatchClause",
                        ComprehensionBlock: "ComprehensionBlock",
                        ComprehensionExpression: "ComprehensionExpression",
                        ConditionalExpression: "ConditionalExpression",
                        ContinueStatement: "ContinueStatement",
                        DirectiveStatement: "DirectiveStatement",
                        DoWhileStatement: "DoWhileStatement",
                        DebuggerStatement: "DebuggerStatement",
                        EmptyStatement: "EmptyStatement",
                        ExpressionStatement: "ExpressionStatement",
                        ForStatement: "ForStatement",
                        ForInStatement: "ForInStatement",
                        FunctionDeclaration: "FunctionDeclaration",
                        FunctionExpression: "FunctionExpression",
                        Identifier: "Identifier",
                        IfStatement: "IfStatement",
                        Literal: "Literal",
                        LabeledStatement: "LabeledStatement",
                        LogicalExpression: "LogicalExpression",
                        MemberExpression: "MemberExpression",
                        NewExpression: "NewExpression",
                        ObjectExpression: "ObjectExpression",
                        ObjectPattern: "ObjectPattern",
                        Program: "Program",
                        Property: "Property",
                        ReturnStatement: "ReturnStatement",
                        SequenceExpression: "SequenceExpression",
                        SwitchStatement: "SwitchStatement",
                        SwitchCase: "SwitchCase",
                        ThisExpression: "ThisExpression",
                        ThrowStatement: "ThrowStatement",
                        TryStatement: "TryStatement",
                        UnaryExpression: "UnaryExpression",
                        UpdateExpression: "UpdateExpression",
                        VariableDeclaration: "VariableDeclaration",
                        VariableDeclarator: "VariableDeclarator",
                        WhileStatement: "WhileStatement",
                        WithStatement: "WithStatement",
                        YieldExpression: "YieldExpression"
                    };
                    Precedence = {
                        Sequence: 0,
                        Yield: 1,
                        Assignment: 1,
                        Conditional: 2,
                        ArrowFunction: 2,
                        LogicalOR: 3,
                        LogicalAND: 4,
                        BitwiseOR: 5,
                        BitwiseXOR: 6,
                        BitwiseAND: 7,
                        Equality: 8,
                        Relational: 9,
                        BitwiseSHIFT: 10,
                        Additive: 11,
                        Multiplicative: 12,
                        Unary: 13,
                        Postfix: 14,
                        Call: 15,
                        New: 16,
                        Member: 17,
                        Primary: 18
                    };
                    BinaryPrecedence = {
                        "||": Precedence.LogicalOR,
                        "&&": Precedence.LogicalAND,
                        "|": Precedence.BitwiseOR,
                        "^": Precedence.BitwiseXOR,
                        "&": Precedence.BitwiseAND,
                        "==": Precedence.Equality,
                        "!=": Precedence.Equality,
                        "===": Precedence.Equality,
                        "!==": Precedence.Equality,
                        is: Precedence.Equality,
                        isnt: Precedence.Equality,
                        "<": Precedence.Relational,
                        ">": Precedence.Relational,
                        "<=": Precedence.Relational,
                        ">=": Precedence.Relational,
                        "in": Precedence.Relational,
                        "instanceof": Precedence.Relational,
                        "<<": Precedence.BitwiseSHIFT,
                        ">>": Precedence.BitwiseSHIFT,
                        ">>>": Precedence.BitwiseSHIFT,
                        "+": Precedence.Additive,
                        "-": Precedence.Additive,
                        "*": Precedence.Multiplicative,
                        "%": Precedence.Multiplicative,
                        "/": Precedence.Multiplicative
                    };
                    function getDefaultOptions() {
                        return {
                            indent: null,
                            base: null,
                            parse: null,
                            comment: false,
                            format: {
                                indent: {
                                    style: "    ",
                                    base: 0,
                                    adjustMultilineComment: false
                                },
                                newline: "\n",
                                space: " ",
                                json: false,
                                renumber: false,
                                hexadecimal: false,
                                quotes: "single",
                                escapeless: false,
                                compact: false,
                                parentheses: true,
                                semicolons: true,
                                safeConcatenation: false
                            },
                            moz: {
                                starlessGenerator: false,
                                parenthesizedComprehensionBlock: false
                            },
                            sourceMap: null,
                            sourceMapRoot: null,
                            sourceMapWithCode: false,
                            directive: false,
                            verbatim: null
                        };
                    }
                    function stringRepeat(str, num) {
                        var result = "";
                        for (num |= 0; num > 0; num >>>= 1, str += str) {
                            if (num & 1) {
                                result += str;
                            }
                        }
                        return result;
                    }
                    isArray = Array.isArray;
                    if (!isArray) {
                        isArray = function isArray(array) {
                            return Object.prototype.toString.call(array) === "[object Array]";
                        };
                    }
                    function SourceNodeMock(line, column, filename, chunk) {
                        var result = [];
                        function flatten(input) {
                            var i, iz;
                            if (isArray(input)) {
                                for (i = 0, iz = input.length; i < iz; ++i) {
                                    flatten(input[i]);
                                }
                            } else {
                                if (input instanceof SourceNodeMock) {
                                    result.push(input);
                                } else {
                                    if (typeof input === "string" && input) {
                                        result.push(input);
                                    }
                                }
                            }
                        }
                        flatten(chunk);
                        this.children = result;
                    }
                    SourceNodeMock.prototype.toString = function toString() {
                        var res = "", i, iz, node;
                        for (i = 0, iz = this.children.length; i < iz; ++i) {
                            node = this.children[i];
                            if (node instanceof SourceNodeMock) {
                                res += node.toString();
                            } else {
                                res += node;
                            }
                        }
                        return res;
                    };
                    SourceNodeMock.prototype.replaceRight = function replaceRight(pattern, replacement) {
                        var last = this.children[this.children.length - 1];
                        if (last instanceof SourceNodeMock) {
                            last.replaceRight(pattern, replacement);
                        } else {
                            if (typeof last === "string") {
                                this.children[this.children.length - 1] = last.replace(pattern, replacement);
                            } else {
                                this.children.push("".replace(pattern, replacement));
                            }
                        }
                        return this;
                    };
                    SourceNodeMock.prototype.join = function join(sep) {
                        var i, iz, result;
                        result = [];
                        iz = this.children.length;
                        if (iz > 0) {
                            --iz;
                            for (i = 0; i < iz; ++i) {
                                result.push(this.children[i], sep);
                            }
                            result.push(this.children[iz]);
                            this.children = result;
                        }
                        return this;
                    };
                    function hasLineTerminator(str) {
                        return /[\r\n]/g.test(str);
                    }
                    function endsWithLineTerminator(str) {
                        var len = str.length;
                        return len && esutils.code.isLineTerminator(str.charCodeAt(len - 1));
                    }
                    function updateDeeply(target, override) {
                        var key, val;
                        function isHashObject(target) {
                            return typeof target === "object" && target instanceof Object && !(target instanceof RegExp);
                        }
                        for (key in override) {
                            if (override.hasOwnProperty(key)) {
                                val = override[key];
                                if (isHashObject(val)) {
                                    if (isHashObject(target[key])) {
                                        updateDeeply(target[key], val);
                                    } else {
                                        target[key] = updateDeeply({}, val);
                                    }
                                } else {
                                    target[key] = val;
                                }
                            }
                        }
                        return target;
                    }
                    function generateNumber(value) {
                        var result, point, temp, exponent, pos;
                        if (value !== value) {
                            throw new Error("Numeric literal whose value is NaN");
                        }
                        if (value < 0 || value === 0 && 1 / value < 0) {
                            throw new Error("Numeric literal whose value is negative");
                        }
                        if (value === 1 / 0) {
                            return json ? "null" : renumber ? "1e400" : "1e+400";
                        }
                        result = "" + value;
                        if (!renumber || result.length < 3) {
                            return result;
                        }
                        point = result.indexOf(".");
                        if (!json && result.charCodeAt(0) === 48 && point === 1) {
                            point = 0;
                            result = result.slice(1);
                        }
                        temp = result;
                        result = result.replace("e+", "e");
                        exponent = 0;
                        if ((pos = temp.indexOf("e")) > 0) {
                            exponent = +temp.slice(pos + 1);
                            temp = temp.slice(0, pos);
                        }
                        if (point >= 0) {
                            exponent -= temp.length - point - 1;
                            temp = +(temp.slice(0, point) + temp.slice(point + 1)) + "";
                        }
                        pos = 0;
                        while (temp.charCodeAt(temp.length + pos - 1) === 48) {
                            --pos;
                        }
                        if (pos !== 0) {
                            exponent -= pos;
                            temp = temp.slice(0, pos);
                        }
                        if (exponent !== 0) {
                            temp += "e" + exponent;
                        }
                        if ((temp.length < result.length || hexadecimal && value > 1e12 && Math.floor(value) === value && (temp = "0x" + value.toString(16)).length < result.length) && +temp === value) {
                            result = temp;
                        }
                        return result;
                    }
                    function escapeRegExpCharacter(ch, previousIsBackslash) {
                        if ((ch & ~1) === 8232) {
                            return (previousIsBackslash ? "u" : "\\u") + (ch === 8232 ? "2028" : "2029");
                        } else {
                            if (ch === 10 || ch === 13) {
                                return (previousIsBackslash ? "" : "\\") + (ch === 10 ? "n" : "r");
                            }
                        }
                        return String.fromCharCode(ch);
                    }
                    function generateRegExp(reg) {
                        var match, result, flags, i, iz, ch, characterInBrack, previousIsBackslash;
                        result = reg.toString();
                        if (reg.source) {
                            match = result.match(/\/([^/]*)$/);
                            if (!match) {
                                return result;
                            }
                            flags = match[1];
                            result = "";
                            characterInBrack = false;
                            previousIsBackslash = false;
                            for (i = 0, iz = reg.source.length; i < iz; ++i) {
                                ch = reg.source.charCodeAt(i);
                                if (!previousIsBackslash) {
                                    if (characterInBrack) {
                                        if (ch === 93) {
                                            characterInBrack = false;
                                        }
                                    } else {
                                        if (ch === 47) {
                                            result += "\\";
                                        } else {
                                            if (ch === 91) {
                                                characterInBrack = true;
                                            }
                                        }
                                    }
                                    result += escapeRegExpCharacter(ch, previousIsBackslash);
                                    previousIsBackslash = ch === 92;
                                } else {
                                    result += escapeRegExpCharacter(ch, previousIsBackslash);
                                    previousIsBackslash = false;
                                }
                            }
                            return "/" + result + "/" + flags;
                        }
                        return result;
                    }
                    function escapeAllowedCharacter(code, next) {
                        var hex, result = "\\";
                        switch (code) {
                          case 8:
                            result += "b";
                            break;

                          case 12:
                            result += "f";
                            break;

                          case 9:
                            result += "t";
                            break;

                          default:
                            hex = code.toString(16);
                            if (json || code > 255) {
                                result += "u" + "0000".slice(hex.length) + hex;
                            } else {
                                if (code === 0 && !esutils.code.isDecimalDigit(next)) {
                                    result += "0";
                                } else {
                                    if (code === 11) {
                                        result += "x0B";
                                    } else {
                                        result += "x" + "00".slice(hex.length) + hex;
                                    }
                                }
                            }
                            break;
                        }
                        return result;
                    }
                    function escapeDisallowedCharacter(code) {
                        var result = "\\";
                        switch (code) {
                          case 92:
                            result += "\\";
                            break;

                          case 10:
                            result += "n";
                            break;

                          case 13:
                            result += "r";
                            break;

                          case 8232:
                            result += "u2028";
                            break;

                          case 8233:
                            result += "u2029";
                            break;

                          default:
                            throw new Error("Incorrectly classified character");
                        }
                        return result;
                    }
                    function escapeDirective(str) {
                        var i, iz, code, quote;
                        quote = quotes === "double" ? '"' : "'";
                        for (i = 0, iz = str.length; i < iz; ++i) {
                            code = str.charCodeAt(i);
                            if (code === 39) {
                                quote = '"';
                                break;
                            } else {
                                if (code === 34) {
                                    quote = "'";
                                    break;
                                } else {
                                    if (code === 92) {
                                        ++i;
                                    }
                                }
                            }
                        }
                        return quote + str + quote;
                    }
                    function escapeString(str) {
                        var result = "", i, len, code, singleQuotes = 0, doubleQuotes = 0, single, quote;
                        for (i = 0, len = str.length; i < len; ++i) {
                            code = str.charCodeAt(i);
                            if (code === 39) {
                                ++singleQuotes;
                            } else {
                                if (code === 34) {
                                    ++doubleQuotes;
                                } else {
                                    if (code === 47 && json) {
                                        result += "\\";
                                    } else {
                                        if (esutils.code.isLineTerminator(code) || code === 92) {
                                            result += escapeDisallowedCharacter(code);
                                            continue;
                                        } else {
                                            if (json && code < 32 || !(json || escapeless || code >= 32 && code <= 126)) {
                                                result += escapeAllowedCharacter(code, str.charCodeAt(i + 1));
                                                continue;
                                            }
                                        }
                                    }
                                }
                            }
                            result += String.fromCharCode(code);
                        }
                        single = !(quotes === "double" || quotes === "auto" && doubleQuotes < singleQuotes);
                        quote = single ? "'" : '"';
                        if (!(single ? singleQuotes : doubleQuotes)) {
                            return quote + result + quote;
                        }
                        str = result;
                        result = quote;
                        for (i = 0, len = str.length; i < len; ++i) {
                            code = str.charCodeAt(i);
                            if (code === 39 && single || code === 34 && !single) {
                                result += "\\";
                            }
                            result += String.fromCharCode(code);
                        }
                        return result + quote;
                    }
                    function toSourceNode(generated, node) {
                        if (node == null) {
                            if (generated instanceof SourceNode) {
                                return generated;
                            } else {
                                node = {};
                            }
                        }
                        if (node.loc == null) {
                            return new SourceNode(null, null, sourceMap, generated, node.name || null);
                        }
                        return new SourceNode(node.loc.start.line, node.loc.start.column, sourceMap === true ? node.loc.source || null : sourceMap, generated, node.name || null);
                    }
                    function noEmptySpace() {
                        return space ? space : " ";
                    }
                    function join(left, right) {
                        var leftSource = toSourceNode(left).toString(), rightSource = toSourceNode(right).toString(), leftCharCode = leftSource.charCodeAt(leftSource.length - 1), rightCharCode = rightSource.charCodeAt(0);
                        if ((leftCharCode === 43 || leftCharCode === 45) && leftCharCode === rightCharCode || esutils.code.isIdentifierPart(leftCharCode) && esutils.code.isIdentifierPart(rightCharCode) || leftCharCode === 47 && rightCharCode === 105) {
                            return [ left, noEmptySpace(), right ];
                        } else {
                            if (esutils.code.isWhiteSpace(leftCharCode) || esutils.code.isLineTerminator(leftCharCode) || esutils.code.isWhiteSpace(rightCharCode) || esutils.code.isLineTerminator(rightCharCode)) {
                                return [ left, right ];
                            }
                        }
                        return [ left, space, right ];
                    }
                    function addIndent(stmt) {
                        return [ base, stmt ];
                    }
                    function withIndent(fn) {
                        var previousBase, result;
                        previousBase = base;
                        base += indent;
                        result = fn.call(this, base);
                        base = previousBase;
                        return result;
                    }
                    function calculateSpaces(str) {
                        var i;
                        for (i = str.length - 1; i >= 0; --i) {
                            if (esutils.code.isLineTerminator(str.charCodeAt(i))) {
                                break;
                            }
                        }
                        return str.length - 1 - i;
                    }
                    function adjustMultilineComment(value, specialBase) {
                        var array, i, len, line, j, spaces, previousBase;
                        array = value.split(/\r\n|[\r\n]/);
                        spaces = Number.MAX_VALUE;
                        for (i = 1, len = array.length; i < len; ++i) {
                            line = array[i];
                            j = 0;
                            while (j < line.length && esutils.code.isWhiteSpace(line.charCodeAt(j))) {
                                ++j;
                            }
                            if (spaces > j) {
                                spaces = j;
                            }
                        }
                        if (typeof specialBase !== "undefined") {
                            previousBase = base;
                            if (array[1][spaces] === "*") {
                                specialBase += " ";
                            }
                            base = specialBase;
                        } else {
                            if (spaces & 1) {
                                --spaces;
                            }
                            previousBase = base;
                        }
                        for (i = 1, len = array.length; i < len; ++i) {
                            array[i] = toSourceNode(addIndent(array[i].slice(spaces))).join("");
                        }
                        base = previousBase;
                        return array.join("\n");
                    }
                    function generateComment(comment, specialBase) {
                        if (comment.type === "Line") {
                            if (endsWithLineTerminator(comment.value)) {
                                return "//" + comment.value;
                            } else {
                                return "//" + comment.value + "\n";
                            }
                        }
                        if (extra.format.indent.adjustMultilineComment && /[\n\r]/.test(comment.value)) {
                            return adjustMultilineComment("/*" + comment.value + "*/", specialBase);
                        }
                        return "/*" + comment.value + "*/";
                    }
                    function addCommentsToStatement(stmt, result) {
                        var i, len, comment, save, tailingToStatement, specialBase, fragment;
                        if (stmt.leadingComments && stmt.leadingComments.length > 0) {
                            save = result;
                            comment = stmt.leadingComments[0];
                            result = [];
                            if (safeConcatenation && stmt.type === Syntax.Program && stmt.body.length === 0) {
                                result.push("\n");
                            }
                            result.push(generateComment(comment));
                            if (!endsWithLineTerminator(toSourceNode(result).toString())) {
                                result.push("\n");
                            }
                            for (i = 1, len = stmt.leadingComments.length; i < len; ++i) {
                                comment = stmt.leadingComments[i];
                                fragment = [ generateComment(comment) ];
                                if (!endsWithLineTerminator(toSourceNode(fragment).toString())) {
                                    fragment.push("\n");
                                }
                                result.push(addIndent(fragment));
                            }
                            result.push(addIndent(save));
                        }
                        if (stmt.trailingComments) {
                            tailingToStatement = !endsWithLineTerminator(toSourceNode(result).toString());
                            specialBase = stringRepeat(" ", calculateSpaces(toSourceNode([ base, result, indent ]).toString()));
                            for (i = 0, len = stmt.trailingComments.length; i < len; ++i) {
                                comment = stmt.trailingComments[i];
                                if (tailingToStatement) {
                                    if (i === 0) {
                                        result = [ result, indent ];
                                    } else {
                                        result = [ result, specialBase ];
                                    }
                                    result.push(generateComment(comment, specialBase));
                                } else {
                                    result = [ result, addIndent(generateComment(comment)) ];
                                }
                                if (i !== len - 1 && !endsWithLineTerminator(toSourceNode(result).toString())) {
                                    result = [ result, "\n" ];
                                }
                            }
                        }
                        return result;
                    }
                    function parenthesize(text, current, should) {
                        if (current < should) {
                            return [ "(", text, ")" ];
                        }
                        return text;
                    }
                    function maybeBlock(stmt, semicolonOptional, functionBody) {
                        var result, noLeadingComment;
                        noLeadingComment = !extra.comment || !stmt.leadingComments;
                        if (stmt.type === Syntax.BlockStatement && noLeadingComment) {
                            return [ space, generateStatement(stmt, {
                                functionBody: functionBody
                            }) ];
                        }
                        if (stmt.type === Syntax.EmptyStatement && noLeadingComment) {
                            return ";";
                        }
                        withIndent(function() {
                            result = [ newline, addIndent(generateStatement(stmt, {
                                semicolonOptional: semicolonOptional,
                                functionBody: functionBody
                            })) ];
                        });
                        return result;
                    }
                    function maybeBlockSuffix(stmt, result) {
                        var ends = endsWithLineTerminator(toSourceNode(result).toString());
                        if (stmt.type === Syntax.BlockStatement && (!extra.comment || !stmt.leadingComments) && !ends) {
                            return [ result, space ];
                        }
                        if (ends) {
                            return [ result, base ];
                        }
                        return [ result, newline, base ];
                    }
                    function generateVerbatim(expr, option) {
                        var i, result;
                        result = expr[extra.verbatim].split(/\r\n|\n/);
                        for (i = 1; i < result.length; i++) {
                            result[i] = newline + base + result[i];
                        }
                        result = parenthesize(result, Precedence.Sequence, option.precedence);
                        return toSourceNode(result, expr);
                    }
                    function generateIdentifier(node) {
                        return toSourceNode(node.name, node);
                    }
                    function generateFunctionBody(node) {
                        var result, i, len, expr, arrow;
                        arrow = node.type === Syntax.ArrowFunctionExpression;
                        if (arrow && node.params.length === 1 && node.params[0].type === Syntax.Identifier) {
                            result = [ generateIdentifier(node.params[0]) ];
                        } else {
                            result = [ "(" ];
                            for (i = 0, len = node.params.length; i < len; ++i) {
                                result.push(generateIdentifier(node.params[i]));
                                if (i + 1 < len) {
                                    result.push("," + space);
                                }
                            }
                            result.push(")");
                        }
                        if (arrow) {
                            result.push(space, "=>");
                        }
                        if (node.expression) {
                            result.push(space);
                            expr = generateExpression(node.body, {
                                precedence: Precedence.Assignment,
                                allowIn: true,
                                allowCall: true
                            });
                            if (expr.toString().charAt(0) === "{") {
                                expr = [ "(", expr, ")" ];
                            }
                            result.push(expr);
                        } else {
                            result.push(maybeBlock(node.body, false, true));
                        }
                        return result;
                    }
                    function generateExpression(expr, option) {
                        var result, precedence, type, currentPrecedence, i, len, raw, fragment, multiline, leftCharCode, leftSource, rightCharCode, allowIn, allowCall, allowUnparenthesizedNew, property;
                        precedence = option.precedence;
                        allowIn = option.allowIn;
                        allowCall = option.allowCall;
                        type = expr.type || option.type;
                        if (extra.verbatim && expr.hasOwnProperty(extra.verbatim)) {
                            return generateVerbatim(expr, option);
                        }
                        switch (type) {
                          case Syntax.SequenceExpression:
                            result = [];
                            allowIn |= Precedence.Sequence < precedence;
                            for (i = 0, len = expr.expressions.length; i < len; ++i) {
                                result.push(generateExpression(expr.expressions[i], {
                                    precedence: Precedence.Assignment,
                                    allowIn: allowIn,
                                    allowCall: true
                                }));
                                if (i + 1 < len) {
                                    result.push("," + space);
                                }
                            }
                            result = parenthesize(result, Precedence.Sequence, precedence);
                            break;

                          case Syntax.AssignmentExpression:
                            allowIn |= Precedence.Assignment < precedence;
                            result = parenthesize([ generateExpression(expr.left, {
                                precedence: Precedence.Call,
                                allowIn: allowIn,
                                allowCall: true
                            }), space + expr.operator + space, generateExpression(expr.right, {
                                precedence: Precedence.Assignment,
                                allowIn: allowIn,
                                allowCall: true
                            }) ], Precedence.Assignment, precedence);
                            break;

                          case Syntax.ArrowFunctionExpression:
                            allowIn |= Precedence.ArrowFunction < precedence;
                            result = parenthesize(generateFunctionBody(expr), Precedence.ArrowFunction, precedence);
                            break;

                          case Syntax.ConditionalExpression:
                            allowIn |= Precedence.Conditional < precedence;
                            result = parenthesize([ generateExpression(expr.test, {
                                precedence: Precedence.LogicalOR,
                                allowIn: allowIn,
                                allowCall: true
                            }), space + "?" + space, generateExpression(expr.consequent, {
                                precedence: Precedence.Assignment,
                                allowIn: allowIn,
                                allowCall: true
                            }), space + ":" + space, generateExpression(expr.alternate, {
                                precedence: Precedence.Assignment,
                                allowIn: allowIn,
                                allowCall: true
                            }) ], Precedence.Conditional, precedence);
                            break;

                          case Syntax.LogicalExpression:
                          case Syntax.BinaryExpression:
                            currentPrecedence = BinaryPrecedence[expr.operator];
                            allowIn |= currentPrecedence < precedence;
                            fragment = generateExpression(expr.left, {
                                precedence: currentPrecedence,
                                allowIn: allowIn,
                                allowCall: true
                            });
                            leftSource = fragment.toString();
                            if (leftSource.charCodeAt(leftSource.length - 1) === 47 && esutils.code.isIdentifierPart(expr.operator.charCodeAt(0))) {
                                result = [ fragment, noEmptySpace(), expr.operator ];
                            } else {
                                result = join(fragment, expr.operator);
                            }
                            fragment = generateExpression(expr.right, {
                                precedence: currentPrecedence + 1,
                                allowIn: allowIn,
                                allowCall: true
                            });
                            if (expr.operator === "/" && fragment.toString().charAt(0) === "/" || expr.operator.slice(-1) === "<" && fragment.toString().slice(0, 3) === "!--") {
                                result.push(noEmptySpace(), fragment);
                            } else {
                                result = join(result, fragment);
                            }
                            if (expr.operator === "in" && !allowIn) {
                                result = [ "(", result, ")" ];
                            } else {
                                result = parenthesize(result, currentPrecedence, precedence);
                            }
                            break;

                          case Syntax.CallExpression:
                            result = [ generateExpression(expr.callee, {
                                precedence: Precedence.Call,
                                allowIn: true,
                                allowCall: true,
                                allowUnparenthesizedNew: false
                            }) ];
                            result.push("(");
                            for (i = 0, len = expr["arguments"].length; i < len; ++i) {
                                result.push(generateExpression(expr["arguments"][i], {
                                    precedence: Precedence.Assignment,
                                    allowIn: true,
                                    allowCall: true
                                }));
                                if (i + 1 < len) {
                                    result.push("," + space);
                                }
                            }
                            result.push(")");
                            if (!allowCall) {
                                result = [ "(", result, ")" ];
                            } else {
                                result = parenthesize(result, Precedence.Call, precedence);
                            }
                            break;

                          case Syntax.NewExpression:
                            len = expr["arguments"].length;
                            allowUnparenthesizedNew = option.allowUnparenthesizedNew === undefined || option.allowUnparenthesizedNew;
                            result = join("new", generateExpression(expr.callee, {
                                precedence: Precedence.New,
                                allowIn: true,
                                allowCall: false,
                                allowUnparenthesizedNew: allowUnparenthesizedNew && !parentheses && len === 0
                            }));
                            if (!allowUnparenthesizedNew || parentheses || len > 0) {
                                result.push("(");
                                for (i = 0; i < len; ++i) {
                                    result.push(generateExpression(expr["arguments"][i], {
                                        precedence: Precedence.Assignment,
                                        allowIn: true,
                                        allowCall: true
                                    }));
                                    if (i + 1 < len) {
                                        result.push("," + space);
                                    }
                                }
                                result.push(")");
                            }
                            result = parenthesize(result, Precedence.New, precedence);
                            break;

                          case Syntax.MemberExpression:
                            result = [ generateExpression(expr.object, {
                                precedence: Precedence.Call,
                                allowIn: true,
                                allowCall: allowCall,
                                allowUnparenthesizedNew: false
                            }) ];
                            if (expr.computed) {
                                result.push("[", generateExpression(expr.property, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: allowCall
                                }), "]");
                            } else {
                                if (expr.object.type === Syntax.Literal && typeof expr.object.value === "number") {
                                    fragment = toSourceNode(result).toString();
                                    if (fragment.indexOf(".") < 0 && !/[eExX]/.test(fragment) && esutils.code.isDecimalDigit(fragment.charCodeAt(fragment.length - 1)) && !(fragment.length >= 2 && fragment.charCodeAt(0) === 48)) {
                                        result.push(".");
                                    }
                                }
                                result.push(".", generateIdentifier(expr.property));
                            }
                            result = parenthesize(result, Precedence.Member, precedence);
                            break;

                          case Syntax.UnaryExpression:
                            fragment = generateExpression(expr.argument, {
                                precedence: Precedence.Unary,
                                allowIn: true,
                                allowCall: true
                            });
                            if (space === "") {
                                result = join(expr.operator, fragment);
                            } else {
                                result = [ expr.operator ];
                                if (expr.operator.length > 2) {
                                    result = join(result, fragment);
                                } else {
                                    leftSource = toSourceNode(result).toString();
                                    leftCharCode = leftSource.charCodeAt(leftSource.length - 1);
                                    rightCharCode = fragment.toString().charCodeAt(0);
                                    if ((leftCharCode === 43 || leftCharCode === 45) && leftCharCode === rightCharCode || esutils.code.isIdentifierPart(leftCharCode) && esutils.code.isIdentifierPart(rightCharCode)) {
                                        result.push(noEmptySpace(), fragment);
                                    } else {
                                        result.push(fragment);
                                    }
                                }
                            }
                            result = parenthesize(result, Precedence.Unary, precedence);
                            break;

                          case Syntax.YieldExpression:
                            if (expr.delegate) {
                                result = "yield*";
                            } else {
                                result = "yield";
                            }
                            if (expr.argument) {
                                result = join(result, generateExpression(expr.argument, {
                                    precedence: Precedence.Yield,
                                    allowIn: true,
                                    allowCall: true
                                }));
                            }
                            result = parenthesize(result, Precedence.Yield, precedence);
                            break;

                          case Syntax.UpdateExpression:
                            if (expr.prefix) {
                                result = parenthesize([ expr.operator, generateExpression(expr.argument, {
                                    precedence: Precedence.Unary,
                                    allowIn: true,
                                    allowCall: true
                                }) ], Precedence.Unary, precedence);
                            } else {
                                result = parenthesize([ generateExpression(expr.argument, {
                                    precedence: Precedence.Postfix,
                                    allowIn: true,
                                    allowCall: true
                                }), expr.operator ], Precedence.Postfix, precedence);
                            }
                            break;

                          case Syntax.FunctionExpression:
                            result = "function";
                            if (expr.id) {
                                result = [ result, noEmptySpace(), generateIdentifier(expr.id), generateFunctionBody(expr) ];
                            } else {
                                result = [ result + space, generateFunctionBody(expr) ];
                            }
                            break;

                          case Syntax.ArrayPattern:
                          case Syntax.ArrayExpression:
                            if (!expr.elements.length) {
                                result = "[]";
                                break;
                            }
                            multiline = expr.elements.length > 1;
                            result = [ "[", multiline ? newline : "" ];
                            withIndent(function(indent) {
                                for (i = 0, len = expr.elements.length; i < len; ++i) {
                                    if (!expr.elements[i]) {
                                        if (multiline) {
                                            result.push(indent);
                                        }
                                        if (i + 1 === len) {
                                            result.push(",");
                                        }
                                    } else {
                                        result.push(multiline ? indent : "", generateExpression(expr.elements[i], {
                                            precedence: Precedence.Assignment,
                                            allowIn: true,
                                            allowCall: true
                                        }));
                                    }
                                    if (i + 1 < len) {
                                        result.push("," + (multiline ? newline : space));
                                    }
                                }
                            });
                            if (multiline && !endsWithLineTerminator(toSourceNode(result).toString())) {
                                result.push(newline);
                            }
                            result.push(multiline ? base : "", "]");
                            break;

                          case Syntax.Property:
                            if (expr.kind === "get" || expr.kind === "set") {
                                result = [ expr.kind, noEmptySpace(), generateExpression(expr.key, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true
                                }), generateFunctionBody(expr.value) ];
                            } else {
                                if (expr.shorthand) {
                                    result = generateExpression(expr.key, {
                                        precedence: Precedence.Sequence,
                                        allowIn: true,
                                        allowCall: true
                                    });
                                } else {
                                    if (expr.method) {
                                        result = [];
                                        if (expr.value.generator) {
                                            result.push("*");
                                        }
                                        result.push(generateExpression(expr.key, {
                                            precedence: Precedence.Sequence,
                                            allowIn: true,
                                            allowCall: true
                                        }), generateFunctionBody(expr.value));
                                    } else {
                                        result = [ generateExpression(expr.key, {
                                            precedence: Precedence.Sequence,
                                            allowIn: true,
                                            allowCall: true
                                        }), ":" + space, generateExpression(expr.value, {
                                            precedence: Precedence.Assignment,
                                            allowIn: true,
                                            allowCall: true
                                        }) ];
                                    }
                                }
                            }
                            break;

                          case Syntax.ObjectExpression:
                            if (!expr.properties.length) {
                                result = "{}";
                                break;
                            }
                            multiline = expr.properties.length > 1;
                            withIndent(function() {
                                fragment = generateExpression(expr.properties[0], {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true,
                                    type: Syntax.Property
                                });
                            });
                            if (!multiline) {
                                if (!hasLineTerminator(toSourceNode(fragment).toString())) {
                                    result = [ "{", space, fragment, space, "}" ];
                                    break;
                                }
                            }
                            withIndent(function(indent) {
                                result = [ "{", newline, indent, fragment ];
                                if (multiline) {
                                    result.push("," + newline);
                                    for (i = 1, len = expr.properties.length; i < len; ++i) {
                                        result.push(indent, generateExpression(expr.properties[i], {
                                            precedence: Precedence.Sequence,
                                            allowIn: true,
                                            allowCall: true,
                                            type: Syntax.Property
                                        }));
                                        if (i + 1 < len) {
                                            result.push("," + newline);
                                        }
                                    }
                                }
                            });
                            if (!endsWithLineTerminator(toSourceNode(result).toString())) {
                                result.push(newline);
                            }
                            result.push(base, "}");
                            break;

                          case Syntax.ObjectPattern:
                            if (!expr.properties.length) {
                                result = "{}";
                                break;
                            }
                            multiline = false;
                            if (expr.properties.length === 1) {
                                property = expr.properties[0];
                                if (property.value.type !== Syntax.Identifier) {
                                    multiline = true;
                                }
                            } else {
                                for (i = 0, len = expr.properties.length; i < len; ++i) {
                                    property = expr.properties[i];
                                    if (!property.shorthand) {
                                        multiline = true;
                                        break;
                                    }
                                }
                            }
                            result = [ "{", multiline ? newline : "" ];
                            withIndent(function(indent) {
                                for (i = 0, len = expr.properties.length; i < len; ++i) {
                                    result.push(multiline ? indent : "", generateExpression(expr.properties[i], {
                                        precedence: Precedence.Sequence,
                                        allowIn: true,
                                        allowCall: true
                                    }));
                                    if (i + 1 < len) {
                                        result.push("," + (multiline ? newline : space));
                                    }
                                }
                            });
                            if (multiline && !endsWithLineTerminator(toSourceNode(result).toString())) {
                                result.push(newline);
                            }
                            result.push(multiline ? base : "", "}");
                            break;

                          case Syntax.ThisExpression:
                            result = "this";
                            break;

                          case Syntax.Identifier:
                            result = generateIdentifier(expr);
                            break;

                          case Syntax.Literal:
                            if (expr.hasOwnProperty("raw")) {
                                result = expr.raw;
                                break;
                            }
                            if (expr.value === null) {
                                result = "null";
                                break;
                            }
                            if (typeof expr.value === "string") {
                                result = escapeString(expr.value);
                                break;
                            }
                            if (typeof expr.value === "number") {
                                result = generateNumber(expr.value);
                                break;
                            }
                            if (typeof expr.value === "boolean") {
                                result = expr.value ? "true" : "false";
                                break;
                            }
                            result = generateRegExp(expr.value);
                            break;

                          case Syntax.ComprehensionExpression:
                            result = [ "[", generateExpression(expr.body, {
                                precedence: Precedence.Assignment,
                                allowIn: true,
                                allowCall: true
                            }) ];
                            if (expr.blocks) {
                                for (i = 0, len = expr.blocks.length; i < len; ++i) {
                                    fragment = generateExpression(expr.blocks[i], {
                                        precedence: Precedence.Sequence,
                                        allowIn: true,
                                        allowCall: true
                                    });
                                    result = join(result, fragment);
                                }
                            }
                            if (expr.filter) {
                                result = join(result, "if" + space);
                                fragment = generateExpression(expr.filter, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true
                                });
                                if (extra.moz.parenthesizedComprehensionBlock) {
                                    result = join(result, [ "(", fragment, ")" ]);
                                } else {
                                    result = join(result, fragment);
                                }
                            }
                            result.push("]");
                            break;

                          case Syntax.ComprehensionBlock:
                            if (expr.left.type === Syntax.VariableDeclaration) {
                                fragment = [ expr.left.kind, noEmptySpace(), generateStatement(expr.left.declarations[0], {
                                    allowIn: false
                                }) ];
                            } else {
                                fragment = generateExpression(expr.left, {
                                    precedence: Precedence.Call,
                                    allowIn: true,
                                    allowCall: true
                                });
                            }
                            fragment = join(fragment, expr.of ? "of" : "in");
                            fragment = join(fragment, generateExpression(expr.right, {
                                precedence: Precedence.Sequence,
                                allowIn: true,
                                allowCall: true
                            }));
                            if (extra.moz.parenthesizedComprehensionBlock) {
                                result = [ "for" + space + "(", fragment, ")" ];
                            } else {
                                result = join("for" + space, fragment);
                            }
                            break;

                          default:
                            throw new Error("Unknown expression type: " + expr.type);
                        }
                        return toSourceNode(result, expr);
                    }
                    function generateStatement(stmt, option) {
                        var i, len, result, node, allowIn, functionBody, directiveContext, fragment, semicolon;
                        allowIn = true;
                        semicolon = ";";
                        functionBody = false;
                        directiveContext = false;
                        if (option) {
                            allowIn = option.allowIn === undefined || option.allowIn;
                            if (!semicolons && option.semicolonOptional === true) {
                                semicolon = "";
                            }
                            functionBody = option.functionBody;
                            directiveContext = option.directiveContext;
                        }
                        switch (stmt.type) {
                          case Syntax.BlockStatement:
                            result = [ "{", newline ];
                            withIndent(function() {
                                for (i = 0, len = stmt.body.length; i < len; ++i) {
                                    fragment = addIndent(generateStatement(stmt.body[i], {
                                        semicolonOptional: i === len - 1,
                                        directiveContext: functionBody
                                    }));
                                    result.push(fragment);
                                    if (!endsWithLineTerminator(toSourceNode(fragment).toString())) {
                                        result.push(newline);
                                    }
                                }
                            });
                            result.push(addIndent("}"));
                            break;

                          case Syntax.BreakStatement:
                            if (stmt.label) {
                                result = "break " + stmt.label.name + semicolon;
                            } else {
                                result = "break" + semicolon;
                            }
                            break;

                          case Syntax.ContinueStatement:
                            if (stmt.label) {
                                result = "continue " + stmt.label.name + semicolon;
                            } else {
                                result = "continue" + semicolon;
                            }
                            break;

                          case Syntax.DirectiveStatement:
                            if (stmt.raw) {
                                result = stmt.raw + semicolon;
                            } else {
                                result = escapeDirective(stmt.directive) + semicolon;
                            }
                            break;

                          case Syntax.DoWhileStatement:
                            result = join("do", maybeBlock(stmt.body));
                            result = maybeBlockSuffix(stmt.body, result);
                            result = join(result, [ "while" + space + "(", generateExpression(stmt.test, {
                                precedence: Precedence.Sequence,
                                allowIn: true,
                                allowCall: true
                            }), ")" + semicolon ]);
                            break;

                          case Syntax.CatchClause:
                            withIndent(function() {
                                result = [ "catch" + space + "(", generateExpression(stmt.param, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true
                                }), ")" ];
                            });
                            result.push(maybeBlock(stmt.body));
                            break;

                          case Syntax.DebuggerStatement:
                            result = "debugger" + semicolon;
                            break;

                          case Syntax.EmptyStatement:
                            result = ";";
                            break;

                          case Syntax.ExpressionStatement:
                            result = [ generateExpression(stmt.expression, {
                                precedence: Precedence.Sequence,
                                allowIn: true,
                                allowCall: true
                            }) ];
                            fragment = toSourceNode(result).toString();
                            if (fragment.charAt(0) === "{" || fragment.slice(0, 8) === "function" && " (".indexOf(fragment.charAt(8)) >= 0 || directive && directiveContext && stmt.expression.type === Syntax.Literal && typeof stmt.expression.value === "string") {
                                result = [ "(", result, ")" + semicolon ];
                            } else {
                                result.push(semicolon);
                            }
                            break;

                          case Syntax.VariableDeclarator:
                            if (stmt.init) {
                                result = [ generateExpression(stmt.id, {
                                    precedence: Precedence.Assignment,
                                    allowIn: allowIn,
                                    allowCall: true
                                }), space, "=", space, generateExpression(stmt.init, {
                                    precedence: Precedence.Assignment,
                                    allowIn: allowIn,
                                    allowCall: true
                                }) ];
                            } else {
                                result = generateIdentifier(stmt.id);
                            }
                            break;

                          case Syntax.VariableDeclaration:
                            result = [ stmt.kind ];
                            if (stmt.declarations.length === 1 && stmt.declarations[0].init && stmt.declarations[0].init.type === Syntax.FunctionExpression) {
                                result.push(noEmptySpace(), generateStatement(stmt.declarations[0], {
                                    allowIn: allowIn
                                }));
                            } else {
                                withIndent(function() {
                                    node = stmt.declarations[0];
                                    if (extra.comment && node.leadingComments) {
                                        result.push("\n", addIndent(generateStatement(node, {
                                            allowIn: allowIn
                                        })));
                                    } else {
                                        result.push(noEmptySpace(), generateStatement(node, {
                                            allowIn: allowIn
                                        }));
                                    }
                                    for (i = 1, len = stmt.declarations.length; i < len; ++i) {
                                        node = stmt.declarations[i];
                                        if (extra.comment && node.leadingComments) {
                                            result.push("," + newline, addIndent(generateStatement(node, {
                                                allowIn: allowIn
                                            })));
                                        } else {
                                            result.push("," + space, generateStatement(node, {
                                                allowIn: allowIn
                                            }));
                                        }
                                    }
                                });
                            }
                            result.push(semicolon);
                            break;

                          case Syntax.ThrowStatement:
                            result = [ join("throw", generateExpression(stmt.argument, {
                                precedence: Precedence.Sequence,
                                allowIn: true,
                                allowCall: true
                            })), semicolon ];
                            break;

                          case Syntax.TryStatement:
                            result = [ "try", maybeBlock(stmt.block) ];
                            result = maybeBlockSuffix(stmt.block, result);
                            if (stmt.handlers) {
                                for (i = 0, len = stmt.handlers.length; i < len; ++i) {
                                    result = join(result, generateStatement(stmt.handlers[i]));
                                    if (stmt.finalizer || i + 1 !== len) {
                                        result = maybeBlockSuffix(stmt.handlers[i].body, result);
                                    }
                                }
                            } else {
                                if (stmt.handler) {
                                    result = join(result, generateStatement(stmt.handler));
                                    if (stmt.finalizer || stmt.guardedHandlers.length > 0) {
                                        result = maybeBlockSuffix(stmt.handler.body, result);
                                    }
                                }
                                for (i = 0, len = stmt.guardedHandlers.length; i < len; ++i) {
                                    result = join(result, generateStatement(stmt.guardedHandlers[i]));
                                    if (stmt.finalizer || i + 1 !== len) {
                                        result = maybeBlockSuffix(stmt.guardedHandlers[i].body, result);
                                    }
                                }
                            }
                            if (stmt.finalizer) {
                                result = join(result, [ "finally", maybeBlock(stmt.finalizer) ]);
                            }
                            break;

                          case Syntax.SwitchStatement:
                            withIndent(function() {
                                result = [ "switch" + space + "(", generateExpression(stmt.discriminant, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true
                                }), ")" + space + "{" + newline ];
                            });
                            if (stmt.cases) {
                                for (i = 0, len = stmt.cases.length; i < len; ++i) {
                                    fragment = addIndent(generateStatement(stmt.cases[i], {
                                        semicolonOptional: i === len - 1
                                    }));
                                    result.push(fragment);
                                    if (!endsWithLineTerminator(toSourceNode(fragment).toString())) {
                                        result.push(newline);
                                    }
                                }
                            }
                            result.push(addIndent("}"));
                            break;

                          case Syntax.SwitchCase:
                            withIndent(function() {
                                if (stmt.test) {
                                    result = [ join("case", generateExpression(stmt.test, {
                                        precedence: Precedence.Sequence,
                                        allowIn: true,
                                        allowCall: true
                                    })), ":" ];
                                } else {
                                    result = [ "default:" ];
                                }
                                i = 0;
                                len = stmt.consequent.length;
                                if (len && stmt.consequent[0].type === Syntax.BlockStatement) {
                                    fragment = maybeBlock(stmt.consequent[0]);
                                    result.push(fragment);
                                    i = 1;
                                }
                                if (i !== len && !endsWithLineTerminator(toSourceNode(result).toString())) {
                                    result.push(newline);
                                }
                                for (;i < len; ++i) {
                                    fragment = addIndent(generateStatement(stmt.consequent[i], {
                                        semicolonOptional: i === len - 1 && semicolon === ""
                                    }));
                                    result.push(fragment);
                                    if (i + 1 !== len && !endsWithLineTerminator(toSourceNode(fragment).toString())) {
                                        result.push(newline);
                                    }
                                }
                            });
                            break;

                          case Syntax.IfStatement:
                            withIndent(function() {
                                result = [ "if" + space + "(", generateExpression(stmt.test, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true
                                }), ")" ];
                            });
                            if (stmt.alternate) {
                                result.push(maybeBlock(stmt.consequent));
                                result = maybeBlockSuffix(stmt.consequent, result);
                                if (stmt.alternate.type === Syntax.IfStatement) {
                                    result = join(result, [ "else ", generateStatement(stmt.alternate, {
                                        semicolonOptional: semicolon === ""
                                    }) ]);
                                } else {
                                    result = join(result, join("else", maybeBlock(stmt.alternate, semicolon === "")));
                                }
                            } else {
                                result.push(maybeBlock(stmt.consequent, semicolon === ""));
                            }
                            break;

                          case Syntax.ForStatement:
                            withIndent(function() {
                                result = [ "for" + space + "(" ];
                                if (stmt.init) {
                                    if (stmt.init.type === Syntax.VariableDeclaration) {
                                        result.push(generateStatement(stmt.init, {
                                            allowIn: false
                                        }));
                                    } else {
                                        result.push(generateExpression(stmt.init, {
                                            precedence: Precedence.Sequence,
                                            allowIn: false,
                                            allowCall: true
                                        }), ";");
                                    }
                                } else {
                                    result.push(";");
                                }
                                if (stmt.test) {
                                    result.push(space, generateExpression(stmt.test, {
                                        precedence: Precedence.Sequence,
                                        allowIn: true,
                                        allowCall: true
                                    }), ";");
                                } else {
                                    result.push(";");
                                }
                                if (stmt.update) {
                                    result.push(space, generateExpression(stmt.update, {
                                        precedence: Precedence.Sequence,
                                        allowIn: true,
                                        allowCall: true
                                    }), ")");
                                } else {
                                    result.push(")");
                                }
                            });
                            result.push(maybeBlock(stmt.body, semicolon === ""));
                            break;

                          case Syntax.ForInStatement:
                            result = [ "for" + space + "(" ];
                            withIndent(function() {
                                if (stmt.left.type === Syntax.VariableDeclaration) {
                                    withIndent(function() {
                                        result.push(stmt.left.kind + noEmptySpace(), generateStatement(stmt.left.declarations[0], {
                                            allowIn: false
                                        }));
                                    });
                                } else {
                                    result.push(generateExpression(stmt.left, {
                                        precedence: Precedence.Call,
                                        allowIn: true,
                                        allowCall: true
                                    }));
                                }
                                result = join(result, "in");
                                result = [ join(result, generateExpression(stmt.right, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true
                                })), ")" ];
                            });
                            result.push(maybeBlock(stmt.body, semicolon === ""));
                            break;

                          case Syntax.LabeledStatement:
                            result = [ stmt.label.name + ":", maybeBlock(stmt.body, semicolon === "") ];
                            break;

                          case Syntax.Program:
                            len = stmt.body.length;
                            result = [ safeConcatenation && len > 0 ? "\n" : "" ];
                            for (i = 0; i < len; ++i) {
                                fragment = addIndent(generateStatement(stmt.body[i], {
                                    semicolonOptional: !safeConcatenation && i === len - 1,
                                    directiveContext: true
                                }));
                                result.push(fragment);
                                if (i + 1 < len && !endsWithLineTerminator(toSourceNode(fragment).toString())) {
                                    result.push(newline);
                                }
                            }
                            break;

                          case Syntax.FunctionDeclaration:
                            result = [ stmt.generator && !extra.moz.starlessGenerator ? "function* " : "function ", generateIdentifier(stmt.id), generateFunctionBody(stmt) ];
                            break;

                          case Syntax.ReturnStatement:
                            if (stmt.argument) {
                                result = [ join("return", generateExpression(stmt.argument, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true
                                })), semicolon ];
                            } else {
                                result = [ "return" + semicolon ];
                            }
                            break;

                          case Syntax.WhileStatement:
                            withIndent(function() {
                                result = [ "while" + space + "(", generateExpression(stmt.test, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true
                                }), ")" ];
                            });
                            result.push(maybeBlock(stmt.body, semicolon === ""));
                            break;

                          case Syntax.WithStatement:
                            withIndent(function() {
                                result = [ "with" + space + "(", generateExpression(stmt.object, {
                                    precedence: Precedence.Sequence,
                                    allowIn: true,
                                    allowCall: true
                                }), ")" ];
                            });
                            result.push(maybeBlock(stmt.body, semicolon === ""));
                            break;

                          default:
                            throw new Error("Unknown statement type: " + stmt.type);
                        }
                        if (extra.comment) {
                            result = addCommentsToStatement(stmt, result);
                        }
                        fragment = toSourceNode(result).toString();
                        if (stmt.type === Syntax.Program && !safeConcatenation && newline === "" && fragment.charAt(fragment.length - 1) === "\n") {
                            result = toSourceNode(result).replaceRight(/\s+$/, "");
                        }
                        return toSourceNode(result, stmt);
                    }
                    function generate(node, options) {
                        var defaultOptions = getDefaultOptions(), result, pair;
                        if (options != null) {
                            if (typeof options.indent === "string") {
                                defaultOptions.format.indent.style = options.indent;
                            }
                            if (typeof options.base === "number") {
                                defaultOptions.format.indent.base = options.base;
                            }
                            options = updateDeeply(defaultOptions, options);
                            indent = options.format.indent.style;
                            if (typeof options.base === "string") {
                                base = options.base;
                            } else {
                                base = stringRepeat(indent, options.format.indent.base);
                            }
                        } else {
                            options = defaultOptions;
                            indent = options.format.indent.style;
                            base = stringRepeat(indent, options.format.indent.base);
                        }
                        json = options.format.json;
                        renumber = options.format.renumber;
                        hexadecimal = json ? false : options.format.hexadecimal;
                        quotes = json ? "double" : options.format.quotes;
                        escapeless = options.format.escapeless;
                        newline = options.format.newline;
                        space = options.format.space;
                        if (options.format.compact) {
                            newline = space = indent = base = "";
                        }
                        parentheses = options.format.parentheses;
                        semicolons = options.format.semicolons;
                        safeConcatenation = options.format.safeConcatenation;
                        directive = options.directive;
                        parse = json ? null : options.parse;
                        sourceMap = options.sourceMap;
                        extra = options;
                        if (sourceMap) {
                            if (!exports.browser) {
                                SourceNode = require("/node_modules/source-map/lib/source-map.js", module).SourceNode;
                            } else {
                                SourceNode = global.sourceMap.SourceNode;
                            }
                        } else {
                            SourceNode = SourceNodeMock;
                        }
                        switch (node.type) {
                          case Syntax.BlockStatement:
                          case Syntax.BreakStatement:
                          case Syntax.CatchClause:
                          case Syntax.ContinueStatement:
                          case Syntax.DirectiveStatement:
                          case Syntax.DoWhileStatement:
                          case Syntax.DebuggerStatement:
                          case Syntax.EmptyStatement:
                          case Syntax.ExpressionStatement:
                          case Syntax.ForStatement:
                          case Syntax.ForInStatement:
                          case Syntax.FunctionDeclaration:
                          case Syntax.IfStatement:
                          case Syntax.LabeledStatement:
                          case Syntax.Program:
                          case Syntax.ReturnStatement:
                          case Syntax.SwitchStatement:
                          case Syntax.SwitchCase:
                          case Syntax.ThrowStatement:
                          case Syntax.TryStatement:
                          case Syntax.VariableDeclaration:
                          case Syntax.VariableDeclarator:
                          case Syntax.WhileStatement:
                          case Syntax.WithStatement:
                            result = generateStatement(node);
                            break;

                          case Syntax.AssignmentExpression:
                          case Syntax.ArrayExpression:
                          case Syntax.ArrayPattern:
                          case Syntax.BinaryExpression:
                          case Syntax.CallExpression:
                          case Syntax.ConditionalExpression:
                          case Syntax.FunctionExpression:
                          case Syntax.Identifier:
                          case Syntax.Literal:
                          case Syntax.LogicalExpression:
                          case Syntax.MemberExpression:
                          case Syntax.NewExpression:
                          case Syntax.ObjectExpression:
                          case Syntax.ObjectPattern:
                          case Syntax.Property:
                          case Syntax.SequenceExpression:
                          case Syntax.ThisExpression:
                          case Syntax.UnaryExpression:
                          case Syntax.UpdateExpression:
                          case Syntax.YieldExpression:
                            result = generateExpression(node, {
                                precedence: Precedence.Sequence,
                                allowIn: true,
                                allowCall: true
                            });
                            break;

                          default:
                            throw new Error("Unknown node type: " + node.type);
                        }
                        if (!sourceMap) {
                            return result.toString();
                        }
                        pair = result.toStringWithSourceMap({
                            file: options.file,
                            sourceRoot: options.sourceMapRoot
                        });
                        if (options.sourceContent) {
                            pair.map.setSourceContent(options.sourceMap, options.sourceContent);
                        }
                        if (options.sourceMapWithCode) {
                            return pair;
                        }
                        return pair.map.toString();
                    }
                    FORMAT_MINIFY = {
                        indent: {
                            style: "",
                            base: 0
                        },
                        renumber: true,
                        hexadecimal: true,
                        quotes: "auto",
                        escapeless: true,
                        compact: true,
                        parentheses: false,
                        semicolons: false
                    };
                    FORMAT_DEFAULTS = getDefaultOptions().format;
                    exports.version = require("/package.json", module).version;
                    exports.generate = generate;
                    exports.attachComments = estraverse.attachComments;
                    exports.browser = false;
                    exports.FORMAT_MINIFY = FORMAT_MINIFY;
                    exports.FORMAT_DEFAULTS = FORMAT_DEFAULTS;
                })();
            });
            require.define("/package.json", function(module, exports, __dirname, __filename) {
                module.exports = {
                    name: "escodegen",
                    description: "ECMAScript code generator",
                    homepage: "http://github.com/Constellation/escodegen",
                    main: "escodegen.js",
                    bin: {
                        esgenerate: "./bin/esgenerate.js",
                        escodegen: "./bin/escodegen.js"
                    },
                    version: "1.0.1-dev",
                    engines: {
                        node: ">=0.4.0"
                    },
                    maintainers: [ {
                        name: "Yusuke Suzuki",
                        email: "utatane.tea@gmail.com",
                        web: "http://github.com/Constellation"
                    } ],
                    repository: {
                        type: "git",
                        url: "http://github.com/Constellation/escodegen.git"
                    },
                    dependencies: {
                        esprima: "~1.0.4",
                        estraverse: "~1.5.0",
                        esutils: "~1.0.0"
                    },
                    optionalDependencies: {
                        "source-map": "~0.1.30"
                    },
                    devDependencies: {
                        "esprima-moz": "*",
                        "commonjs-everywhere": "~0.8.0",
                        q: "*",
                        bower: "*",
                        semver: "*",
                        chai: "~1.7.2",
                        "grunt-contrib-jshint": "~0.5.0",
                        "grunt-cli": "~0.1.9",
                        grunt: "~0.4.1",
                        "grunt-mocha-test": "~0.6.2"
                    },
                    licenses: [ {
                        type: "BSD",
                        url: "http://github.com/Constellation/escodegen/raw/master/LICENSE.BSD"
                    } ],
                    scripts: {
                        test: "grunt travis",
                        "unit-test": "grunt test",
                        lint: "grunt lint",
                        release: "node tools/release.js",
                        "build-min": "./node_modules/.bin/cjsify -ma path: tools/entry-point.js > escodegen.browser.min.js",
                        build: "./node_modules/.bin/cjsify -a path: tools/entry-point.js > escodegen.browser.js"
                    }
                };
            });
            require.define("/node_modules/source-map/lib/source-map.js", function(module, exports, __dirname, __filename) {
                exports.SourceMapGenerator = require("/node_modules/source-map/lib/source-map/source-map-generator.js", module).SourceMapGenerator;
                exports.SourceMapConsumer = require("/node_modules/source-map/lib/source-map/source-map-consumer.js", module).SourceMapConsumer;
                exports.SourceNode = require("/node_modules/source-map/lib/source-map/source-node.js", module).SourceNode;
            });
            require.define("/node_modules/source-map/lib/source-map/source-node.js", function(module, exports, __dirname, __filename) {
                if (typeof define !== "function") {
                    var define = require("/node_modules/source-map/node_modules/amdefine/amdefine.js", module)(module, require);
                }
                define(function(require, exports, module) {
                    var SourceMapGenerator = require("/node_modules/source-map/lib/source-map/source-map-generator.js", module).SourceMapGenerator;
                    var util = require("/node_modules/source-map/lib/source-map/util.js", module);
                    function SourceNode(aLine, aColumn, aSource, aChunks, aName) {
                        this.children = [];
                        this.sourceContents = {};
                        this.line = aLine === undefined ? null : aLine;
                        this.column = aColumn === undefined ? null : aColumn;
                        this.source = aSource === undefined ? null : aSource;
                        this.name = aName === undefined ? null : aName;
                        if (aChunks != null) {
                            this.add(aChunks);
                        }
                    }
                    SourceNode.fromStringWithSourceMap = function SourceNode_fromStringWithSourceMap(aGeneratedCode, aSourceMapConsumer) {
                        var node = new SourceNode();
                        var remainingLines = aGeneratedCode.split("\n");
                        var lastGeneratedLine = 1, lastGeneratedColumn = 0;
                        var lastMapping = null;
                        aSourceMapConsumer.eachMapping(function(mapping) {
                            if (lastMapping === null) {
                                while (lastGeneratedLine < mapping.generatedLine) {
                                    node.add(remainingLines.shift() + "\n");
                                    lastGeneratedLine++;
                                }
                                if (lastGeneratedColumn < mapping.generatedColumn) {
                                    var nextLine = remainingLines[0];
                                    node.add(nextLine.substr(0, mapping.generatedColumn));
                                    remainingLines[0] = nextLine.substr(mapping.generatedColumn);
                                    lastGeneratedColumn = mapping.generatedColumn;
                                }
                            } else {
                                if (lastGeneratedLine < mapping.generatedLine) {
                                    var code = "";
                                    do {
                                        code += remainingLines.shift() + "\n";
                                        lastGeneratedLine++;
                                        lastGeneratedColumn = 0;
                                    } while (lastGeneratedLine < mapping.generatedLine);
                                    if (lastGeneratedColumn < mapping.generatedColumn) {
                                        var nextLine = remainingLines[0];
                                        code += nextLine.substr(0, mapping.generatedColumn);
                                        remainingLines[0] = nextLine.substr(mapping.generatedColumn);
                                        lastGeneratedColumn = mapping.generatedColumn;
                                    }
                                    addMappingWithCode(lastMapping, code);
                                } else {
                                    var nextLine = remainingLines[0];
                                    var code = nextLine.substr(0, mapping.generatedColumn - lastGeneratedColumn);
                                    remainingLines[0] = nextLine.substr(mapping.generatedColumn - lastGeneratedColumn);
                                    lastGeneratedColumn = mapping.generatedColumn;
                                    addMappingWithCode(lastMapping, code);
                                }
                            }
                            lastMapping = mapping;
                        }, this);
                        addMappingWithCode(lastMapping, remainingLines.join("\n"));
                        aSourceMapConsumer.sources.forEach(function(sourceFile) {
                            var content = aSourceMapConsumer.sourceContentFor(sourceFile);
                            if (content) {
                                node.setSourceContent(sourceFile, content);
                            }
                        });
                        return node;
                        function addMappingWithCode(mapping, code) {
                            if (mapping === null || mapping.source === undefined) {
                                node.add(code);
                            } else {
                                node.add(new SourceNode(mapping.originalLine, mapping.originalColumn, mapping.source, code, mapping.name));
                            }
                        }
                    };
                    SourceNode.prototype.add = function SourceNode_add(aChunk) {
                        if (Array.isArray(aChunk)) {
                            aChunk.forEach(function(chunk) {
                                this.add(chunk);
                            }, this);
                        } else {
                            if (aChunk instanceof SourceNode || typeof aChunk === "string") {
                                if (aChunk) {
                                    this.children.push(aChunk);
                                }
                            } else {
                                throw new TypeError("Expected a SourceNode, string, or an array of SourceNodes and strings. Got " + aChunk);
                            }
                        }
                        return this;
                    };
                    SourceNode.prototype.prepend = function SourceNode_prepend(aChunk) {
                        if (Array.isArray(aChunk)) {
                            for (var i = aChunk.length - 1; i >= 0; i--) {
                                this.prepend(aChunk[i]);
                            }
                        } else {
                            if (aChunk instanceof SourceNode || typeof aChunk === "string") {
                                this.children.unshift(aChunk);
                            } else {
                                throw new TypeError("Expected a SourceNode, string, or an array of SourceNodes and strings. Got " + aChunk);
                            }
                        }
                        return this;
                    };
                    SourceNode.prototype.walk = function SourceNode_walk(aFn) {
                        var chunk;
                        for (var i = 0, len = this.children.length; i < len; i++) {
                            chunk = this.children[i];
                            if (chunk instanceof SourceNode) {
                                chunk.walk(aFn);
                            } else {
                                if (chunk !== "") {
                                    aFn(chunk, {
                                        source: this.source,
                                        line: this.line,
                                        column: this.column,
                                        name: this.name
                                    });
                                }
                            }
                        }
                    };
                    SourceNode.prototype.join = function SourceNode_join(aSep) {
                        var newChildren;
                        var i;
                        var len = this.children.length;
                        if (len > 0) {
                            newChildren = [];
                            for (i = 0; i < len - 1; i++) {
                                newChildren.push(this.children[i]);
                                newChildren.push(aSep);
                            }
                            newChildren.push(this.children[i]);
                            this.children = newChildren;
                        }
                        return this;
                    };
                    SourceNode.prototype.replaceRight = function SourceNode_replaceRight(aPattern, aReplacement) {
                        var lastChild = this.children[this.children.length - 1];
                        if (lastChild instanceof SourceNode) {
                            lastChild.replaceRight(aPattern, aReplacement);
                        } else {
                            if (typeof lastChild === "string") {
                                this.children[this.children.length - 1] = lastChild.replace(aPattern, aReplacement);
                            } else {
                                this.children.push("".replace(aPattern, aReplacement));
                            }
                        }
                        return this;
                    };
                    SourceNode.prototype.setSourceContent = function SourceNode_setSourceContent(aSourceFile, aSourceContent) {
                        this.sourceContents[util.toSetString(aSourceFile)] = aSourceContent;
                    };
                    SourceNode.prototype.walkSourceContents = function SourceNode_walkSourceContents(aFn) {
                        for (var i = 0, len = this.children.length; i < len; i++) {
                            if (this.children[i] instanceof SourceNode) {
                                this.children[i].walkSourceContents(aFn);
                            }
                        }
                        var sources = Object.keys(this.sourceContents);
                        for (var i = 0, len = sources.length; i < len; i++) {
                            aFn(util.fromSetString(sources[i]), this.sourceContents[sources[i]]);
                        }
                    };
                    SourceNode.prototype.toString = function SourceNode_toString() {
                        var str = "";
                        this.walk(function(chunk) {
                            str += chunk;
                        });
                        return str;
                    };
                    SourceNode.prototype.toStringWithSourceMap = function SourceNode_toStringWithSourceMap(aArgs) {
                        var generated = {
                            code: "",
                            line: 1,
                            column: 0
                        };
                        var map = new SourceMapGenerator(aArgs);
                        var sourceMappingActive = false;
                        var lastOriginalSource = null;
                        var lastOriginalLine = null;
                        var lastOriginalColumn = null;
                        var lastOriginalName = null;
                        this.walk(function(chunk, original) {
                            generated.code += chunk;
                            if (original.source !== null && original.line !== null && original.column !== null) {
                                if (lastOriginalSource !== original.source || lastOriginalLine !== original.line || lastOriginalColumn !== original.column || lastOriginalName !== original.name) {
                                    map.addMapping({
                                        source: original.source,
                                        original: {
                                            line: original.line,
                                            column: original.column
                                        },
                                        generated: {
                                            line: generated.line,
                                            column: generated.column
                                        },
                                        name: original.name
                                    });
                                }
                                lastOriginalSource = original.source;
                                lastOriginalLine = original.line;
                                lastOriginalColumn = original.column;
                                lastOriginalName = original.name;
                                sourceMappingActive = true;
                            } else {
                                if (sourceMappingActive) {
                                    map.addMapping({
                                        generated: {
                                            line: generated.line,
                                            column: generated.column
                                        }
                                    });
                                    lastOriginalSource = null;
                                    sourceMappingActive = false;
                                }
                            }
                            chunk.split("").forEach(function(ch) {
                                if (ch === "\n") {
                                    generated.line++;
                                    generated.column = 0;
                                } else {
                                    generated.column++;
                                }
                            });
                        });
                        this.walkSourceContents(function(sourceFile, sourceContent) {
                            map.setSourceContent(sourceFile, sourceContent);
                        });
                        return {
                            code: generated.code,
                            map: map
                        };
                    };
                    exports.SourceNode = SourceNode;
                });
            });
            require.define("/node_modules/source-map/lib/source-map/util.js", function(module, exports, __dirname, __filename) {
                if (typeof define !== "function") {
                    var define = require("/node_modules/source-map/node_modules/amdefine/amdefine.js", module)(module, require);
                }
                define(function(require, exports, module) {
                    function getArg(aArgs, aName, aDefaultValue) {
                        if (aName in aArgs) {
                            return aArgs[aName];
                        } else {
                            if (arguments.length === 3) {
                                return aDefaultValue;
                            } else {
                                throw new Error('"' + aName + '" is a required argument.');
                            }
                        }
                    }
                    exports.getArg = getArg;
                    var urlRegexp = /([\w+\-.]+):\/\/((\w+:\w+)@)?([\w.]+)?(:(\d+))?(\S+)?/;
                    var dataUrlRegexp = /^data:.+\,.+/;
                    function urlParse(aUrl) {
                        var match = aUrl.match(urlRegexp);
                        if (!match) {
                            return null;
                        }
                        return {
                            scheme: match[1],
                            auth: match[3],
                            host: match[4],
                            port: match[6],
                            path: match[7]
                        };
                    }
                    exports.urlParse = urlParse;
                    function urlGenerate(aParsedUrl) {
                        var url = aParsedUrl.scheme + "://";
                        if (aParsedUrl.auth) {
                            url += aParsedUrl.auth + "@";
                        }
                        if (aParsedUrl.host) {
                            url += aParsedUrl.host;
                        }
                        if (aParsedUrl.port) {
                            url += ":" + aParsedUrl.port;
                        }
                        if (aParsedUrl.path) {
                            url += aParsedUrl.path;
                        }
                        return url;
                    }
                    exports.urlGenerate = urlGenerate;
                    function join(aRoot, aPath) {
                        var url;
                        if (aPath.match(urlRegexp) || aPath.match(dataUrlRegexp)) {
                            return aPath;
                        }
                        if (aPath.charAt(0) === "/" && (url = urlParse(aRoot))) {
                            url.path = aPath;
                            return urlGenerate(url);
                        }
                        return aRoot.replace(/\/$/, "") + "/" + aPath;
                    }
                    exports.join = join;
                    function toSetString(aStr) {
                        return "$" + aStr;
                    }
                    exports.toSetString = toSetString;
                    function fromSetString(aStr) {
                        return aStr.substr(1);
                    }
                    exports.fromSetString = fromSetString;
                    function relative(aRoot, aPath) {
                        aRoot = aRoot.replace(/\/$/, "");
                        var url = urlParse(aRoot);
                        if (aPath.charAt(0) == "/" && url && url.path == "/") {
                            return aPath.slice(1);
                        }
                        return aPath.indexOf(aRoot + "/") === 0 ? aPath.substr(aRoot.length + 1) : aPath;
                    }
                    exports.relative = relative;
                    function strcmp(aStr1, aStr2) {
                        var s1 = aStr1 || "";
                        var s2 = aStr2 || "";
                        return (s1 > s2) - (s1 < s2);
                    }
                    function compareByOriginalPositions(mappingA, mappingB, onlyCompareOriginal) {
                        var cmp;
                        cmp = strcmp(mappingA.source, mappingB.source);
                        if (cmp) {
                            return cmp;
                        }
                        cmp = mappingA.originalLine - mappingB.originalLine;
                        if (cmp) {
                            return cmp;
                        }
                        cmp = mappingA.originalColumn - mappingB.originalColumn;
                        if (cmp || onlyCompareOriginal) {
                            return cmp;
                        }
                        cmp = strcmp(mappingA.name, mappingB.name);
                        if (cmp) {
                            return cmp;
                        }
                        cmp = mappingA.generatedLine - mappingB.generatedLine;
                        if (cmp) {
                            return cmp;
                        }
                        return mappingA.generatedColumn - mappingB.generatedColumn;
                    }
                    exports.compareByOriginalPositions = compareByOriginalPositions;
                    function compareByGeneratedPositions(mappingA, mappingB, onlyCompareGenerated) {
                        var cmp;
                        cmp = mappingA.generatedLine - mappingB.generatedLine;
                        if (cmp) {
                            return cmp;
                        }
                        cmp = mappingA.generatedColumn - mappingB.generatedColumn;
                        if (cmp || onlyCompareGenerated) {
                            return cmp;
                        }
                        cmp = strcmp(mappingA.source, mappingB.source);
                        if (cmp) {
                            return cmp;
                        }
                        cmp = mappingA.originalLine - mappingB.originalLine;
                        if (cmp) {
                            return cmp;
                        }
                        cmp = mappingA.originalColumn - mappingB.originalColumn;
                        if (cmp) {
                            return cmp;
                        }
                        return strcmp(mappingA.name, mappingB.name);
                    }
                    exports.compareByGeneratedPositions = compareByGeneratedPositions;
                });
            });
            require.define("/node_modules/source-map/node_modules/amdefine/amdefine.js", function(module, exports, __dirname, __filename) {
                "use strict";
                function amdefine(module, requireFn) {
                    "use strict";
                    var defineCache = {}, loaderCache = {}, alreadyCalled = false, path = require("path", module), makeRequire, stringRequire;
                    function trimDots(ary) {
                        var i, part;
                        for (i = 0; ary[i]; i += 1) {
                            part = ary[i];
                            if (part === ".") {
                                ary.splice(i, 1);
                                i -= 1;
                            } else {
                                if (part === "..") {
                                    if (i === 1 && (ary[2] === ".." || ary[0] === "..")) {
                                        break;
                                    } else {
                                        if (i > 0) {
                                            ary.splice(i - 1, 2);
                                            i -= 2;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    function normalize(name, baseName) {
                        var baseParts;
                        if (name && name.charAt(0) === ".") {
                            if (baseName) {
                                baseParts = baseName.split("/");
                                baseParts = baseParts.slice(0, baseParts.length - 1);
                                baseParts = baseParts.concat(name.split("/"));
                                trimDots(baseParts);
                                name = baseParts.join("/");
                            }
                        }
                        return name;
                    }
                    function makeNormalize(relName) {
                        return function(name) {
                            return normalize(name, relName);
                        };
                    }
                    function makeLoad(id) {
                        function load(value) {
                            loaderCache[id] = value;
                        }
                        load.fromText = function(id, text) {
                            throw new Error("amdefine does not implement load.fromText");
                        };
                        return load;
                    }
                    makeRequire = function(systemRequire, exports, module, relId) {
                        function amdRequire(deps, callback) {
                            if (typeof deps === "string") {
                                return stringRequire(systemRequire, exports, module, deps, relId);
                            } else {
                                deps = deps.map(function(depName) {
                                    return stringRequire(systemRequire, exports, module, depName, relId);
                                });
                                process.nextTick(function() {
                                    callback.apply(null, deps);
                                });
                            }
                        }
                        amdRequire.toUrl = function(filePath) {
                            if (filePath.indexOf(".") === 0) {
                                return normalize(filePath, path.dirname(module.filename));
                            } else {
                                return filePath;
                            }
                        };
                        return amdRequire;
                    };
                    requireFn = requireFn || function req() {
                        return module.require.apply(module, arguments);
                    };
                    function runFactory(id, deps, factory) {
                        var r, e, m, result;
                        if (id) {
                            e = loaderCache[id] = {};
                            m = {
                                id: id,
                                uri: __filename,
                                exports: e
                            };
                            r = makeRequire(requireFn, e, m, id);
                        } else {
                            if (alreadyCalled) {
                                throw new Error("amdefine with no module ID cannot be called more than once per file.");
                            }
                            alreadyCalled = true;
                            e = module.exports;
                            m = module;
                            r = makeRequire(requireFn, e, m, module.id);
                        }
                        if (deps) {
                            deps = deps.map(function(depName) {
                                return r(depName);
                            });
                        }
                        if (typeof factory === "function") {
                            result = factory.apply(m.exports, deps);
                        } else {
                            result = factory;
                        }
                        if (result !== undefined) {
                            m.exports = result;
                            if (id) {
                                loaderCache[id] = m.exports;
                            }
                        }
                    }
                    stringRequire = function(systemRequire, exports, module, id, relId) {
                        var index = id.indexOf("!"), originalId = id, prefix, plugin;
                        if (index === -1) {
                            id = normalize(id, relId);
                            if (id === "require") {
                                return makeRequire(systemRequire, exports, module, relId);
                            } else {
                                if (id === "exports") {
                                    return exports;
                                } else {
                                    if (id === "module") {
                                        return module;
                                    } else {
                                        if (loaderCache.hasOwnProperty(id)) {
                                            return loaderCache[id];
                                        } else {
                                            if (defineCache[id]) {
                                                runFactory.apply(null, defineCache[id]);
                                                return loaderCache[id];
                                            } else {
                                                if (systemRequire) {
                                                    return systemRequire(originalId);
                                                } else {
                                                    throw new Error("No module with ID: " + id);
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        } else {
                            prefix = id.substring(0, index);
                            id = id.substring(index + 1, id.length);
                            plugin = stringRequire(systemRequire, exports, module, prefix, relId);
                            if (plugin.normalize) {
                                id = plugin.normalize(id, makeNormalize(relId));
                            } else {
                                id = normalize(id, relId);
                            }
                            if (loaderCache[id]) {
                                return loaderCache[id];
                            } else {
                                plugin.load(id, makeRequire(systemRequire, exports, module, relId), makeLoad(id), {});
                                return loaderCache[id];
                            }
                        }
                    };
                    function define(id, deps, factory) {
                        if (Array.isArray(id)) {
                            factory = deps;
                            deps = id;
                            id = undefined;
                        } else {
                            if (typeof id !== "string") {
                                factory = id;
                                id = deps = undefined;
                            }
                        }
                        if (deps && !Array.isArray(deps)) {
                            factory = deps;
                            deps = undefined;
                        }
                        if (!deps) {
                            deps = [ "require", "exports", "module" ];
                        }
                        if (id) {
                            defineCache[id] = [ id, deps, factory ];
                        } else {
                            runFactory(id, deps, factory);
                        }
                    }
                    define.require = function(id) {
                        if (loaderCache[id]) {
                            return loaderCache[id];
                        }
                        if (defineCache[id]) {
                            runFactory.apply(null, defineCache[id]);
                            return loaderCache[id];
                        }
                    };
                    define.amd = {};
                    return define;
                }
                module.exports = amdefine;
            });
            require.define("/node_modules/source-map/lib/source-map/source-map-generator.js", function(module, exports, __dirname, __filename) {
                if (typeof define !== "function") {
                    var define = require("/node_modules/source-map/node_modules/amdefine/amdefine.js", module)(module, require);
                }
                define(function(require, exports, module) {
                    var base64VLQ = require("/node_modules/source-map/lib/source-map/base64-vlq.js", module);
                    var util = require("/node_modules/source-map/lib/source-map/util.js", module);
                    var ArraySet = require("/node_modules/source-map/lib/source-map/array-set.js", module).ArraySet;
                    function SourceMapGenerator(aArgs) {
                        this._file = util.getArg(aArgs, "file");
                        this._sourceRoot = util.getArg(aArgs, "sourceRoot", null);
                        this._sources = new ArraySet();
                        this._names = new ArraySet();
                        this._mappings = [];
                        this._sourcesContents = null;
                    }
                    SourceMapGenerator.prototype._version = 3;
                    SourceMapGenerator.fromSourceMap = function SourceMapGenerator_fromSourceMap(aSourceMapConsumer) {
                        var sourceRoot = aSourceMapConsumer.sourceRoot;
                        var generator = new SourceMapGenerator({
                            file: aSourceMapConsumer.file,
                            sourceRoot: sourceRoot
                        });
                        aSourceMapConsumer.eachMapping(function(mapping) {
                            var newMapping = {
                                generated: {
                                    line: mapping.generatedLine,
                                    column: mapping.generatedColumn
                                }
                            };
                            if (mapping.source) {
                                newMapping.source = mapping.source;
                                if (sourceRoot) {
                                    newMapping.source = util.relative(sourceRoot, newMapping.source);
                                }
                                newMapping.original = {
                                    line: mapping.originalLine,
                                    column: mapping.originalColumn
                                };
                                if (mapping.name) {
                                    newMapping.name = mapping.name;
                                }
                            }
                            generator.addMapping(newMapping);
                        });
                        aSourceMapConsumer.sources.forEach(function(sourceFile) {
                            var content = aSourceMapConsumer.sourceContentFor(sourceFile);
                            if (content) {
                                generator.setSourceContent(sourceFile, content);
                            }
                        });
                        return generator;
                    };
                    SourceMapGenerator.prototype.addMapping = function SourceMapGenerator_addMapping(aArgs) {
                        var generated = util.getArg(aArgs, "generated");
                        var original = util.getArg(aArgs, "original", null);
                        var source = util.getArg(aArgs, "source", null);
                        var name = util.getArg(aArgs, "name", null);
                        this._validateMapping(generated, original, source, name);
                        if (source && !this._sources.has(source)) {
                            this._sources.add(source);
                        }
                        if (name && !this._names.has(name)) {
                            this._names.add(name);
                        }
                        this._mappings.push({
                            generatedLine: generated.line,
                            generatedColumn: generated.column,
                            originalLine: original != null && original.line,
                            originalColumn: original != null && original.column,
                            source: source,
                            name: name
                        });
                    };
                    SourceMapGenerator.prototype.setSourceContent = function SourceMapGenerator_setSourceContent(aSourceFile, aSourceContent) {
                        var source = aSourceFile;
                        if (this._sourceRoot) {
                            source = util.relative(this._sourceRoot, source);
                        }
                        if (aSourceContent !== null) {
                            if (!this._sourcesContents) {
                                this._sourcesContents = {};
                            }
                            this._sourcesContents[util.toSetString(source)] = aSourceContent;
                        } else {
                            delete this._sourcesContents[util.toSetString(source)];
                            if (Object.keys(this._sourcesContents).length === 0) {
                                this._sourcesContents = null;
                            }
                        }
                    };
                    SourceMapGenerator.prototype.applySourceMap = function SourceMapGenerator_applySourceMap(aSourceMapConsumer, aSourceFile) {
                        if (!aSourceFile) {
                            aSourceFile = aSourceMapConsumer.file;
                        }
                        var sourceRoot = this._sourceRoot;
                        if (sourceRoot) {
                            aSourceFile = util.relative(sourceRoot, aSourceFile);
                        }
                        var newSources = new ArraySet();
                        var newNames = new ArraySet();
                        this._mappings.forEach(function(mapping) {
                            if (mapping.source === aSourceFile && mapping.originalLine) {
                                var original = aSourceMapConsumer.originalPositionFor({
                                    line: mapping.originalLine,
                                    column: mapping.originalColumn
                                });
                                if (original.source !== null) {
                                    if (sourceRoot) {
                                        mapping.source = util.relative(sourceRoot, original.source);
                                    } else {
                                        mapping.source = original.source;
                                    }
                                    mapping.originalLine = original.line;
                                    mapping.originalColumn = original.column;
                                    if (original.name !== null && mapping.name !== null) {
                                        mapping.name = original.name;
                                    }
                                }
                            }
                            var source = mapping.source;
                            if (source && !newSources.has(source)) {
                                newSources.add(source);
                            }
                            var name = mapping.name;
                            if (name && !newNames.has(name)) {
                                newNames.add(name);
                            }
                        }, this);
                        this._sources = newSources;
                        this._names = newNames;
                        aSourceMapConsumer.sources.forEach(function(sourceFile) {
                            var content = aSourceMapConsumer.sourceContentFor(sourceFile);
                            if (content) {
                                if (sourceRoot) {
                                    sourceFile = util.relative(sourceRoot, sourceFile);
                                }
                                this.setSourceContent(sourceFile, content);
                            }
                        }, this);
                    };
                    SourceMapGenerator.prototype._validateMapping = function SourceMapGenerator_validateMapping(aGenerated, aOriginal, aSource, aName) {
                        if (aGenerated && "line" in aGenerated && "column" in aGenerated && aGenerated.line > 0 && aGenerated.column >= 0 && !aOriginal && !aSource && !aName) {
                            return;
                        } else {
                            if (aGenerated && "line" in aGenerated && "column" in aGenerated && aOriginal && "line" in aOriginal && "column" in aOriginal && aGenerated.line > 0 && aGenerated.column >= 0 && aOriginal.line > 0 && aOriginal.column >= 0 && aSource) {
                                return;
                            } else {
                                throw new Error("Invalid mapping: " + JSON.stringify({
                                    generated: aGenerated,
                                    source: aSource,
                                    orginal: aOriginal,
                                    name: aName
                                }));
                            }
                        }
                    };
                    SourceMapGenerator.prototype._serializeMappings = function SourceMapGenerator_serializeMappings() {
                        var previousGeneratedColumn = 0;
                        var previousGeneratedLine = 1;
                        var previousOriginalColumn = 0;
                        var previousOriginalLine = 0;
                        var previousName = 0;
                        var previousSource = 0;
                        var result = "";
                        var mapping;
                        this._mappings.sort(util.compareByGeneratedPositions);
                        for (var i = 0, len = this._mappings.length; i < len; i++) {
                            mapping = this._mappings[i];
                            if (mapping.generatedLine !== previousGeneratedLine) {
                                previousGeneratedColumn = 0;
                                while (mapping.generatedLine !== previousGeneratedLine) {
                                    result += ";";
                                    previousGeneratedLine++;
                                }
                            } else {
                                if (i > 0) {
                                    if (!util.compareByGeneratedPositions(mapping, this._mappings[i - 1])) {
                                        continue;
                                    }
                                    result += ",";
                                }
                            }
                            result += base64VLQ.encode(mapping.generatedColumn - previousGeneratedColumn);
                            previousGeneratedColumn = mapping.generatedColumn;
                            if (mapping.source) {
                                result += base64VLQ.encode(this._sources.indexOf(mapping.source) - previousSource);
                                previousSource = this._sources.indexOf(mapping.source);
                                result += base64VLQ.encode(mapping.originalLine - 1 - previousOriginalLine);
                                previousOriginalLine = mapping.originalLine - 1;
                                result += base64VLQ.encode(mapping.originalColumn - previousOriginalColumn);
                                previousOriginalColumn = mapping.originalColumn;
                                if (mapping.name) {
                                    result += base64VLQ.encode(this._names.indexOf(mapping.name) - previousName);
                                    previousName = this._names.indexOf(mapping.name);
                                }
                            }
                        }
                        return result;
                    };
                    SourceMapGenerator.prototype._generateSourcesContent = function SourceMapGenerator_generateSourcesContent(aSources, aSourceRoot) {
                        return aSources.map(function(source) {
                            if (!this._sourcesContents) {
                                return null;
                            }
                            if (aSourceRoot) {
                                source = util.relative(aSourceRoot, source);
                            }
                            var key = util.toSetString(source);
                            return Object.prototype.hasOwnProperty.call(this._sourcesContents, key) ? this._sourcesContents[key] : null;
                        }, this);
                    };
                    SourceMapGenerator.prototype.toJSON = function SourceMapGenerator_toJSON() {
                        var map = {
                            version: this._version,
                            file: this._file,
                            sources: this._sources.toArray(),
                            names: this._names.toArray(),
                            mappings: this._serializeMappings()
                        };
                        if (this._sourceRoot) {
                            map.sourceRoot = this._sourceRoot;
                        }
                        if (this._sourcesContents) {
                            map.sourcesContent = this._generateSourcesContent(map.sources, map.sourceRoot);
                        }
                        return map;
                    };
                    SourceMapGenerator.prototype.toString = function SourceMapGenerator_toString() {
                        return JSON.stringify(this);
                    };
                    exports.SourceMapGenerator = SourceMapGenerator;
                });
            });
            require.define("/node_modules/source-map/lib/source-map/array-set.js", function(module, exports, __dirname, __filename) {
                if (typeof define !== "function") {
                    var define = require("/node_modules/source-map/node_modules/amdefine/amdefine.js", module)(module, require);
                }
                define(function(require, exports, module) {
                    var util = require("/node_modules/source-map/lib/source-map/util.js", module);
                    function ArraySet() {
                        this._array = [];
                        this._set = {};
                    }
                    ArraySet.fromArray = function ArraySet_fromArray(aArray, aAllowDuplicates) {
                        var set = new ArraySet();
                        for (var i = 0, len = aArray.length; i < len; i++) {
                            set.add(aArray[i], aAllowDuplicates);
                        }
                        return set;
                    };
                    ArraySet.prototype.add = function ArraySet_add(aStr, aAllowDuplicates) {
                        var isDuplicate = this.has(aStr);
                        var idx = this._array.length;
                        if (!isDuplicate || aAllowDuplicates) {
                            this._array.push(aStr);
                        }
                        if (!isDuplicate) {
                            this._set[util.toSetString(aStr)] = idx;
                        }
                    };
                    ArraySet.prototype.has = function ArraySet_has(aStr) {
                        return Object.prototype.hasOwnProperty.call(this._set, util.toSetString(aStr));
                    };
                    ArraySet.prototype.indexOf = function ArraySet_indexOf(aStr) {
                        if (this.has(aStr)) {
                            return this._set[util.toSetString(aStr)];
                        }
                        throw new Error('"' + aStr + '" is not in the set.');
                    };
                    ArraySet.prototype.at = function ArraySet_at(aIdx) {
                        if (aIdx >= 0 && aIdx < this._array.length) {
                            return this._array[aIdx];
                        }
                        throw new Error("No element indexed by " + aIdx);
                    };
                    ArraySet.prototype.toArray = function ArraySet_toArray() {
                        return this._array.slice();
                    };
                    exports.ArraySet = ArraySet;
                });
            });
            require.define("/node_modules/source-map/lib/source-map/base64-vlq.js", function(module, exports, __dirname, __filename) {
                if (typeof define !== "function") {
                    var define = require("/node_modules/source-map/node_modules/amdefine/amdefine.js", module)(module, require);
                }
                define(function(require, exports, module) {
                    var base64 = require("/node_modules/source-map/lib/source-map/base64.js", module);
                    var VLQ_BASE_SHIFT = 5;
                    var VLQ_BASE = 1 << VLQ_BASE_SHIFT;
                    var VLQ_BASE_MASK = VLQ_BASE - 1;
                    var VLQ_CONTINUATION_BIT = VLQ_BASE;
                    function toVLQSigned(aValue) {
                        return aValue < 0 ? (-aValue << 1) + 1 : (aValue << 1) + 0;
                    }
                    function fromVLQSigned(aValue) {
                        var isNegative = (aValue & 1) === 1;
                        var shifted = aValue >> 1;
                        return isNegative ? -shifted : shifted;
                    }
                    exports.encode = function base64VLQ_encode(aValue) {
                        var encoded = "";
                        var digit;
                        var vlq = toVLQSigned(aValue);
                        do {
                            digit = vlq & VLQ_BASE_MASK;
                            vlq >>>= VLQ_BASE_SHIFT;
                            if (vlq > 0) {
                                digit |= VLQ_CONTINUATION_BIT;
                            }
                            encoded += base64.encode(digit);
                        } while (vlq > 0);
                        return encoded;
                    };
                    exports.decode = function base64VLQ_decode(aStr) {
                        var i = 0;
                        var strLen = aStr.length;
                        var result = 0;
                        var shift = 0;
                        var continuation, digit;
                        do {
                            if (i >= strLen) {
                                throw new Error("Expected more digits in base 64 VLQ value.");
                            }
                            digit = base64.decode(aStr.charAt(i++));
                            continuation = !!(digit & VLQ_CONTINUATION_BIT);
                            digit &= VLQ_BASE_MASK;
                            result = result + (digit << shift);
                            shift += VLQ_BASE_SHIFT;
                        } while (continuation);
                        return {
                            value: fromVLQSigned(result),
                            rest: aStr.slice(i)
                        };
                    };
                });
            });
            require.define("/node_modules/source-map/lib/source-map/base64.js", function(module, exports, __dirname, __filename) {
                if (typeof define !== "function") {
                    var define = require("/node_modules/source-map/node_modules/amdefine/amdefine.js", module)(module, require);
                }
                define(function(require, exports, module) {
                    var charToIntMap = {};
                    var intToCharMap = {};
                    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".split("").forEach(function(ch, index) {
                        charToIntMap[ch] = index;
                        intToCharMap[index] = ch;
                    });
                    exports.encode = function base64_encode(aNumber) {
                        if (aNumber in intToCharMap) {
                            return intToCharMap[aNumber];
                        }
                        throw new TypeError("Must be between 0 and 63: " + aNumber);
                    };
                    exports.decode = function base64_decode(aChar) {
                        if (aChar in charToIntMap) {
                            return charToIntMap[aChar];
                        }
                        throw new TypeError("Not a valid base 64 digit: " + aChar);
                    };
                });
            });
            require.define("/node_modules/source-map/lib/source-map/source-map-consumer.js", function(module, exports, __dirname, __filename) {
                if (typeof define !== "function") {
                    var define = require("/node_modules/source-map/node_modules/amdefine/amdefine.js", module)(module, require);
                }
                define(function(require, exports, module) {
                    var util = require("/node_modules/source-map/lib/source-map/util.js", module);
                    var binarySearch = require("/node_modules/source-map/lib/source-map/binary-search.js", module);
                    var ArraySet = require("/node_modules/source-map/lib/source-map/array-set.js", module).ArraySet;
                    var base64VLQ = require("/node_modules/source-map/lib/source-map/base64-vlq.js", module);
                    function SourceMapConsumer(aSourceMap) {
                        var sourceMap = aSourceMap;
                        if (typeof aSourceMap === "string") {
                            sourceMap = JSON.parse(aSourceMap.replace(/^\)\]\}'/, ""));
                        }
                        var version = util.getArg(sourceMap, "version");
                        var sources = util.getArg(sourceMap, "sources");
                        var names = util.getArg(sourceMap, "names", []);
                        var sourceRoot = util.getArg(sourceMap, "sourceRoot", null);
                        var sourcesContent = util.getArg(sourceMap, "sourcesContent", null);
                        var mappings = util.getArg(sourceMap, "mappings");
                        var file = util.getArg(sourceMap, "file", null);
                        if (version != this._version) {
                            throw new Error("Unsupported version: " + version);
                        }
                        this._names = ArraySet.fromArray(names, true);
                        this._sources = ArraySet.fromArray(sources, true);
                        this.sourceRoot = sourceRoot;
                        this.sourcesContent = sourcesContent;
                        this._mappings = mappings;
                        this.file = file;
                    }
                    SourceMapConsumer.fromSourceMap = function SourceMapConsumer_fromSourceMap(aSourceMap) {
                        var smc = Object.create(SourceMapConsumer.prototype);
                        smc._names = ArraySet.fromArray(aSourceMap._names.toArray(), true);
                        smc._sources = ArraySet.fromArray(aSourceMap._sources.toArray(), true);
                        smc.sourceRoot = aSourceMap._sourceRoot;
                        smc.sourcesContent = aSourceMap._generateSourcesContent(smc._sources.toArray(), smc.sourceRoot);
                        smc.file = aSourceMap._file;
                        smc.__generatedMappings = aSourceMap._mappings.slice().sort(util.compareByGeneratedPositions);
                        smc.__originalMappings = aSourceMap._mappings.slice().sort(util.compareByOriginalPositions);
                        return smc;
                    };
                    SourceMapConsumer.prototype._version = 3;
                    Object.defineProperty(SourceMapConsumer.prototype, "sources", {
                        get: function() {
                            return this._sources.toArray().map(function(s) {
                                return this.sourceRoot ? util.join(this.sourceRoot, s) : s;
                            }, this);
                        }
                    });
                    SourceMapConsumer.prototype.__generatedMappings = null;
                    Object.defineProperty(SourceMapConsumer.prototype, "_generatedMappings", {
                        get: function() {
                            if (!this.__generatedMappings) {
                                this.__generatedMappings = [];
                                this.__originalMappings = [];
                                this._parseMappings(this._mappings, this.sourceRoot);
                            }
                            return this.__generatedMappings;
                        }
                    });
                    SourceMapConsumer.prototype.__originalMappings = null;
                    Object.defineProperty(SourceMapConsumer.prototype, "_originalMappings", {
                        get: function() {
                            if (!this.__originalMappings) {
                                this.__generatedMappings = [];
                                this.__originalMappings = [];
                                this._parseMappings(this._mappings, this.sourceRoot);
                            }
                            return this.__originalMappings;
                        }
                    });
                    SourceMapConsumer.prototype._parseMappings = function SourceMapConsumer_parseMappings(aStr, aSourceRoot) {
                        var generatedLine = 1;
                        var previousGeneratedColumn = 0;
                        var previousOriginalLine = 0;
                        var previousOriginalColumn = 0;
                        var previousSource = 0;
                        var previousName = 0;
                        var mappingSeparator = /^[,;]/;
                        var str = aStr;
                        var mapping;
                        var temp;
                        while (str.length > 0) {
                            if (str.charAt(0) === ";") {
                                generatedLine++;
                                str = str.slice(1);
                                previousGeneratedColumn = 0;
                            } else {
                                if (str.charAt(0) === ",") {
                                    str = str.slice(1);
                                } else {
                                    mapping = {};
                                    mapping.generatedLine = generatedLine;
                                    temp = base64VLQ.decode(str);
                                    mapping.generatedColumn = previousGeneratedColumn + temp.value;
                                    previousGeneratedColumn = mapping.generatedColumn;
                                    str = temp.rest;
                                    if (str.length > 0 && !mappingSeparator.test(str.charAt(0))) {
                                        temp = base64VLQ.decode(str);
                                        mapping.source = this._sources.at(previousSource + temp.value);
                                        previousSource += temp.value;
                                        str = temp.rest;
                                        if (str.length === 0 || mappingSeparator.test(str.charAt(0))) {
                                            throw new Error("Found a source, but no line and column");
                                        }
                                        temp = base64VLQ.decode(str);
                                        mapping.originalLine = previousOriginalLine + temp.value;
                                        previousOriginalLine = mapping.originalLine;
                                        mapping.originalLine += 1;
                                        str = temp.rest;
                                        if (str.length === 0 || mappingSeparator.test(str.charAt(0))) {
                                            throw new Error("Found a source and line, but no column");
                                        }
                                        temp = base64VLQ.decode(str);
                                        mapping.originalColumn = previousOriginalColumn + temp.value;
                                        previousOriginalColumn = mapping.originalColumn;
                                        str = temp.rest;
                                        if (str.length > 0 && !mappingSeparator.test(str.charAt(0))) {
                                            temp = base64VLQ.decode(str);
                                            mapping.name = this._names.at(previousName + temp.value);
                                            previousName += temp.value;
                                            str = temp.rest;
                                        }
                                    }
                                    this.__generatedMappings.push(mapping);
                                    if (typeof mapping.originalLine === "number") {
                                        this.__originalMappings.push(mapping);
                                    }
                                }
                            }
                        }
                        this.__originalMappings.sort(util.compareByOriginalPositions);
                    };
                    SourceMapConsumer.prototype._findMapping = function SourceMapConsumer_findMapping(aNeedle, aMappings, aLineName, aColumnName, aComparator) {
                        if (aNeedle[aLineName] <= 0) {
                            throw new TypeError("Line must be greater than or equal to 1, got " + aNeedle[aLineName]);
                        }
                        if (aNeedle[aColumnName] < 0) {
                            throw new TypeError("Column must be greater than or equal to 0, got " + aNeedle[aColumnName]);
                        }
                        return binarySearch.search(aNeedle, aMappings, aComparator);
                    };
                    SourceMapConsumer.prototype.originalPositionFor = function SourceMapConsumer_originalPositionFor(aArgs) {
                        var needle = {
                            generatedLine: util.getArg(aArgs, "line"),
                            generatedColumn: util.getArg(aArgs, "column")
                        };
                        var mapping = this._findMapping(needle, this._generatedMappings, "generatedLine", "generatedColumn", util.compareByGeneratedPositions);
                        if (mapping) {
                            var source = util.getArg(mapping, "source", null);
                            if (source && this.sourceRoot) {
                                source = util.join(this.sourceRoot, source);
                            }
                            return {
                                source: source,
                                line: util.getArg(mapping, "originalLine", null),
                                column: util.getArg(mapping, "originalColumn", null),
                                name: util.getArg(mapping, "name", null)
                            };
                        }
                        return {
                            source: null,
                            line: null,
                            column: null,
                            name: null
                        };
                    };
                    SourceMapConsumer.prototype.sourceContentFor = function SourceMapConsumer_sourceContentFor(aSource) {
                        if (!this.sourcesContent) {
                            return null;
                        }
                        if (this.sourceRoot) {
                            aSource = util.relative(this.sourceRoot, aSource);
                        }
                        if (this._sources.has(aSource)) {
                            return this.sourcesContent[this._sources.indexOf(aSource)];
                        }
                        var url;
                        if (this.sourceRoot && (url = util.urlParse(this.sourceRoot))) {
                            var fileUriAbsPath = aSource.replace(/^file:\/\//, "");
                            if (url.scheme == "file" && this._sources.has(fileUriAbsPath)) {
                                return this.sourcesContent[this._sources.indexOf(fileUriAbsPath)];
                            }
                            if ((!url.path || url.path == "/") && this._sources.has("/" + aSource)) {
                                return this.sourcesContent[this._sources.indexOf("/" + aSource)];
                            }
                        }
                        throw new Error('"' + aSource + '" is not in the SourceMap.');
                    };
                    SourceMapConsumer.prototype.generatedPositionFor = function SourceMapConsumer_generatedPositionFor(aArgs) {
                        var needle = {
                            source: util.getArg(aArgs, "source"),
                            originalLine: util.getArg(aArgs, "line"),
                            originalColumn: util.getArg(aArgs, "column")
                        };
                        if (this.sourceRoot) {
                            needle.source = util.relative(this.sourceRoot, needle.source);
                        }
                        var mapping = this._findMapping(needle, this._originalMappings, "originalLine", "originalColumn", util.compareByOriginalPositions);
                        if (mapping) {
                            return {
                                line: util.getArg(mapping, "generatedLine", null),
                                column: util.getArg(mapping, "generatedColumn", null)
                            };
                        }
                        return {
                            line: null,
                            column: null
                        };
                    };
                    SourceMapConsumer.GENERATED_ORDER = 1;
                    SourceMapConsumer.ORIGINAL_ORDER = 2;
                    SourceMapConsumer.prototype.eachMapping = function SourceMapConsumer_eachMapping(aCallback, aContext, aOrder) {
                        var context = aContext || null;
                        var order = aOrder || SourceMapConsumer.GENERATED_ORDER;
                        var mappings;
                        switch (order) {
                          case SourceMapConsumer.GENERATED_ORDER:
                            mappings = this._generatedMappings;
                            break;

                          case SourceMapConsumer.ORIGINAL_ORDER:
                            mappings = this._originalMappings;
                            break;

                          default:
                            throw new Error("Unknown order of iteration.");
                        }
                        var sourceRoot = this.sourceRoot;
                        mappings.map(function(mapping) {
                            var source = mapping.source;
                            if (source && sourceRoot) {
                                source = util.join(sourceRoot, source);
                            }
                            return {
                                source: source,
                                generatedLine: mapping.generatedLine,
                                generatedColumn: mapping.generatedColumn,
                                originalLine: mapping.originalLine,
                                originalColumn: mapping.originalColumn,
                                name: mapping.name
                            };
                        }).forEach(aCallback, context);
                    };
                    exports.SourceMapConsumer = SourceMapConsumer;
                });
            });
            require.define("/node_modules/source-map/lib/source-map/binary-search.js", function(module, exports, __dirname, __filename) {
                if (typeof define !== "function") {
                    var define = require("/node_modules/source-map/node_modules/amdefine/amdefine.js", module)(module, require);
                }
                define(function(require, exports, module) {
                    function recursiveSearch(aLow, aHigh, aNeedle, aHaystack, aCompare) {
                        var mid = Math.floor((aHigh - aLow) / 2) + aLow;
                        var cmp = aCompare(aNeedle, aHaystack[mid], true);
                        if (cmp === 0) {
                            return aHaystack[mid];
                        } else {
                            if (cmp > 0) {
                                if (aHigh - mid > 1) {
                                    return recursiveSearch(mid, aHigh, aNeedle, aHaystack, aCompare);
                                }
                                return aHaystack[mid];
                            } else {
                                if (mid - aLow > 1) {
                                    return recursiveSearch(aLow, mid, aNeedle, aHaystack, aCompare);
                                }
                                return aLow < 0 ? null : aHaystack[aLow];
                            }
                        }
                    }
                    exports.search = function search(aNeedle, aHaystack, aCompare) {
                        return aHaystack.length > 0 ? recursiveSearch(-1, aHaystack.length, aNeedle, aHaystack, aCompare) : null;
                    };
                });
            });
            require.define("/node_modules/esutils/lib/utils.js", function(module, exports, __dirname, __filename) {
                (function() {
                    "use strict";
                    exports.code = require("/node_modules/esutils/lib/code.js", module);
                    exports.keyword = require("/node_modules/esutils/lib/keyword.js", module);
                })();
            });
            require.define("/node_modules/esutils/lib/keyword.js", function(module, exports, __dirname, __filename) {
                (function() {
                    "use strict";
                    var code = require("/node_modules/esutils/lib/code.js", module);
                    function isStrictModeReservedWordES6(id) {
                        switch (id) {
                          case "implements":
                          case "interface":
                          case "package":
                          case "private":
                          case "protected":
                          case "public":
                          case "static":
                          case "let":
                            return true;

                          default:
                            return false;
                        }
                    }
                    function isKeywordES5(id, strict) {
                        if (!strict && id === "yield") {
                            return false;
                        }
                        return isKeywordES6(id, strict);
                    }
                    function isKeywordES6(id, strict) {
                        if (strict && isStrictModeReservedWordES6(id)) {
                            return true;
                        }
                        switch (id.length) {
                          case 2:
                            return id === "if" || id === "in" || id === "do";

                          case 3:
                            return id === "var" || id === "for" || id === "new" || id === "try";

                          case 4:
                            return id === "this" || id === "else" || id === "case" || id === "void" || id === "with" || id === "enum";

                          case 5:
                            return id === "while" || id === "break" || id === "catch" || id === "throw" || id === "const" || id === "yield" || id === "class" || id === "super";

                          case 6:
                            return id === "return" || id === "typeof" || id === "delete" || id === "switch" || id === "export" || id === "import";

                          case 7:
                            return id === "default" || id === "finally" || id === "extends";

                          case 8:
                            return id === "function" || id === "continue" || id === "debugger";

                          case 10:
                            return id === "instanceof";

                          default:
                            return false;
                        }
                    }
                    function isRestrictedWord(id) {
                        return id === "eval" || id === "arguments";
                    }
                    function isIdentifierName(id) {
                        var i, iz, ch;
                        if (id.length === 0) {
                            return false;
                        }
                        ch = id.charCodeAt(0);
                        if (!code.isIdentifierStart(ch) || ch === 92) {
                            return false;
                        }
                        for (i = 1, iz = id.length; i < iz; ++i) {
                            ch = id.charCodeAt(i);
                            if (!code.isIdentifierPart(ch) || ch === 92) {
                                return false;
                            }
                        }
                        return true;
                    }
                    module.exports = {
                        isKeywordES5: isKeywordES5,
                        isKeywordES6: isKeywordES6,
                        isRestrictedWord: isRestrictedWord,
                        isIdentifierName: isIdentifierName
                    };
                })();
            });
            require.define("/node_modules/esutils/lib/code.js", function(module, exports, __dirname, __filename) {
                (function() {
                    "use strict";
                    var Regex;
                    Regex = {
                        NonAsciiIdentifierStart: new RegExp("[ªµºÀ-ÖØ-öø-ˁˆ-ˑˠ-ˤˬˮͰ-ʹͶͷͺ-ͽΆΈ-ΊΌΎ-ΡΣ-ϵϷ-ҁҊ-ԧԱ-Ֆՙա-ևא-תװ-ײؠ-يٮٯٱ-ۓەۥۦۮۯۺ-ۼۿܐܒ-ܯݍ-ޥޱߊ-ߪߴߵߺࠀ-ࠕࠚࠤࠨࡀ-ࡘࢠࢢ-ࢬऄ-हऽॐक़-ॡॱ-ॷॹ-ॿঅ-ঌএঐও-নপ-রলশ-হঽৎড়ঢ়য়-ৡৰৱਅ-ਊਏਐਓ-ਨਪ-ਰਲਲ਼ਵਸ਼ਸਹਖ਼-ੜਫ਼ੲ-ੴઅ-ઍએ-ઑઓ-નપ-રલળવ-હઽૐૠૡଅ-ଌଏଐଓ-ନପ-ରଲଳଵ-ହଽଡ଼ଢ଼ୟ-ୡୱஃஅ-ஊஎ-ஐஒ-கஙசஜஞடணதந-பம-ஹௐఅ-ఌఎ-ఐఒ-నప-ళవ-హఽౘౙౠౡಅ-ಌಎ-ಐಒ-ನಪ-ಳವ-ಹಽೞೠೡೱೲഅ-ഌഎ-ഐഒ-ഺഽൎൠൡൺ-ൿඅ-ඖක-නඳ-රලව-ෆก-ะาำเ-ๆກຂຄງຈຊຍດ-ທນ-ຟມ-ຣລວສຫອ-ະາຳຽເ-ໄໆໜ-ໟༀཀ-ཇཉ-ཬྈ-ྌက-ဪဿၐ-ၕၚ-ၝၡၥၦၮ-ၰၵ-ႁႎႠ-ჅჇჍა-ჺჼ-ቈቊ-ቍቐ-ቖቘቚ-ቝበ-ኈኊ-ኍነ-ኰኲ-ኵኸ-ኾዀዂ-ዅወ-ዖዘ-ጐጒ-ጕጘ-ፚᎀ-ᎏᎠ-Ᏼᐁ-ᙬᙯ-ᙿᚁ-ᚚᚠ-ᛪᛮ-ᛰᜀ-ᜌᜎ-ᜑᜠ-ᜱᝀ-ᝑᝠ-ᝬᝮ-ᝰក-ឳៗៜᠠ-ᡷᢀ-ᢨᢪᢰ-ᣵᤀ-ᤜᥐ-ᥭᥰ-ᥴᦀ-ᦫᧁ-ᧇᨀ-ᨖᨠ-ᩔᪧᬅ-ᬳᭅ-ᭋᮃ-ᮠᮮᮯᮺ-ᯥᰀ-ᰣᱍ-ᱏᱚ-ᱽᳩ-ᳬᳮ-ᳱᳵᳶᴀ-ᶿḀ-ἕἘ-Ἕἠ-ὅὈ-Ὅὐ-ὗὙὛὝὟ-ώᾀ-ᾴᾶ-ᾼιῂ-ῄῆ-ῌῐ-ΐῖ-Ίῠ-Ῥῲ-ῴῶ-ῼⁱⁿₐ-ₜℂℇℊ-ℓℕℙ-ℝℤΩℨK-ℭℯ-ℹℼ-ℿⅅ-ⅉⅎⅠ-ↈⰀ-Ⱞⰰ-ⱞⱠ-ⳤⳫ-ⳮⳲⳳⴀ-ⴥⴧⴭⴰ-ⵧⵯⶀ-ⶖⶠ-ⶦⶨ-ⶮⶰ-ⶶⶸ-ⶾⷀ-ⷆⷈ-ⷎⷐ-ⷖⷘ-ⷞⸯ々-〇〡-〩〱-〵〸-〼ぁ-ゖゝ-ゟァ-ヺー-ヿㄅ-ㄭㄱ-ㆎㆠ-ㆺㇰ-ㇿ㐀-䶵一-鿌ꀀ-ꒌꓐ-ꓽꔀ-ꘌꘐ-ꘟꘪꘫꙀ-ꙮꙿ-ꚗꚠ-ꛯꜗ-ꜟꜢ-ꞈꞋ-ꞎꞐ-ꞓꞠ-Ɦꟸ-ꠁꠃ-ꠅꠇ-ꠊꠌ-ꠢꡀ-ꡳꢂ-ꢳꣲ-ꣷꣻꤊ-ꤥꤰ-ꥆꥠ-ꥼꦄ-ꦲꧏꨀ-ꨨꩀ-ꩂꩄ-ꩋꩠ-ꩶꩺꪀ-ꪯꪱꪵꪶꪹ-ꪽꫀꫂꫛ-ꫝꫠ-ꫪꫲ-ꫴꬁ-ꬆꬉ-ꬎꬑ-ꬖꬠ-ꬦꬨ-ꬮꯀ-ꯢ가-힣ힰ-ퟆퟋ-ퟻ豈-舘並-龎ﬀ-ﬆﬓ-ﬗיִײַ-ﬨשׁ-זּטּ-לּמּנּסּףּפּצּ-ﮱﯓ-ﴽﵐ-ﶏﶒ-ﷇﷰ-ﷻﹰ-ﹴﹶ-ﻼＡ-Ｚａ-ｚｦ-ﾾￂ-ￇￊ-ￏￒ-ￗￚ-ￜ]"),
                        NonAsciiIdentifierPart: new RegExp("[ªµºÀ-ÖØ-öø-ˁˆ-ˑˠ-ˤˬˮ̀-ʹͶͷͺ-ͽΆΈ-ΊΌΎ-ΡΣ-ϵϷ-ҁ҃-҇Ҋ-ԧԱ-Ֆՙա-և֑-ׇֽֿׁׂׅׄא-תװ-ײؐ-ؚؠ-٩ٮ-ۓە-ۜ۟-۪ۨ-ۼۿܐ-݊ݍ-ޱ߀-ߵߺࠀ-࠭ࡀ-࡛ࢠࢢ-ࢬࣤ-ࣾऀ-ॣ०-९ॱ-ॷॹ-ॿঁ-ঃঅ-ঌএঐও-নপ-রলশ-হ়-ৄেৈো-ৎৗড়ঢ়য়-ৣ০-ৱਁ-ਃਅ-ਊਏਐਓ-ਨਪ-ਰਲਲ਼ਵਸ਼ਸਹ਼ਾ-ੂੇੈੋ-੍ੑਖ਼-ੜਫ਼੦-ੵઁ-ઃઅ-ઍએ-ઑઓ-નપ-રલળવ-હ઼-ૅે-ૉો-્ૐૠ-ૣ૦-૯ଁ-ଃଅ-ଌଏଐଓ-ନପ-ରଲଳଵ-ହ଼-ୄେୈୋ-୍ୖୗଡ଼ଢ଼ୟ-ୣ୦-୯ୱஂஃஅ-ஊஎ-ஐஒ-கஙசஜஞடணதந-பம-ஹா-ூெ-ைொ-்ௐௗ௦-௯ఁ-ఃఅ-ఌఎ-ఐఒ-నప-ళవ-హఽ-ౄె-ైొ-్ౕౖౘౙౠ-ౣ౦-౯ಂಃಅ-ಌಎ-ಐಒ-ನಪ-ಳವ-ಹ಼-ೄೆ-ೈೊ-್ೕೖೞೠ-ೣ೦-೯ೱೲംഃഅ-ഌഎ-ഐഒ-ഺഽ-ൄെ-ൈൊ-ൎൗൠ-ൣ൦-൯ൺ-ൿංඃඅ-ඖක-නඳ-රලව-ෆ්ා-ුූෘ-ෟෲෳก-ฺเ-๎๐-๙ກຂຄງຈຊຍດ-ທນ-ຟມ-ຣລວສຫອ-ູົ-ຽເ-ໄໆ່-ໍ໐-໙ໜ-ໟༀ༘༙༠-༩༹༵༷༾-ཇཉ-ཬཱ-྄྆-ྗྙ-ྼ࿆က-၉ၐ-ႝႠ-ჅჇჍა-ჺჼ-ቈቊ-ቍቐ-ቖቘቚ-ቝበ-ኈኊ-ኍነ-ኰኲ-ኵኸ-ኾዀዂ-ዅወ-ዖዘ-ጐጒ-ጕጘ-ፚ፝-፟ᎀ-ᎏᎠ-Ᏼᐁ-ᙬᙯ-ᙿᚁ-ᚚᚠ-ᛪᛮ-ᛰᜀ-ᜌᜎ-᜔ᜠ-᜴ᝀ-ᝓᝠ-ᝬᝮ-ᝰᝲᝳក-៓ៗៜ៝០-៩᠋-᠍᠐-᠙ᠠ-ᡷᢀ-ᢪᢰ-ᣵᤀ-ᤜᤠ-ᤫᤰ-᤻᥆-ᥭᥰ-ᥴᦀ-ᦫᦰ-ᧉ᧐-᧙ᨀ-ᨛᨠ-ᩞ᩠-᩿᩼-᪉᪐-᪙ᪧᬀ-ᭋ᭐-᭙᭫-᭳ᮀ-᯳ᰀ-᰷᱀-᱉ᱍ-ᱽ᳐-᳔᳒-ᳶᴀ-ᷦ᷼-ἕἘ-Ἕἠ-ὅὈ-Ὅὐ-ὗὙὛὝὟ-ώᾀ-ᾴᾶ-ᾼιῂ-ῄῆ-ῌῐ-ΐῖ-Ίῠ-Ῥῲ-ῴῶ-ῼ‌‍‿⁀⁔ⁱⁿₐ-ₜ⃐-⃥⃜⃡-⃰ℂℇℊ-ℓℕℙ-ℝℤΩℨK-ℭℯ-ℹℼ-ℿⅅ-ⅉⅎⅠ-ↈⰀ-Ⱞⰰ-ⱞⱠ-ⳤⳫ-ⳳⴀ-ⴥⴧⴭⴰ-ⵧⵯ⵿-ⶖⶠ-ⶦⶨ-ⶮⶰ-ⶶⶸ-ⶾⷀ-ⷆⷈ-ⷎⷐ-ⷖⷘ-ⷞⷠ-ⷿⸯ々-〇〡-〯〱-〵〸-〼ぁ-ゖ゙゚ゝ-ゟァ-ヺー-ヿㄅ-ㄭㄱ-ㆎㆠ-ㆺㇰ-ㇿ㐀-䶵一-鿌ꀀ-ꒌꓐ-ꓽꔀ-ꘌꘐ-ꘫꙀ-꙯ꙴ-꙽ꙿ-ꚗꚟ-꛱ꜗ-ꜟꜢ-ꞈꞋ-ꞎꞐ-ꞓꞠ-Ɦꟸ-ꠧꡀ-ꡳꢀ-꣄꣐-꣙꣠-ꣷꣻ꤀-꤭ꤰ-꥓ꥠ-ꥼꦀ-꧀ꧏ-꧙ꨀ-ꨶꩀ-ꩍ꩐-꩙ꩠ-ꩶꩺꩻꪀ-ꫂꫛ-ꫝꫠ-ꫯꫲ-꫶ꬁ-ꬆꬉ-ꬎꬑ-ꬖꬠ-ꬦꬨ-ꬮꯀ-ꯪ꯬꯭꯰-꯹가-힣ힰ-ퟆퟋ-ퟻ豈-舘並-龎ﬀ-ﬆﬓ-ﬗיִ-ﬨשׁ-זּטּ-לּמּנּסּףּפּצּ-ﮱﯓ-ﴽﵐ-ﶏﶒ-ﷇﷰ-ﷻ︀-️︠-︦︳︴﹍-﹏ﹰ-ﹴﹶ-ﻼ０-９Ａ-Ｚ＿ａ-ｚｦ-ﾾￂ-ￇￊ-ￏￒ-ￗￚ-ￜ]")
                    };
                    function isDecimalDigit(ch) {
                        return ch >= 48 && ch <= 57;
                    }
                    function isHexDigit(ch) {
                        return isDecimalDigit(ch) || 97 <= ch && ch <= 102 || 65 <= ch && ch <= 70;
                    }
                    function isOctalDigit(ch) {
                        return ch >= 48 && ch <= 55;
                    }
                    function isWhiteSpace(ch) {
                        return ch === 32 || ch === 9 || ch === 11 || ch === 12 || ch === 160 || ch >= 5760 && [ 5760, 6158, 8192, 8193, 8194, 8195, 8196, 8197, 8198, 8199, 8200, 8201, 8202, 8239, 8287, 12288, 65279 ].indexOf(ch) >= 0;
                    }
                    function isLineTerminator(ch) {
                        return ch === 10 || ch === 13 || ch === 8232 || ch === 8233;
                    }
                    function isIdentifierStart(ch) {
                        return ch === 36 || ch === 95 || ch >= 65 && ch <= 90 || ch >= 97 && ch <= 122 || ch === 92 || ch >= 128 && Regex.NonAsciiIdentifierStart.test(String.fromCharCode(ch));
                    }
                    function isIdentifierPart(ch) {
                        return ch === 36 || ch === 95 || ch >= 65 && ch <= 90 || ch >= 97 && ch <= 122 || ch >= 48 && ch <= 57 || ch === 92 || ch >= 128 && Regex.NonAsciiIdentifierPart.test(String.fromCharCode(ch));
                    }
                    module.exports = {
                        isDecimalDigit: isDecimalDigit,
                        isHexDigit: isHexDigit,
                        isOctalDigit: isOctalDigit,
                        isWhiteSpace: isWhiteSpace,
                        isLineTerminator: isLineTerminator,
                        isIdentifierStart: isIdentifierStart,
                        isIdentifierPart: isIdentifierPart
                    };
                })();
            });
            require.define("/node_modules/estraverse/estraverse.js", function(module, exports, __dirname, __filename) {
                (function(root, factory) {
                    "use strict";
                    if (typeof define === "function" && define.amd) {
                        define([ "exports" ], factory);
                    } else {
                        if (typeof exports !== "undefined") {
                            factory(exports);
                        } else {
                            factory(root.estraverse = {});
                        }
                    }
                })(this, function(exports) {
                    "use strict";
                    var Syntax, isArray, VisitorOption, VisitorKeys, BREAK, SKIP;
                    Syntax = {
                        AssignmentExpression: "AssignmentExpression",
                        ArrayExpression: "ArrayExpression",
                        ArrayPattern: "ArrayPattern",
                        ArrowFunctionExpression: "ArrowFunctionExpression",
                        BlockStatement: "BlockStatement",
                        BinaryExpression: "BinaryExpression",
                        BreakStatement: "BreakStatement",
                        CallExpression: "CallExpression",
                        CatchClause: "CatchClause",
                        ClassBody: "ClassBody",
                        ClassDeclaration: "ClassDeclaration",
                        ClassExpression: "ClassExpression",
                        ConditionalExpression: "ConditionalExpression",
                        ContinueStatement: "ContinueStatement",
                        DebuggerStatement: "DebuggerStatement",
                        DirectiveStatement: "DirectiveStatement",
                        DoWhileStatement: "DoWhileStatement",
                        EmptyStatement: "EmptyStatement",
                        ExpressionStatement: "ExpressionStatement",
                        ForStatement: "ForStatement",
                        ForInStatement: "ForInStatement",
                        FunctionDeclaration: "FunctionDeclaration",
                        FunctionExpression: "FunctionExpression",
                        Identifier: "Identifier",
                        IfStatement: "IfStatement",
                        Literal: "Literal",
                        LabeledStatement: "LabeledStatement",
                        LogicalExpression: "LogicalExpression",
                        MemberExpression: "MemberExpression",
                        MethodDefinition: "MethodDefinition",
                        NewExpression: "NewExpression",
                        ObjectExpression: "ObjectExpression",
                        ObjectPattern: "ObjectPattern",
                        Program: "Program",
                        Property: "Property",
                        ReturnStatement: "ReturnStatement",
                        SequenceExpression: "SequenceExpression",
                        SwitchStatement: "SwitchStatement",
                        SwitchCase: "SwitchCase",
                        ThisExpression: "ThisExpression",
                        ThrowStatement: "ThrowStatement",
                        TryStatement: "TryStatement",
                        UnaryExpression: "UnaryExpression",
                        UpdateExpression: "UpdateExpression",
                        VariableDeclaration: "VariableDeclaration",
                        VariableDeclarator: "VariableDeclarator",
                        WhileStatement: "WhileStatement",
                        WithStatement: "WithStatement",
                        YieldExpression: "YieldExpression"
                    };
                    function ignoreJSHintError() {}
                    isArray = Array.isArray;
                    if (!isArray) {
                        isArray = function isArray(array) {
                            return Object.prototype.toString.call(array) === "[object Array]";
                        };
                    }
                    function deepCopy(obj) {
                        var ret = {}, key, val;
                        for (key in obj) {
                            if (obj.hasOwnProperty(key)) {
                                val = obj[key];
                                if (typeof val === "object" && val !== null) {
                                    ret[key] = deepCopy(val);
                                } else {
                                    ret[key] = val;
                                }
                            }
                        }
                        return ret;
                    }
                    function shallowCopy(obj) {
                        var ret = {}, key;
                        for (key in obj) {
                            if (obj.hasOwnProperty(key)) {
                                ret[key] = obj[key];
                            }
                        }
                        return ret;
                    }
                    ignoreJSHintError(shallowCopy);
                    function upperBound(array, func) {
                        var diff, len, i, current;
                        len = array.length;
                        i = 0;
                        while (len) {
                            diff = len >>> 1;
                            current = i + diff;
                            if (func(array[current])) {
                                len = diff;
                            } else {
                                i = current + 1;
                                len -= diff + 1;
                            }
                        }
                        return i;
                    }
                    function lowerBound(array, func) {
                        var diff, len, i, current;
                        len = array.length;
                        i = 0;
                        while (len) {
                            diff = len >>> 1;
                            current = i + diff;
                            if (func(array[current])) {
                                i = current + 1;
                                len -= diff + 1;
                            } else {
                                len = diff;
                            }
                        }
                        return i;
                    }
                    ignoreJSHintError(lowerBound);
                    VisitorKeys = {
                        AssignmentExpression: [ "left", "right" ],
                        ArrayExpression: [ "elements" ],
                        ArrayPattern: [ "elements" ],
                        ArrowFunctionExpression: [ "params", "defaults", "rest", "body" ],
                        BlockStatement: [ "body" ],
                        BinaryExpression: [ "left", "right" ],
                        BreakStatement: [ "label" ],
                        CallExpression: [ "callee", "arguments" ],
                        CatchClause: [ "param", "body" ],
                        ClassBody: [ "body" ],
                        ClassDeclaration: [ "id", "body", "superClass" ],
                        ClassExpression: [ "id", "body", "superClass" ],
                        ConditionalExpression: [ "test", "consequent", "alternate" ],
                        ContinueStatement: [ "label" ],
                        DebuggerStatement: [],
                        DirectiveStatement: [],
                        DoWhileStatement: [ "body", "test" ],
                        EmptyStatement: [],
                        ExpressionStatement: [ "expression" ],
                        ForStatement: [ "init", "test", "update", "body" ],
                        ForInStatement: [ "left", "right", "body" ],
                        FunctionDeclaration: [ "id", "params", "defaults", "rest", "body" ],
                        FunctionExpression: [ "id", "params", "defaults", "rest", "body" ],
                        Identifier: [],
                        IfStatement: [ "test", "consequent", "alternate" ],
                        Literal: [],
                        LabeledStatement: [ "label", "body" ],
                        LogicalExpression: [ "left", "right" ],
                        MemberExpression: [ "object", "property" ],
                        MethodDefinition: [ "key", "value" ],
                        NewExpression: [ "callee", "arguments" ],
                        ObjectExpression: [ "properties" ],
                        ObjectPattern: [ "properties" ],
                        Program: [ "body" ],
                        Property: [ "key", "value" ],
                        ReturnStatement: [ "argument" ],
                        SequenceExpression: [ "expressions" ],
                        SwitchStatement: [ "discriminant", "cases" ],
                        SwitchCase: [ "test", "consequent" ],
                        ThisExpression: [],
                        ThrowStatement: [ "argument" ],
                        TryStatement: [ "block", "handlers", "handler", "guardedHandlers", "finalizer" ],
                        UnaryExpression: [ "argument" ],
                        UpdateExpression: [ "argument" ],
                        VariableDeclaration: [ "declarations" ],
                        VariableDeclarator: [ "id", "init" ],
                        WhileStatement: [ "test", "body" ],
                        WithStatement: [ "object", "body" ],
                        YieldExpression: [ "argument" ]
                    };
                    BREAK = {};
                    SKIP = {};
                    VisitorOption = {
                        Break: BREAK,
                        Skip: SKIP
                    };
                    function Reference(parent, key) {
                        this.parent = parent;
                        this.key = key;
                    }
                    Reference.prototype.replace = function replace(node) {
                        this.parent[this.key] = node;
                    };
                    function Element(node, path, wrap, ref) {
                        this.node = node;
                        this.path = path;
                        this.wrap = wrap;
                        this.ref = ref;
                    }
                    function Controller() {}
                    Controller.prototype.path = function path() {
                        var i, iz, j, jz, result, element;
                        function addToPath(result, path) {
                            if (isArray(path)) {
                                for (j = 0, jz = path.length; j < jz; ++j) {
                                    result.push(path[j]);
                                }
                            } else {
                                result.push(path);
                            }
                        }
                        if (!this.__current.path) {
                            return null;
                        }
                        result = [];
                        for (i = 2, iz = this.__leavelist.length; i < iz; ++i) {
                            element = this.__leavelist[i];
                            addToPath(result, element.path);
                        }
                        addToPath(result, this.__current.path);
                        return result;
                    };
                    Controller.prototype.parents = function parents() {
                        var i, iz, result;
                        result = [];
                        for (i = 1, iz = this.__leavelist.length; i < iz; ++i) {
                            result.push(this.__leavelist[i].node);
                        }
                        return result;
                    };
                    Controller.prototype.current = function current() {
                        return this.__current.node;
                    };
                    Controller.prototype.__execute = function __execute(callback, element) {
                        var previous, result;
                        result = undefined;
                        previous = this.__current;
                        this.__current = element;
                        this.__state = null;
                        if (callback) {
                            result = callback.call(this, element.node, this.__leavelist[this.__leavelist.length - 1].node);
                        }
                        this.__current = previous;
                        return result;
                    };
                    Controller.prototype.notify = function notify(flag) {
                        this.__state = flag;
                    };
                    Controller.prototype.skip = function() {
                        this.notify(SKIP);
                    };
                    Controller.prototype["break"] = function() {
                        this.notify(BREAK);
                    };
                    Controller.prototype.__initialize = function(root, visitor) {
                        this.visitor = visitor;
                        this.root = root;
                        this.__worklist = [];
                        this.__leavelist = [];
                        this.__current = null;
                        this.__state = null;
                    };
                    Controller.prototype.traverse = function traverse(root, visitor) {
                        var worklist, leavelist, element, node, nodeType, ret, key, current, current2, candidates, candidate, sentinel;
                        this.__initialize(root, visitor);
                        sentinel = {};
                        worklist = this.__worklist;
                        leavelist = this.__leavelist;
                        worklist.push(new Element(root, null, null, null));
                        leavelist.push(new Element(null, null, null, null));
                        while (worklist.length) {
                            element = worklist.pop();
                            if (element === sentinel) {
                                element = leavelist.pop();
                                ret = this.__execute(visitor.leave, element);
                                if (this.__state === BREAK || ret === BREAK) {
                                    return;
                                }
                                continue;
                            }
                            if (element.node) {
                                ret = this.__execute(visitor.enter, element);
                                if (this.__state === BREAK || ret === BREAK) {
                                    return;
                                }
                                worklist.push(sentinel);
                                leavelist.push(element);
                                if (this.__state === SKIP || ret === SKIP) {
                                    continue;
                                }
                                node = element.node;
                                nodeType = element.wrap || node.type;
                                candidates = VisitorKeys[nodeType];
                                current = candidates.length;
                                while ((current -= 1) >= 0) {
                                    key = candidates[current];
                                    candidate = node[key];
                                    if (!candidate) {
                                        continue;
                                    }
                                    if (!isArray(candidate)) {
                                        worklist.push(new Element(candidate, key, null, null));
                                        continue;
                                    }
                                    current2 = candidate.length;
                                    while ((current2 -= 1) >= 0) {
                                        if (!candidate[current2]) {
                                            continue;
                                        }
                                        if ((nodeType === Syntax.ObjectExpression || nodeType === Syntax.ObjectPattern) && "properties" === candidates[current]) {
                                            element = new Element(candidate[current2], [ key, current2 ], "Property", null);
                                        } else {
                                            element = new Element(candidate[current2], [ key, current2 ], null, null);
                                        }
                                        worklist.push(element);
                                    }
                                }
                            }
                        }
                    };
                    Controller.prototype.replace = function replace(root, visitor) {
                        var worklist, leavelist, node, nodeType, target, element, current, current2, candidates, candidate, sentinel, outer, key;
                        this.__initialize(root, visitor);
                        sentinel = {};
                        worklist = this.__worklist;
                        leavelist = this.__leavelist;
                        outer = {
                            root: root
                        };
                        element = new Element(root, null, null, new Reference(outer, "root"));
                        worklist.push(element);
                        leavelist.push(element);
                        while (worklist.length) {
                            element = worklist.pop();
                            if (element === sentinel) {
                                element = leavelist.pop();
                                target = this.__execute(visitor.leave, element);
                                if (target !== undefined && target !== BREAK && target !== SKIP) {
                                    element.ref.replace(target);
                                }
                                if (this.__state === BREAK || target === BREAK) {
                                    return outer.root;
                                }
                                continue;
                            }
                            target = this.__execute(visitor.enter, element);
                            if (target !== undefined && target !== BREAK && target !== SKIP) {
                                element.ref.replace(target);
                                element.node = target;
                            }
                            if (this.__state === BREAK || target === BREAK) {
                                return outer.root;
                            }
                            node = element.node;
                            if (!node) {
                                continue;
                            }
                            worklist.push(sentinel);
                            leavelist.push(element);
                            if (this.__state === SKIP || target === SKIP) {
                                continue;
                            }
                            nodeType = element.wrap || node.type;
                            candidates = VisitorKeys[nodeType];
                            current = candidates.length;
                            while ((current -= 1) >= 0) {
                                key = candidates[current];
                                candidate = node[key];
                                if (!candidate) {
                                    continue;
                                }
                                if (!isArray(candidate)) {
                                    worklist.push(new Element(candidate, key, null, new Reference(node, key)));
                                    continue;
                                }
                                current2 = candidate.length;
                                while ((current2 -= 1) >= 0) {
                                    if (!candidate[current2]) {
                                        continue;
                                    }
                                    if (nodeType === Syntax.ObjectExpression && "properties" === candidates[current]) {
                                        element = new Element(candidate[current2], [ key, current2 ], "Property", new Reference(candidate, current2));
                                    } else {
                                        element = new Element(candidate[current2], [ key, current2 ], null, new Reference(candidate, current2));
                                    }
                                    worklist.push(element);
                                }
                            }
                        }
                        return outer.root;
                    };
                    function traverse(root, visitor) {
                        var controller = new Controller();
                        return controller.traverse(root, visitor);
                    }
                    function replace(root, visitor) {
                        var controller = new Controller();
                        return controller.replace(root, visitor);
                    }
                    function extendCommentRange(comment, tokens) {
                        var target;
                        target = upperBound(tokens, function search(token) {
                            return token.range[0] > comment.range[0];
                        });
                        comment.extendedRange = [ comment.range[0], comment.range[1] ];
                        if (target !== tokens.length) {
                            comment.extendedRange[1] = tokens[target].range[0];
                        }
                        target -= 1;
                        if (target >= 0) {
                            comment.extendedRange[0] = tokens[target].range[1];
                        }
                        return comment;
                    }
                    function attachComments(tree, providedComments, tokens) {
                        var comments = [], comment, len, i, cursor;
                        if (!tree.range) {
                            throw new Error("attachComments needs range information");
                        }
                        if (!tokens.length) {
                            if (providedComments.length) {
                                for (i = 0, len = providedComments.length; i < len; i += 1) {
                                    comment = deepCopy(providedComments[i]);
                                    comment.extendedRange = [ 0, tree.range[0] ];
                                    comments.push(comment);
                                }
                                tree.leadingComments = comments;
                            }
                            return tree;
                        }
                        for (i = 0, len = providedComments.length; i < len; i += 1) {
                            comments.push(extendCommentRange(deepCopy(providedComments[i]), tokens));
                        }
                        cursor = 0;
                        traverse(tree, {
                            enter: function(node) {
                                var comment;
                                while (cursor < comments.length) {
                                    comment = comments[cursor];
                                    if (comment.extendedRange[1] > node.range[0]) {
                                        break;
                                    }
                                    if (comment.extendedRange[1] === node.range[0]) {
                                        if (!node.leadingComments) {
                                            node.leadingComments = [];
                                        }
                                        node.leadingComments.push(comment);
                                        comments.splice(cursor, 1);
                                    } else {
                                        cursor += 1;
                                    }
                                }
                                if (cursor === comments.length) {
                                    return VisitorOption.Break;
                                }
                                if (comments[cursor].extendedRange[0] > node.range[1]) {
                                    return VisitorOption.Skip;
                                }
                            }
                        });
                        cursor = 0;
                        traverse(tree, {
                            leave: function(node) {
                                var comment;
                                while (cursor < comments.length) {
                                    comment = comments[cursor];
                                    if (node.range[1] < comment.extendedRange[0]) {
                                        break;
                                    }
                                    if (node.range[1] === comment.extendedRange[0]) {
                                        if (!node.trailingComments) {
                                            node.trailingComments = [];
                                        }
                                        node.trailingComments.push(comment);
                                        comments.splice(cursor, 1);
                                    } else {
                                        cursor += 1;
                                    }
                                }
                                if (cursor === comments.length) {
                                    return VisitorOption.Break;
                                }
                                if (comments[cursor].extendedRange[0] > node.range[1]) {
                                    return VisitorOption.Skip;
                                }
                            }
                        });
                        return tree;
                    }
                    exports.version = "1.3.3-dev";
                    exports.Syntax = Syntax;
                    exports.traverse = traverse;
                    exports.replace = replace;
                    exports.attachComments = attachComments;
                    exports.VisitorKeys = VisitorKeys;
                    exports.VisitorOption = VisitorOption;
                    exports.Controller = Controller;
                });
            });
            require("/tools/entry-point.js");
        }).call(exports, exports);
    })(codeGetExports);
    JSParsingTools.generate = codeGetExports.escodegen.generate;
    if (typeof module !== "undefined" && module.exports) {
        module.exports = JSParsingTools;
    } else {
        HammerheadClient.define("Shared.JSParsingTools", function() {
            this.exports = JSParsingTools;
        });
    }
})();